/*
  keycode.c : I/F of Fep Bridge 1.33
    Copyright (C) Yutaka Tamiya (tamy@flab.fujitsu.co) 1996/11/13
*/

static struct {
    unsigned short	noShift;
    unsigned short	shift;
    unsigned short	control;
    unsigned short	alt;
} VirtualKeyCodeToBiosKeyCode[256] = {
	{0, 	0,	0,	0},	/* 0x00 None	 	*/
	{0,	0,	0,	0},	/* 0x01 VK_LBUTTON	*/
	{0,	0,	0,	0},	/* 0x02 VK_RBUTTON	*/
	{0,	0,	0,	0},	/* 0x03 VK_CANCEL	*/
	{0,	0,	0,	0},	/* 0x04 VK_MBUTTON	*/
	{0,	0,	0,	0},	/* 0x05 Undefined	*/
	{0,	0,	0,	0},	/* 0x06 Undefined	*/
	{0,	0,	0,	0},	/* 0x07 Undefined	*/
	{0x53,	0x53,	0x93,	0x93},	/* 0x08 VK_BACK		*/
	{0x0f,	0x0f,	0x94,	0x95},	/* 0x09 VK_TAB	 	*/
	{0,	0,	0,	0},	/* 0x0a Undefined	*/
	{0,	0,	0,	0},	/* 0x0b Undefined	*/
	{0x4c,	0x4c,	0x4c,	0x4c},	/* 0x0c VK_CLEAR	*/
	{0x1c,	0x1c,	0x1c,	0x96},	/* 0x0d VK_ENTER	*/
	{0,	0,	0,	0},	/* 0x0e Undefined	*/
	{0,	0,	0,	0},	/* 0x0f Undefined	*/
	{0x2a,	0x2a,	0x2a,	0x2a},	/* 0x10 VK_SHIFT	*/
	{0x1d,	0x1d,	0x1d,	0x1d},	/* 0x11 VK_CONTROL	*/
	{0x38,	0x38,	0x38,	0x38},	/* 0x12 VK_MENU		*/
	{0x65,	0x65,	0x65,	0x65},	/* 0x13 VK_PAUSE	*/
	{0xb3,	0x3a,	0xb4,	0xb5},	/* 0x14 VK_CAPITAL	*/
	{0xb6,	0xb7,	0xb8,	0xb9},	/* 0x15 VK_KANA		*/
	{0,	0,	0,	0},	/* 0x16 Reserved for Kanji systems */
	{0,	0,	0,	0},	/* 0x17 Reserved for Kanji systems */
	{0,	0,	0,	0},	/* 0x18 VK_FINAL (Korea) (??)	*/
	{0xaf,	0xb0,	0xb1,	0xb2},	/* 0x19 VK_KANJI	*/
	{0,	0,	0,	0},	/* 0x1a Undefined	*/
	{0x01,	0x01,	0x01,	0x01},	/* 0x1b VK_ESCAPE	*/
	{0xa7,	0xa8,	0xa9,	0xaa},	/* 0x1c VK_CONVERT	*/
	{0xab,	0xac,	0xad,	0xae},	/* 0x1d VK_NONCONVERT	*/
	{0,	0,	0,	0},	/* 0x1e VK_ACCEPT (??)	*/
	{0,	0,	0,	0},	/* 0x1f VK_MODECHANGE (??)	*/
	{0x39,	0x39,	0x39,	0x39},	/* 0x20 VK_SPACE	*/
	{0x49,	0x49,	0x84,	0x99},	/* 0x21 VK_PRIOR	*/
	{0x51,	0x51,	0x76,	0x91},	/* 0x22 VK_NEXT		*/
	{0x4f,	0x4f,	0x75,	0x9f},	/* 0x23 VK_END		*/
	{0x47,	0x47,	0x77,	0x97},	/* 0x24 VK_HOME		*/
	{0x4b,	0x4b,	0x73,	0x9b},	/* 0x25 VK_LEFT		*/
	{0x48,	0x48,	0x8d,	0x98},	/* 0x26 VK_UP		*/
	{0x4d,	0x4d,	0x74,	0x4d},	/* 0x27 VK_RIGHT	*/
	{0x50,	0x50,	0x91,	0x90},	/* 0x28 VK_DOWN		*/
	{0,	0,	0,	0},	/* 0x29 VK_SELECT	*/
	{0,	0,	0,	0},	/* 0x2a OEM specific	*/
	{0,	0,	0,	0},	/* 0x2b VK_EXECUTE	*/
	{0x63,	0x63,	0x63,	0x63},	/* 0x2c VK_SNAPSHOT	*/
	{0x52,	0x52,	0x92,	0x92},	/* 0x2d VK_INSERT	*/
	{0x0e,	0x0e,	0x0e,	0x0e},	/* 0x2e VK_DELETE	*/
	{0,	0,	0,	0},	/* 0x2f VK_HELP		*/
	{0x0b,	0x0b,	0x0b,	0x81},	/* 0x30 VK_0		*/
	{0x02,	0x02,	0x02,	0x78},	/* 0x31 VK_1		*/
	{0x03,	0x03,	0x03,	0x79},	/* 0x32 VK_2		*/
	{0x04,	0x04,	0x04,	0x7a},	/* 0x33 VK_3		*/
	{0x05,	0x05,	0x05,	0x7b},	/* 0x34 VK_4		*/
	{0x06,	0x06,	0x06,	0x7c},	/* 0x35 VK_5		*/
	{0x07,	0x07,	0x07,	0x7d},	/* 0x36 VK_6		*/
	{0x08,	0x08,	0x08,	0x7e},	/* 0x37 VK_7		*/
	{0x09,	0x09,	0x09,	0x7f},	/* 0x38 VK_8		*/
	{0x0a,	0x0a,	0x0a,	0x80},	/* 0x39 VK_9		*/
	{0,	0,	0,	0},	/* 0x3a Undefined	*/
	{0,	0,	0,	0},	/* 0x3b Undefined	*/
	{0,	0,	0,	0},	/* 0x3c Undefined	*/
	{0,	0,	0,	0},	/* 0x3d Undefined	*/
	{0,	0,	0,	0},	/* 0x3e Undefined	*/
	{0,	0,	0,	0},	/* 0x3f Undefined	*/
	{0,	0,	0,	0},	/* 0x40 Undefined	*/
	{0x1e,	0x1e,	0x1e,	0x1e},	/* 0x41 VK_A		*/
	{0x30,	0x30,	0x30,	0x30},	/* 0x42 VK_B		*/
	{0x2e,	0x2e,	0x2e,	0x2e},	/* 0x43 VK_C		*/
	{0x20,	0x20,	0x20,	0x20},	/* 0x44 VK_D		*/
	{0x12,	0x12,	0x12,	0x12},	/* 0x45 VK_E		*/
	{0x21,	0x21,	0x21,	0x21},	/* 0x46 VK_F		*/
	{0x22,	0x22,	0x22,	0x22},	/* 0x47 VK_G		*/
	{0x23,	0x23,	0x23,	0x23},	/* 0x48 VK_H		*/
	{0x17,	0x17,	0x17,	0x17},	/* 0x49 VK_I		*/
	{0x24,	0x24,	0x24,	0x24},	/* 0x4a VK_J		*/
	{0x25,	0x25,	0x25,	0x25},	/* 0x4b VK_K		*/
	{0x26,	0x26,	0x26,	0x26},	/* 0x4c VK_L		*/
	{0x32,	0x32,	0x32,	0x32},	/* 0x4d VK_M		*/
	{0x31,	0x31,	0x31,	0x31},	/* 0x4e VK_N		*/
	{0x18,	0x18,	0x18,	0x18},	/* 0x4f VK_O		*/
	{0x19,	0x19,	0x19,	0x19},	/* 0x50 VK_P		*/
	{0x10,	0x10,	0x10,	0x10},	/* 0x51 VK_Q		*/
	{0x13,	0x13,	0x13,	0x13},	/* 0x52 VK_R		*/
	{0x1f,	0x1f,	0x1f,	0x1f},	/* 0x53 VK_S		*/
	{0x14,	0x14,	0x14,	0x14},	/* 0x54 VK_T		*/
	{0x16,	0x16,	0x16,	0x16},	/* 0x55 VK_U		*/
	{0x2f,	0x2f,	0x2f,	0x2f},	/* 0x56 VK_V		*/
	{0x11,	0x11,	0x11,	0x11},	/* 0x57 VK_W		*/
	{0x2d,	0x2d,	0x2d,	0x2d},	/* 0x58 VK_X		*/
	{0x15,	0x15,	0x15,	0x15},	/* 0x59 VK_Y		*/
	{0x2c,	0x2c,	0x2c,	0x2c},	/* 0x5a VK_Z		*/
	{0,	0,	0,	0},	/* 0x5b Undefined	*/
	{0,	0,	0,	0},	/* 0x5c Undefined	*/
	{0,	0,	0,	0},	/* 0x5d Undefined	*/
	{0,	0,	0,	0},	/* 0x5e Undefined	*/
	{0,	0,	0,	0},	/* 0x5f Undefined	*/
	{0x52,	0x52,	0x52,	0x52},	/* 0x60 VK_NUMPAD0	*/
	{0x4f,	0x4f,	0x4f,	0x4f},	/* 0x61 VK_NUMPAD1	*/
	{0x50,	0x50,	0x50,	0x50},	/* 0x62 VK_NUMPAD2	*/
	{0x51,	0x51,	0x51,	0x51},	/* 0x63 VK_NUMPAD3	*/
	{0x4b,	0x4b,	0x4b,	0x4b},	/* 0x64 VK_NUMPAD4	*/
	{0x4c,	0x4c,	0x4c,	0x4c},	/* 0x65 VK_NUMPAD5	*/
	{0x4d,	0x4d,	0x4d,	0x4d},	/* 0x66 VK_NUMPAD6	*/
	{0x47,	0x47,	0x47,	0x47},	/* 0x67 VK_NUMPAD7	*/
	{0x48,	0x48,	0x48,	0x48},	/* 0x68 VK_NUMPAD8	*/
	{0x49,	0x49,	0x49,	0x49},	/* 0x69 VK_NUMPAD9	*/
	{0x37,	0x37,	0x37,	0x37},	/* 0x6a VK_MULTIPLY	*/
	{0x4e,	0x4e,	0x4e,	0x4e},	/* 0x6b VK_ADD		*/
	{0x00,	0x00,	0x00,	0x00},	/* 0x6c	VK_SEPARATOR	*/
	{0x4a,	0x4a,	0x4a,	0x4a},	/* 0x6d VK_SUBTRACT	*/
	{0x53,	0x53,	0x53,	0x53},	/* 0x6e VK_DECIMAL	*/
	{0x62,	0x62,	0x62,	0x62},	/* 0x6f	VK_DIVICE	*/
	{0x3b,	0x54,	0x5e,	0x68},	/* 0x70 VK_F1		*/
	{0x3c,	0x55,	0x5f,	0x69},	/* 0x71 VK_F2		*/
	{0x3d,	0x56,	0x60,	0x6a},	/* 0x72 VK_F3		*/
	{0x3e,	0x57,	0x61,	0x6b},	/* 0x73 VK_F4		*/
	{0x3f,	0x58,	0x62,	0x6c},	/* 0x74 VK_F5		*/
	{0x40,	0x59,	0x63,	0x6d},	/* 0x75 VK_F6		*/
	{0x41,	0x5a,	0x64,	0x6e},	/* 0x76 VK_F7		*/
	{0x42,	0x5b,	0x65,	0x6f},	/* 0x77 VK_F8		*/
	{0x43,	0x5c,	0x66,	0x70},	/* 0x78 VK_F9		*/
	{0x44,	0x5d,	0x67,	0x71},	/* 0x79 VK_F10		*/
	{0x85,	0x87,	0x89,	0x8b},	/* 0x7a VK_F11		*/
	{0x86,	0x88,	0x8a,	0x8c},	/* 0x7b VK_F12		*/
	{0,	0,	0,	0},	/* 0x7c VK_F13		*/
	{0,	0,	0,	0},	/* 0x7d VK_F14		*/
	{0,	0,	0,	0},	/* 0x7e VK_F15		*/
	{0,	0,	0,	0},	/* 0x7f VK_F16		*/
	{0,	0,	0,	0},	/* 0x80 VK_F17		*/
	{0,	0,	0,	0},	/* 0x81 VK_F18		*/
	{0,	0,	0,	0},	/* 0x82 VK_F19		*/
	{0,	0,	0,	0},	/* 0x83 VK_F20		*/
	{0,	0,	0,	0},	/* 0x84 VK_F21		*/
	{0,	0,	0,	0},	/* 0x85 VK_F22		*/
	{0,	0,	0,	0},	/* 0x86 VK_F23		*/
	{0,	0,	0,	0},	/* 0x87 VK_F24		*/
	{0,	0,	0,	0},	/* 0x88 Unassigned	*/
	{0,	0,	0,	0},	/* 0x89 Unassigned	*/
	{0,	0,	0,	0},	/* 0x8a Unassigned	*/
	{0,	0,	0,	0},	/* 0x8b Unassigned	*/
	{0,	0,	0,	0},	/* 0x8c Unassigned	*/
	{0,	0,	0,	0},	/* 0x8d Unassigned	*/
	{0,	0,	0,	0},	/* 0x8e Unassigned	*/
	{0,	0,	0,	0},	/* 0x8f Unassigned	*/
	{0x45,	0x45,	0x45,	0x45},	/* 0x90 VK_NUMLOCK	*/
	{0x46,	0x46,	0x46,	0x46},	/* 0x91 VK_SCROLL	*/
	{0x27,	0x27,	0x27,	0x27},	/* 0xba VK_US_SEMICOLON	*/
	{0x0d,	0x0d,	0x0d,	0x83},	/* 0xbb VK_US_EQUAL	*/
	{0x33,	0x33,	0x33,	0x33},	/* 0xbc VK_US_COMMA	*/
	{0x0c,	0x0c,	0x0c,	0x82},	/* 0xbd VK_US_MINUS	*/
	{0x34,	0x34,	0x34,	0x34},	/* 0xbe VK_US_PERIOD	*/
	{0x35,	0x35,	0x35,	0x35},	/* 0xbf VK_US_SLASH	*/
	{0x29,	0x29,	0x29,	0x29},	/* 0xc0 VK_US_GRAVE	*/
	{0,	0,	0,	0},	/* 0xc1 Unassigned	*/
	{0,	0,	0,	0},	/* 0xc2 Unassigned	*/
	{0,	0,	0,	0},	/* 0xc3 Unassigned	*/
	{0,	0,	0,	0},	/* 0xc4 Unassigned	*/
	{0,	0,	0,	0},	/* 0xc5 Unassigned	*/
	{0,	0,	0,	0},	/* 0xc6 Unassigned	*/
	{0,	0,	0,	0},	/* 0xc7 Unassigned	*/
	{0,	0,	0,	0},	/* 0xc8 Unassigned	*/
	{0,	0,	0,	0},	/* 0xc9 Unassigned	*/
	{0,	0,	0,	0},	/* 0xca Unassigned	*/
	{0,	0,	0,	0},	/* 0xcb Unassigned	*/
	{0,	0,	0,	0},	/* 0xcc Unassigned	*/
	{0,	0,	0,	0},	/* 0xcd Unassigned	*/
	{0,	0,	0,	0},	/* 0xce Unassigned	*/
	{0,	0,	0,	0},	/* 0xcf Unassigned	*/
	{0,	0,	0,	0},	/* 0xd0 Unassigned	*/
	{0,	0,	0,	0},	/* 0xd1 Unassigned	*/
	{0,	0,	0,	0},	/* 0xd2 Unassigned	*/
	{0,	0,	0,	0},	/* 0xd3 Unassigned	*/
	{0,	0,	0,	0},	/* 0xd4 Unassigned	*/
	{0,	0,	0,	0},	/* 0xd5 Unassigned	*/
	{0,	0,	0,	0},	/* 0xd6 Unassigned	*/
	{0,	0,	0,	0},	/* 0xd7 Unassigned	*/
	{0,	0,	0,	0},	/* 0xd8 Unassigned	*/
	{0,	0,	0,	0},	/* 0xd9 Unassigned	*/
	{0,	0,	0,	0},	/* 0xda Unassigned	*/
	{0x1a,	0x1a,	0x1a,	0x1a},	/* 0xdb VK_US_BRACELEFT	*/
	{0x2b,	0x2b,	0x2b,	0x2b},	/* 0xdc VK_US_BACKSLASH	*/
	{0x1b,	0x1b,	0x1b,	0x1b},	/* 0xdd VK_US_BRACERIGHT*/
	{0x28,	0x28,	0x28,	0x28},	/* 0xde VK_US_APOSTROPHE*/
	{0,	0,	0,	0},	/* 0xdf OEM specific	*/
	{0,	0,	0,	0},	/* 0xe0 OEM specific	*/
	{0,	0,	0,	0},	/* 0xe1 OEM specific	*/
	{0,	0,	0,	0},	/* 0xe2 OEM specific	*/
	{0,	0,	0,	0},	/* 0xe3 OEM specific	*/
	{0,	0,	0,	0},	/* 0xe4 OEM specific	*/
	{0,	0,	0,	0},	/* 0xe5 Unassigned	*/
	{0,	0,	0,	0},	/* 0xe6 OEM specific	*/
	{0,	0,	0,	0},	/* 0xe7 Unassigned	*/
	{0,	0,	0,	0},	/* 0xe8 Unassigned	*/
	{0,	0,	0,	0},	/* 0xe9 OEM specific	*/
	{0,	0,	0,	0},	/* 0xea OEM specific	*/
	{0,	0,	0,	0},	/* 0xeb OEM specific	*/
	{0,	0,	0,	0},	/* 0xec OEM specific	*/
	{0,	0,	0,	0},	/* 0xed OEM specific	*/
	{0,	0,	0,	0},	/* 0xee OEM specific	*/
	{0,	0,	0,	0},	/* 0xef OEM specific	*/
	{0,	0,	0,	0},	/* 0xf0 OEM specific	*/
	{0,	0,	0,	0},	/* 0xf1 OEM specific	*/
	{0,	0,	0,	0},	/* 0xf2 OEM specific	*/
	{0,	0,	0,	0},	/* 0xf3 OEM specific	*/
	{0,	0,	0,	0},	/* 0xf4 OEM specific	*/
	{0,	0,	0,	0},	/* 0xf5 OEM specific	*/
	{0,	0,	0,	0},	/* 0xf6 Unassigned	*/
	{0,	0,	0,	0},	/* 0xf7 Unassigned	*/
	{0,	0,	0,	0},	/* 0xf8 Unassigned	*/
	{0,	0,	0,	0},	/* 0xf9 Unassigned	*/
	{0,	0,	0,	0},	/* 0xfa Unassigned	*/
	{0,	0,	0,	0},	/* 0xfb Unassigned	*/
	{0,	0,	0,	0},	/* 0xfc Unassigned	*/
	{0,	0,	0,	0},	/* 0xfd Unassigned	*/
	{0,	0,	0,	0},	/* 0xfe Unassigned	*/
	{0,	0,	0,	0},	/* 0xff None		*/
};

static unsigned short AsciiToBiosKeyCode[256] = {
	0x00, 0x1e, 0x30, 0x2e, 0x20, 0x12, 0x21, 0x22,
	0x23, 0x17, 0x24, 0x25, 0x26, 0x32, 0x31, 0x18,
	0x19, 0x10, 0x13, 0x1f, 0x14, 0x16, 0x2f, 0x11,
	0x2d, 0x15, 0x2c, 0x1a, 0x2b, 0x1b, 0x00, 0x00,
	0x39, 0x02, 0x28, 0x04, 0x05, 0x06, 0x08, 0x28,
	0x0a, 0x0b, 0x09, 0x0d, 0x33, 0x0c, 0x34, 0x35,
	0x0b, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08,
	0x09, 0x0a, 0x27, 0x27, 0x33, 0x0d, 0x34, 0x35,
	0x03, 0x1e, 0x30, 0x2e, 0x20, 0x12, 0x21, 0x22,
	0x23, 0x17, 0x24, 0x25, 0x26, 0x32, 0x31, 0x18,
	0x19, 0x10, 0x13, 0x1f, 0x14, 0x16, 0x2f, 0x11,
	0x2d, 0x15, 0x2c, 0x1a, 0x2b, 0x1b, 0x07, 0x0c,
	0x29, 0x1e, 0x30, 0x2e, 0x20, 0x12, 0x21, 0x22,
	0x23, 0x17, 0x24, 0x25, 0x26, 0x32, 0x31, 0x18,
	0x19, 0x10, 0x13, 0x1f, 0x14, 0x16, 0x2f, 0x11,
	0x2d, 0x15, 0x2c, 0x1a, 0x2b, 0x1b, 0x29, 0x00,
};

