;;;; $Id: fep-client.el,v 1.12 1995/12/15 04:56:55 ueda Exp ueda $
;;;;
;; Fep server interface for Egg
;; Coded by Y.Ueda (ueda@kyoto-u.ac.jp)
;; Modified by Y. Tamiya (tamy@flab.fujitsu.co.jp) 1996/11/13

;; This file is part of Egg on Mule (Multilingual Environment)

;; Egg is distributed in the forms of patches to GNU
;; Emacs under the terms of the GNU EMACS GENERAL PUBLIC
;; LICENSE which is distributed along with GNU Emacs by the
;; Free Software Foundation.

;; Egg is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied
;; warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
;; PURPOSE.  See the GNU EMACS GENERAL PUBLIC LICENSE for
;; more details.

;; You should have received a copy of the GNU EMACS GENERAL
;; PUBLIC LICENSE along with Nemacs; see the file COPYING.
;; If not, write to the Free Software Foundation, 675 Mass
;; Ave, Cambridge, MA 02139, USA.

;;;
;;; Mule - Fep server interface in elisp
;;;

(provide 'fep-client)

;;;
;;;  Fep deamon command constants
;;;

; Commands of Protocol
(defconst fep:FO_Info	1 "Information of Fep Bridge")
(defconst fep:FO_Ack	2 "Acknoledgement to connect with Fep Bridge")
(defconst fep:FO_Result	5 "Result of japanese conversion")
(defconst fep:FO_Alive	3 "Fep Bridge is alive")
(defconst fep:FO_Dead	4 "Fep Bridge is dead")
(defconst fep:FI_Info	1 "Send me information of Fep Bridge")
(defconst fep:FI_Auth	3 "User authorization check")
(defconst fep:FI_Reset	2 "Reset Fep Bridge")
(defconst fep:FI_Conv1	4 "Request for japanese conversion")
(defconst fep:FI_ConvN	6 "Request for japanese conversion")
(defconst fep:FI_Stat	5 "Is Fep Bridge alive ?")
(defconst fep:FI_Kill	7 "Kill Fep Bridge")

; Virtual key codes of MS-Windows
(defconst fep:VK_SHIFT_MASK	1   "Shift modifier")
(defconst fep:VK_CONTROL_MASK	2   "Control modifier")
(defconst fep:VK_ALT_MASK	4   "Alt modifier")

(defconst fep:VK_LBUTTON	1   "Left mouse button")
(defconst fep:VK_RBUTTON	2   "Right mouse button")
(defconst fep:VK_CANCEL		3   "Used for control-break processing")
(defconst fep:VK_MBUTTON	4   "Middle mouse button (three-button mouse)")
(defconst fep:VK_BACK		8   "BACKSPACE key")
(defconst fep:VK_TAB		9   "TAB key")
(defconst fep:VK_CLEAR		12  "CLEAR key")
(defconst fep:VK_RETURN		13  "ENTER key")
(defconst fep:VK_SHIFT		16  "SHIFT key")
(defconst fep:VK_CONTROL	17  "CTRL key")
(defconst fep:VK_MENU		18  "ALT key")
(defconst fep:VK_PAUSE		19  "PAUSE key")
(defconst fep:VK_CAPITAL	20  "CAPS LOCK key")
(defconst fep:VK_KANA		21  "Kakana/Hiragana key")
(defconst fep:VK_KANJI		25  "Kanji key")
(defconst fep:VK_ESCAPE		27  "ESC key")
(defconst fep:VK_CONVERT	28  "Henkan key")
(defconst fep:VK_NONCONVERT	29  "MuHenkan key")
(defconst fep:VK_SPACE		32  "SPACEBAR")
(defconst fep:VK_PRIOR		33  "PAGE UP key")
(defconst fep:VK_NEXT		34  "PAGE DOWN key")
(defconst fep:VK_END		35  "END key")
(defconst fep:VK_HOME		36  "HOME key")
(defconst fep:VK_LEFT		37  "LEFT ARROW key")
(defconst fep:VK_UP		38  "UP ARROW key")
(defconst fep:VK_RIGHT		39  "RIGHT ARROW key")
(defconst fep:VK_DOWN		40  "DOWN ARROW key")
(defconst fep:VK_SELECT		41  "SELECT key")
(defconst fep:VK_EXECUTE	43  "EXECUTE key")
(defconst fep:VK_SNAPSHOT	44  "PRINT SCREEN key")
(defconst fep:VK_INSERT		45  "INS key")
(defconst fep:VK_DELETE		46  "DEL key")
(defconst fep:VK_HELP		47  "HELP key")
(defconst fep:VK_0		48  "0 key")
(defconst fep:VK_1		49  "1 key")
(defconst fep:VK_2		50  "2 key")
(defconst fep:VK_3		51  "3 key")
(defconst fep:VK_4		52  "4 key")
(defconst fep:VK_5		53  "5 key")
(defconst fep:VK_6		54  "6 key")
(defconst fep:VK_7		55  "7 key")
(defconst fep:VK_8		56  "8 key")
(defconst fep:VK_9		57  "9 key")
(defconst fep:VK_A		65  "A key")
(defconst fep:VK_B		66  "B key")
(defconst fep:VK_C		67  "C key")
(defconst fep:VK_D		68  "D key")
(defconst fep:VK_E		69  "E key")
(defconst fep:VK_F		70  "F key")
(defconst fep:VK_G		71  "G key")
(defconst fep:VK_H		72  "H key")
(defconst fep:VK_I		73  "I key")
(defconst fep:VK_J		74  "J key")
(defconst fep:VK_K		75  "K key")
(defconst fep:VK_L		76  "L key")
(defconst fep:VK_M		77  "M key")
(defconst fep:VK_N		78  "N key")
(defconst fep:VK_O		79  "O key")
(defconst fep:VK_P		80  "P key")
(defconst fep:VK_Q		81  "Q key")
(defconst fep:VK_R		82  "R key")
(defconst fep:VK_S		83  "S key")
(defconst fep:VK_T		84  "T key")
(defconst fep:VK_U		85  "U key")
(defconst fep:VK_V		86  "V key")
(defconst fep:VK_W		87  "W key")
(defconst fep:VK_X		88  "X key")
(defconst fep:VK_Y		89  "Y key")
(defconst fep:VK_Z		90  "Z key")
(defconst fep:VK_NUMPAD0	96  "Numeric keypad 0 key")
(defconst fep:VK_NUMPAD1	97  "Numeric keypad 1 key")
(defconst fep:VK_NUMPAD2	98  "Numeric keypad 2 key")
(defconst fep:VK_NUMPAD3	99  "Numeric keypad 3 key")
(defconst fep:VK_NUMPAD4	100 "Numeric keypad 4 key")
(defconst fep:VK_NUMPAD5	101 "Numeric keypad 5 key")
(defconst fep:VK_NUMPAD6	102 "Numeric keypad 6 key")
(defconst fep:VK_NUMPAD7	103 "Numeric keypad 7 key")
(defconst fep:VK_NUMPAD8	104 "Numeric keypad 8 key")
(defconst fep:VK_NUMPAD9	105 "Numeric keypad 9 key")
(defconst fep:VK_MULTIPLY	106 "Multiply key")
(defconst fep:VK_ADD		107 "Add key")
(defconst fep:VK_SEPARATOR	108 "Separator key")
(defconst fep:VK_SUBTRACT	109 "Subtract key")
(defconst fep:VK_DECIMAL	110 "Decimal key")
(defconst fep:VK_DIVIDE		111 "Divide key")
(defconst fep:VK_F1		112 "F1 key")
(defconst fep:VK_F2		113 "F2 key")
(defconst fep:VK_F3		114 "F3 key")
(defconst fep:VK_F4		115 "F4 key")
(defconst fep:VK_F5		116 "F5 key")
(defconst fep:VK_F6		117 "F6 key")
(defconst fep:VK_F7		118 "F7 key")
(defconst fep:VK_F8		119 "F8 key")
(defconst fep:VK_F9		120 "F9 key")
(defconst fep:VK_F10		121 "F10 key")
(defconst fep:VK_F11		122 "F11 key")
(defconst fep:VK_F12		123 "F12 key")
(defconst fep:VK_F13		124 "F13 key")
(defconst fep:VK_F14		125 "F14 key")
(defconst fep:VK_F15		126 "F15 key")
(defconst fep:VK_F16		127 "F16 key")
(defconst fep:VK_F17		128 "F17 key")
(defconst fep:VK_F18		129 "F18 key")
(defconst fep:VK_F19		130 "F19 key")
(defconst fep:VK_F20		131 "F20 key")
(defconst fep:VK_F21		132 "F21 key")
(defconst fep:VK_F22		133 "F22 key")
(defconst fep:VK_F23		134 "F23 key")
(defconst fep:VK_F24		135 "F24 key")
(defconst fep:VK_NUMLOCK	144 "NUM LOCK key")
(defconst fep:VK_SCROLL		145 "SCROLL LOCK key")
; for US 101 keyboard
(defconst fep:VK_US_SEMICOLON	186 "; key")
(defconst fep:VK_US_EQUAL	187 "= key")
(defconst fep:VK_US_COMMA	188 ", key")
(defconst fep:VK_US_MINUS	189 "- key")
(defconst fep:VK_US_PERIOD	190 ". key")
(defconst fep:VK_US_SLASH	191 "/ key")
(defconst fep:VK_US_GRAVE	192 "` key")
(defconst fep:VK_US_BRACELEFT	219 "[ key")
(defconst fep:VK_US_BACKSLASH	220 "\ key")
(defconst fep:VK_US_BRACERIGHT	221 "] key")
(defconst fep:VK_US_APOSTROPHE	222 "' key")

; Variables for storing conversion results
(defvar fep:*packet-id* 	0   "Packet ID from Fep Bridge.")
(defvar fep:*return-code*	0   "Result value of Fep Bridge.")
(defvar fep:*result-len* 	0   "Length of FEP result string.")
(defvar fep:*result-str*	nil "FEP result string.")
(defvar fep:*mode-len*		0   "Length of FEP mode string.")
(defvar fep:*mode-str*		nil "FEP mode string.")
;(defvar fep:*mode-attr*	nil "Attributes of FEP mode string.")
(defvar fep:*system-len*	0   "Length of FEP system string.")
(defvar fep:*system-str*	nil "FEP system string.")
(defvar fep:*system-attr*	nil "Attributes of FEP system string.")
(defvar fep:*init-system-str*	nil "Initial FEP system string.")
(defvar fep:*buf-len*		0   "Length of FEP buffer string.")
(defvar fep:*buf-str*		nil "FEP buffer string.")
(defvar fep:*buf-attr*		nil "Attributes of FEP buffer string.")

(defvar fep:*buf-attr-mask*	240	"Bit mask for FEP's buffer string")
(defvar fep:*buf-attr-mask2*	  0	"Bit mask for FEP's buffer string")
(defvar fep:*system-attr-mask*	240	"Bit mask for FEP's system string")

(defvar fep:*bunsetu-start*	nil "Start Potion of Current Bunsetu.")
(defvar fep:*bunsetu-end*	nil "End Potion of Current Bunsetu.")

;;;
;;; Variables for Emacs
;;;
(defvar fep:*current-buffer*	nil	"Current buffer for conversion")
(defvar fep:*japanese-mode*	nil	"Fep Bridge japanese mode.")
(defvar fep:*preedit-start* (make-marker)
					"Start point of the preedit region")
(defvar fep:*preedit-end* (make-marker) "End point of the preedit region")

(defvar fep:*attr-underline*	nil	"Underline attr in the preedit region")
(defvar fep:*system-seg-list*	nil	"Segments of system region")
(defvar fep:*buffer*  nil	"Buffer associated with Fep Bridge process.")
(defvar fep:*process* nil	"Fep Bridge process.")
(defvar fep:*command* nil	"Command buffer to be sent to Fep Bridge.")
(defvar fep:*mode-line-format*	nil	"Mode line format for Fep Bridge.")
(defvar fep:*mode-line-format-backup*	nil	"Backup of mode line format.")
(defvar fep:*local-map-backup*	nil	"Backup of keymap")
(defvar fep:*mode-map* (make-keymap)	"Keymap for Fep Bridge")
(defvar fep:*menu-buffer*	nil	"Buffer to display fep:*system-str*")
(defvar fep:*saved-minibuffer*	nil	"Saved buffer displayed in minibuffer")

;;;  Fep server version error
(defconst FEP_DifferentVersion 11 "")
(defvar *fep-current-server-version* nil)

(defvar fep-server-buffer nil  "Buffer associated with Fep server process.")

(defvar fep-server-process nil  "Fep Kana Kanji hankan process.")

(defvar fep-command-tail-position nil)
(defvar fep-command-buffer nil)

(defvar fep-henkan-string nil)

(defvar fep-return-code nil)
(defvar fep-error-code nil)

;;;
;;;  Put data into the command buffer 
;;;
(defun fep:put-1byte (integer)
  (insert (logand 255 integer)))

(defun fep:put-int2 (integer)
  (insert (logand 255 (lsh integer -8))
	  (logand 255 integer)))

;;;
;;; Get data from the process buffer
;;;
(defun fep:get-1byte ()
  (let ((c 0) (point (point)))
    (while (< (point-max) (+ point 1))
      (accept-process-output)
      (if (= c 10000) (error "Count exceed."))
      (setq c (1+ c)))
    (goto-char point))
  (let ((point (point)))
    (prog1
	(char-after (point))
      (goto-char (+ (point) 1)))))

(defun fep:get-int2 ()
  (let ((c 0) (point (point)))
    ;;;(goto-char (point-min))
    (while (< (point-max) (+ point 2))
      (accept-process-output)
      (if (= c 10000) (error "Count exceed."))
      (setq c (1+ c)))
    (goto-char point))
  (let ((point (point)))
    (prog1
	(logior 
	 (lsh (char-after point)       8)
	 (lsh (char-after (+ point 1)) 0))
      (goto-char (+ (point) 2)))))

(defun fep:get-nbyte (len)
  (let ((point (point)) (c 0))
      (while (< (point-max) (+ point len))
	(accept-process-output)
	(if (= c 10000) (error "Count exceed."))
	(setq c (1+ c)))
      (progn 
	(goto-char (+ point len))
	(buffer-substring point (point)))))

;;;
;;; Fep Server Command Primitives
;;;

;; (defun fep-command-start (command)
;;   (set-buffer fep-command-buffer)
;;   (goto-char (point-min))
;;   (if (not (= (point-max) (+ fep-command-tail-position 1024)))
;;       (error "fep command start error"))
;;   (delete-region (point-min) fep-command-tail-position)
;;   (fep-put-4byte command))

(defun fep-command-start (command)
  (set-buffer fep-command-buffer)
  (erase-buffer)
  (fep:put-1byte command)
  (point)
  )

(defun fep-command-reset ()
  (save-excursion
    (progn  
      ;;; for Mule
      (if (fboundp 'set-process-coding-system)
	  (set-process-coding-system fep-server-process *noconv* *noconv*))
      ;;; for Nemacs 3.0 and later
;;      (if (fboundp 'set-process-kanji-code)
;;         (set-process-kanji-code fep-server-process 0))
      (set-buffer fep-command-buffer)
      (setq mc-flag nil)   ;;; for Mule
;;      (setq kanji-flag nil)
;;      (setq kanji-fileio-code 0)   ;;; for Nemacs 2.1
      (buffer-disable-undo fep-command-buffer)
      (erase-buffer)
      (setq fep-command-tail-position (point-min))
      (let ((max 1024) (i 0))
	(while (< i max)
	  (insert 0)
	  (setq i (1+ i)))))))

(defun fep-command-end ()
  (set-buffer fep-server-buffer)
  (erase-buffer)
  (set-buffer fep-command-buffer)
  (process-send-region fep-server-process (point-min) (point)))

;;;
;;; Get conversion result
;;;
(defun fep:get-result ()
  (let ((str nil) (attr nil))
    (set-buffer fep-server-buffer);fep:*buffer*)
    (prog2
     (setq fep:*packet-id*	(fep:get-1byte))
     (setq fep:*return-code*	(fep:get-1byte))
     (if (eq fep:*return-code* 0)
       (progn
	 (setq fep:*result-len*	(fep:get-int2))
	 (setq fep:*mode-len*	(fep:get-int2))
	 (setq fep:*system-len*	(fep:get-int2))
	 (setq fep:*buf-len*	(fep:get-int2))
	 (setq fep:*cursor-pos*	(fep:get-int2))

	 (setq fep:*result-str*
	       (code-convert-string
		(fep:get-nbyte fep:*result-len*) *sjis* *internal*))

	 (setq str	(fep:get-nbyte fep:*mode-len*))
	 (setq attr	(fep:get-nbyte fep:*mode-len*))
	 ;(setq fep:*mode-attr*
	 ;	   (fep:attr-convert fep:*mode-len* str attr))
;         (setq fep:*mode-str*	(code-convert-string str *sjis* *internal*))

	 (setq str	(fep:get-nbyte fep:*system-len*))
	 (setq attr	(fep:get-nbyte fep:*system-len*))
	 (setq fep:*system-attr* (fep:attr-convert fep:*system-len* str attr))
	 (setq fep:*system-str* (code-convert-string str *sjis* *internal*))

	 (setq str	(fep:get-nbyte fep:*buf-len*))
	 (setq attr	(fep:get-nbyte fep:*buf-len*))
	 (setq fep:*buf-attr* (fep:attr-convert fep:*buf-len* str attr))
	 (setq fep:*buf-str*	(code-convert-string str *sjis* *internal*))
	 (if (string-match " $" fep:*buf-str*)
	     (setq fep:*buf-str* (substring fep:*buf-str* 0 -1)))

	 (fep:set-bunsetu-pos)
	 ;(setq fep:*cursor-pos*
	 ;	   (fep:pos-convert str fep:*cursor-pos*))
	 )))
    ))

;;;
;;; Convert attribute string and cursor positoin for Mule buffer
;;;  Kanji, Hankaku-kana, and ascii character take 3, 2, and 1 byte
;;;  in Mule buffer, respectively.
;;;
(defun fep:attr-convert (len str attr)
  (let ((rtn (make-string (* len 2) 0)) ; enough length
	(i 0) (j 0) (c 0) (d 0))
    (while (< i len)
      (progn
	(setq c (aref str i))
	(setq d (aref attr i))
	(cond
	 ((or (and (<= 129 c) (<= c 159)) (and (<= 224 c) (<= c 252)))
	  ; Sjis-Kanji
	  (aset rtn j d)
	  (aset rtn (+ j 1) d)
	  (aset rtn (+ j 2) d)
	  (setq j (+ j 3))
	  (setq i (+ i 2)))
	 ((and (<= 161 c) (<= c 223))
	  ; Hankaku-Katakana
	  (aset rtn j d)
	  (aset rtn (+ j 1) d)
	  (setq j (+ j 2))
	  (setq i (+ i 1)))
	 (t
	  ; Ascii char
	  (aset rtn j d)
	  (setq j (+ j 1))
	  (setq i (+ i 1))))))
    (substring rtn 0 j)))

;;;
;;; Fep Server Reply primitives
;;;

(defun fep-get-result ()
  (set-buffer fep-server-buffer)
  (condition-case ()
      (accept-process-output fep-server-process)
    (error nil))
  (goto-char (point-min)))

(defun fep-get-return-code ()
  (setq fep-return-code (fep-get-4byte))
  (setq fep-error-code  (if (= fep-return-code 0) nil
			  (fep-error-symbol fep-return-code)))
  (if fep-error-code nil
    fep-return-code))

;;;
;;; Fep Server Interface:  fep-server-open
;;;

;(defvar *fep-server-max-kana-string-length* 1000)

(defvar *fep-program-name* "fep-egg-m")
(defvar *fep-service-name* 6622)

(defun fep-server-open (server-host-name login-name)
  (if (fep-server-active-p) t
     (let ((fepserv_name 
	   (if (or (null  server-host-name)
		   (equal server-host-name "")
		   (equal server-host-name "unix"))
	       (system-name)
	     server-host-name))
	  (user_name
	   (if (or (null login-name) (equal login-name ""))
	       (user-login-name)
	     login-name))
	  (host_name (system-name))
	  (program_name 
	   (format "%d.%s" 
		  (string-to-int (substring (make-temp-name "") 1 6))
		  *fep-program-name*)))
      (setq fep-server-process 
	    (condition-case var
		(open-network-stream "*fepbrdg*" " [Fep Bridge Result Buffer] "
				     fepserv_name *fep-service-name* )
	      (error 
	        (cond((string-match "Unknown host" (car (cdr var)))
		      (setq fep-error-code (list ':FEP_UNKNOWN_HOST
						 fepserv_name)))
		     ((string-match "Unknown service" (car (cdr var)))
		      (setq fep-error-code (list ':FEP_UNKNOWN_SERVICE
						 *fep-service-name*)))
		     (t ;;; "Host ... not respoding"
		      (setq fep-error-code ':FEP_SOCK_OPEN_FAIL)))
		     nil)))
      (if (null fep-server-process) nil
	(process-kill-without-query fep-server-process)
	(setq fep:*menu-buffer*
	      (get-buffer-create " [Fep Bridge Menu Buffer]"))
	(setq fep-server-buffer (get-buffer " [Fep Bridge Result Buffer] "))
	(setq fep-command-buffer (get-buffer-create " [Fep Command Buffer] "))

	(save-excursion 
          ;;; for Mule
	  (if (fboundp 'set-process-coding-system)
	      (set-process-coding-system 
	       fep-server-process *noconv* *noconv*))
	  ;;; for Nemacs 3.0 
;;	  (if (fboundp 'set-process-kanji-code)
;;	      (set-process-kanji-code fep-server-process 0))
	  (progn
	    (set-buffer fep-server-buffer)
	    (setq mc-flag nil)   ;;; for Mule
;;	    (setq kanji-flag nil)
	    ;;; for Nemacs 2.1
;;	    (setq kanji-fileio-code 0) 
	    (buffer-disable-undo fep-server-buffer)
	    )
	  (progn
	    (set-buffer fep:*menu-buffer*)
;	    (setq mc-flag nil)   ;;; for Mule
;;	    (setq kanji-flag nil)
	    ;;; for Nemacs 2.1 
;;	    (setq kanji-fileio-code 0)
	    (buffer-disable-undo fep:*menu-buffer*))
	  (progn  
	    (set-buffer fep-command-buffer)
	    (setq mc-flag nil)   ;;; for Mule
;;	    (setq kanji-flag nil)
	    ;;; for Nemacs 2.1
;;	    (setq kanji-fileio-code 0)
	    (buffer-disable-undo fep-command-buffer)
	    (erase-buffer)
	    (setq fep-command-tail-position (point-min))
	    (let ((max 1024) (i 0))
	      (while (< i max)
		(insert 0)
		(setq i (1+ i)))))
	  (setq fep-error-code nil)
	  (if (fep:authority)
	      (progn
		(fep:reset)
		(fep:info))	; get the initial system str
	    (setq fep-server-process nil))
	  )))))

(defun fep:reset ()
  "Reset Fep Bridge."
  (fep-command-start fep:FI_Reset)
  (fep-command-end)
  (fep:get-result)
  )
  
(defun fep:info ()
  "Get information of Fep Bridge."
  (fep-command-start fep:FI_Info)
  (fep-command-end)
  (fep:get-info)
  )
  
(defun fep:authority ()
  "Send the user name to Fep Bridge."
  (let ((name (user-login-name)))
    (fep-command-start fep:FI_Auth)
    (fep:put-int2 (length name))
    (insert name)
    (fep-command-end)
    (set-buffer fep-server-buffer);fep:*buffer*)
    (and (eq (fep:get-1byte) fep:FO_Ack) (eq (fep:get-1byte) 0)))
  )

;;;
;;; Get info
;;;
(defun fep:get-info ()
  (let ((init-mode-len) (init-system-len))
    (set-buffer fep-server-buffer);fep:*buffer*)
    (progn
      (fep:get-nbyte 11)
      (setq init-mode-len	(fep:get-int2))
      (setq init-system-len	(fep:get-int2))
      (fep:get-nbyte (* 2 init-mode-len))
      (setq fep:*init-system-str*
	    (code-convert-string (fep:get-nbyte init-system-len)
				 *sjis* *internal*))
      (fep:get-nbyte init-system-len)
      ))
  )

(defun fep-server-active-p ()
  (and fep-server-process
       (eq (process-status fep-server-process) 'open)))

(defun fep-connection-error ()
  (setq fep-error-code ':fep-no-connection)
  (setq fep-return-code -1)
  nil)

(defun fep-server-close ()
  (if (fep-server-active-p)
      (delete-process fep-server-process))
  (if fep-server-buffer
      (kill-buffer fep-server-buffer))
  (if fep-command-buffer
      (kill-buffer fep-command-buffer))
  (if fep:*menu-buffer*
      (kill-buffer fep:*menu-buffer*))
  (setq fep-server-process nil)
  (setq fep-server-buffer nil)
  (setq fep-command-buffer nil)
  (setq fep:*menu-buffer*))

(or (fboundp 'si:kill-emacs)
    (fset 'si:kill-emacs (symbol-function 'kill-emacs)))

(defun kill-emacs (&optional arg)
  (interactive "P")
  (if (fep-server-active-p)
      (progn
	(fep-server-close)))
  (si:kill-emacs arg))

;;;
;;; entry function
;;;
(defun fep-server-henkan-begin (henkan-string)
  (if (not (fep-server-active-p)) (fep-connection-error)
    (let ((inhibit-quit t) (roma-chr-list (kana-to-roma henkan-string)))
      (save-excursion
	(fep:reset)
;	(setq fep-henkan-string henkan-string)

	(fep-command-start fep:FI_ConvN)
	(fep:put-int2 (length roma-chr-list))
	(mapcar '(lambda (chr)
		   (fep:put-1byte chr)
		   (if (> chr 63) (setq chr (+ chr 32)))
		   (fep:put-1byte chr)
		   (fep:put-1byte 0))
		roma-chr-list)
	(fep-command-end)
	(fep:get-result)
;; 	(mapcar '(lambda (chr)
;; 		   (fep-command-start fep:FI_Conv1)
;; 		   (fep:put-1byte chr)
;; 		   (if (> chr 63) (incf chr 32))
;; 		   (fep:put-1byte chr)
;; 		   (fep:put-1byte 0)
;; 		   (fep-command-end)
;; 		   (fep:get-result))
;; 		roma-chr-list)
	(fep-henkan-begin)
;(deb-break-point)
	(setq fep-error-code fep:*return-code*)
	(if (= fep-error-code 0)  (setq fep-error-code nil))
	t
  ))))

;; $B%V%l!<%/%]%$%s%HMQ%@%_!<4X?t(B
;(defun deb-break-point ())
;(debug-on-entry 'deb-break-point)

(defun fep-send-kcode (kcode)
  "Send key input to FEP and show result of conversion. If FEP does not
   accept the key input, execute Emacs command assigned to the key input."
  (save-excursion
    (fep-command-start fep:FI_Conv1)
    (fep:put-1byte (car kcode))
    (fep:put-1byte (car (cdr kcode)))
    (fep:put-1byte (fep:convert-modifiers (car (cdr (cdr kcode)))))
    (fep-command-end)
    (fep:get-result)))

(defun fep-send-kcode-list (kcodes)
  "Send key input to FEP and show result of conversion. If FEP does not
   accept the key input, execute Emacs command assigned to the key input."
  (save-excursion
    (fep-command-start fep:FI_ConvN)
    (fep:put-int2 (length kcodes))
    (mapcar '(lambda (kcode)
	       (fep:put-1byte (car kcode))
	       (fep:put-1byte (car (cdr kcode)))
	       (fep:put-1byte (fep:convert-modifiers (car (cdr (cdr kcode))))))
	    kcodes)
    (fep-command-end)
    (fep:get-result)))

(defun fep:convert-modifiers (mod-list)
  (let ((c 0))
    (while mod-list
       (if (eq 'shift (car mod-list)) (setq c (+ c fep:VK_SHIFT_MASK)))
       (if (eq 'control (car mod-list)) (setq c (+ c fep:VK_CONTROL_MASK)))
       (if (or (eq 'meta (car mod-list)) (eq 'alt (car mod-list)))
	   (setq c (+ c fep:VK_ALT_MASK)))
       (setq mod-list (cdr mod-list)))
    c
))

;;;
;;; entry function
;;;
(defun fep-server-henkan-quit () t)


;;;
;;; entry function
;;;
(defun fep-server-henkan-end (bunsetu-no)
  (if (not (fep-server-active-p)) (fep-connection-error)
    (fep-kakutei))
)

;;;
;;; entry function
;;;
(defun fep-bunsetu-kanji (bunsetu-no &optional buffer)
  (let ((savebuffer (current-buffer)))
    (unwind-protect 
	(progn
	    (set-buffer buffer)
	    (egg:bunsetu-face-off 0)
	    (let ((p0 (point)))
	      (insert fep:*buf-str*)
	      (if (< bunsetu-no 0)
		  nil
		(egg:bunsetu-face-on 0)
		(if (not window-system)
		    (goto-char egg:*region-end*))
		)))
      (set-buffer savebuffer))))

(defun fep:set-bunsetu-pos ()
  (let ((i 0) (len (length fep:*buf-attr*)))
    (setq fep:*bunsetu-start* nil
	  fep:*bunsetu-end* nil)
    (while (and (not fep:*bunsetu-start*) (< i len))
      (while (and (< i len)
		  (zerop (logand (aref fep:*buf-attr* i) fep:*buf-attr-mask*)))
	(setq i (1+ i)))
      (if (< i len)
	  (progn
	    (setq fep:*bunsetu-start* i)
	    (while
		(and (< i len)
		     (not (zerop (logand (aref fep:*buf-attr* i) fep:*buf-attr-mask*)))
		     (zerop (logand (aref fep:*buf-attr* i) fep:*buf-attr-mask2*)))
	      (setq i (1+ i)))
	    (setq fep:*bunsetu-end* i))))
    ))

(defun fep:make-segments (attr-str mask buf-top)
  "Make list of attribute segments from attr-str."
  (let ((rtn nil) (i 0) (s 0) (len (length attr-str)))
    (while (< i len)
      (while (and (< i len)
		  (zerop (logand (aref attr-str i) mask)))
	(setq i (1+ i)))
      (if (< i len)
	  (progn
	    (setq s i)
	    (while
		(and (< i len)
		     (not (zerop (logand (aref attr-str i) mask))))
	      (setq i (1+ i)))
	    (setq rtn
		  (cons (make-overlay (+ buf-top s) (+ buf-top i) nil t) rtn))
	    )))
    rtn
  ))

(defun fep:attribute-on (seg-list attr)
  (let ((segs seg-list))
    (while segs
      (progn
	(overlay-put (car segs) 'face attr)
	(setq segs (cdr segs)))))
  )

(defun fep:attribute-off (seg-list)
  (let ((segs seg-list))
    (while segs
      (progn
	(and (overlayp (car segs))
	     (delete-overlay (car segs)))
	(setq segs (cdr segs))))
    ))


(defun fep:display-minibuffer ()
  "Display the system str if it differs from the initial system str."
  (if (equal fep:*init-system-str* fep:*system-str*)
      (save-excursion
	(if fep:*saved-minibuffer*
	    ;display the saved-buffer in minibuffer
	    (progn
	      (set-window-buffer (minibuffer-window) fep:*saved-minibuffer*)
	      (setq fep:*saved-minibuffer nil)
	      )))
    (save-excursion
      (if (not fep:*saved-minibuffer*)
	  (setq fep:*saved-minibuffer* (window-buffer (minibuffer-window)))
	)
      (setq minibuffer-preprompt nil)
      (set-window-buffer (minibuffer-window) fep:*menu-buffer*)
      (set-buffer fep:*menu-buffer*)
      (fep:attribute-off fep:*system-seg-list*)
      (delete-region (point-min) (point-max))
      (insert fep:*system-str*)
      (setq fep:*system-seg-list*
	    (fep:make-segments fep:*system-attr*
			       fep:*system-attr-mask* (point-min)))
      (fep:attribute-on fep:*system-seg-list* 'region)
      ))
  )

(defconst *fep-error-alist*
  '(
    (1 :FEP_SERVER_DEAD
       "$B%5!<%P$,;`$s$G$$$^$9!#(B")
    (2 :FEP_SOCK_OPEN_FAIL
       "socket$B$N(Bopen$B$K<:GT$7$^$7$?!#(B")
    (6 :FEP_ALLOC_FAIL
       "$B%a%b%j(Balloc$B$G<:GT$7$^$7$?!#(B")
    (7 :SJ3_ILLEGAL_COMMAND
       "$B%3%^%s%I$,4V0c$C$F$$$k(B")
    (12 :FEP_BAD_HOST
	" $B%[%9%HL>$,$J$$(B ")
    (13 :FEP_BAD_USER
	" $B%f!<%6L>$,$J$$(B ")
    (31 :FEP_NOT_A_DICT
	"$B@5$7$$<-=q$G$O$"$j$^$;$s!#(B")
    (35 :FEP_NO_EXIST     
	"$B%U%!%$%k$,B8:_$7$^$;$s!#(B")
    (37 :FEP_OPENF_ERR
	"$B%U%!%$%k$,%*!<%W%s$G$-$^$;$s!#(B")
    (39 :FEP_PARAMR
	"$B%U%!%$%k$NFI$_9~$_8"8B$,$"$j$^$;$s!#(B")
    (40 :FEP_PARAMW
	"$B%U%!%$%k$N=q$-9~$_8"8B$,$"$j$^$;$s!#(B")
    (71 :FEP_NOT_A_USERDICT
	"$B;XDj$5$l$F<-=q$O!"%f!<%6!<<-=q$G$O$"$j$^$;$s!#(B")
    (72 :FEP_RDONLY
	"$B%j!<%I%*%s%j!<$N<-=q$KEPO?$7$h$&$H$7$^$7$?!#(B")
    (74 :FEP_BAD_YOMI
	"$BFI$_$KITE,Ev$JJ8;z$,4^$^$l$F$$$^$9!#(B")
    (75 :FEP_BAD_KANJI
	"$B4A;z$KITE,Ev$JJ8;z$,4^$^$l$F$$$^$9!#(B")
    (76 :FEP_BAD_HINSHI
	"$B;XDj$5$l$?IJ;lHV9f$,$"$j$^$;$s!#(B")
    (82 :FEP_WORD_ALREADY_EXIST
	"$B;XDj$5$l$?C18l$O$9$G$KB8:_$7$F$$$^$9!#(B")
    (84 :FEP_JISHOTABLE_FULL
	"$B<-=q%F!<%V%k$,0lGU$G$9!#(B")
    (92 :FEP_WORD_NO_EXIST
	"$B;XDj$5$l$?C18l$,B8:_$7$^$;$s!#(B")
    (102 :FEP_MKDIR_FAIL
	" $B%G%#%l%/%H%j$r:n$jB;$J$C$?(B ")
    ))

(defun fep-error-symbol (code)
  (let ((pair (assoc code *fep-error-alist*)))
    (if (null pair)
	(list ':fep-unknown-error-code code)
      (car (cdr pair)))))

;;;; $B%m!<%^;z5UJQ49(B
(setq kana-to-roma-table
'((?$B$"(B ?A)   (?$B$$(B ?I)   (?$B$&(B ?U)   (?$B$((B ?E)   (?$B$*(B ?O)
  (?$B$+(B ?K ?A)(?$B$-(B ?K ?I)(?$B$/(B ?K ?U)(?$B$1(B ?K ?E)(?$B$3(B ?K ?O)
  (?$B$5(B ?S ?A)(?$B$7(B ?S ?I)(?$B$9(B ?S ?U)(?$B$;(B ?S ?E)(?$B$=(B ?S ?O)
  (?$B$?(B ?T ?A)(?$B$A(B ?T ?I)(?$B$D(B ?T ?U)(?$B$F(B ?T ?E)(?$B$H(B ?T ?O)
  (?$B$J(B ?N ?A)(?$B$K(B ?N ?I)(?$B$L(B ?N ?U)(?$B$M(B ?N ?E)(?$B$N(B ?N ?O)
  (?$B$O(B ?H ?A)(?$B$R(B ?H ?I)(?$B$U(B ?H ?U)(?$B$X(B ?H ?E)(?$B$[(B ?H ?O)
  (?$B$^(B ?M ?A)(?$B$_(B ?M ?I)(?$B$`(B ?M ?U)(?$B$a(B ?M ?E)(?$B$b(B ?M ?O)
  (?$B$d(B ?Y ?A)           (?$B$f(B ?Y ?U)           (?$B$h(B ?Y ?O)
  (?$B$i(B ?R ?A)(?$B$j(B ?R ?I)(?$B$k(B ?R ?U)(?$B$l(B ?R ?E)(?$B$m(B ?R ?O)
  (?$B$o(B ?W ?A)(?$B$p(B ?W ?I)           (?$B$q(B ?W ?E)(?$B$r(B ?W ?O)
  (?$B$,(B ?G ?A)(?$B$.(B ?G ?I)(?$B$0(B ?G ?U)(?$B$2(B ?G ?E)(?$B$4(B ?G ?O)
  (?$B$6(B ?Z ?A)(?$B$8(B ?Z ?I)(?$B$:(B ?Z ?U)(?$B$<(B ?Z ?E)(?$B$>(B ?Z ?O)
  (?$B$@(B ?D ?A)(?$B$B(B ?D ?I)(?$B$E(B ?D ?U)(?$B$G(B ?D ?E)(?$B$I(B ?D ?O)
  (?$B$P(B ?B ?A)(?$B$S(B ?B ?I)(?$B$V(B ?B ?U)(?$B$Y(B ?B ?E)(?$B$\(B ?B ?O)
  (?$B$Q(B ?P ?A)(?$B$T(B ?P ?I)(?$B$W(B ?P ?U)(?$B$Z(B ?P ?E)(?$B$](B ?P ?O)
  (?$B$!(B ?L ?A)(?$B$#(B ?L ?I)(?$B$%(B ?L ?U)(?$B$'(B ?L ?E)(?$B$)(B ?L ?O)
  (?$B$c(B ?L ?Y ?A)(?$B$e(B ?L ?Y ?U)(?$B$g(B ?L ?Y ?O)(?$B$C(B ?L ?T ?U)(?$B$s(B ?N ?N)
  (?$B!%(B ?.)(?$B!$(B ?,)(?$B!#(B ?.)(?$B!"(B ?,)(?$B!<(B ?-)
  (?$B#0(B ?0)(?$B#1(B ?1)(?$B#2(B ?2)(?$B#3(B ?3)(?$B#4(B ?4)
  (?$B#5(B ?5)(?$B#6(B ?6)(?$B#7(B ?7)(?$B#8(B ?8)(?$B#9(B ?9)
  (?$B!*(B ?!)(?$B!w(B ?@)(?$B!t(B ?#)(?$B!p(B ?$)(?$B!s(B ?%)
  (?$B!0(B ?^)(?$B!u(B ?&)(?$B!v(B ?*)(?$B!J(B ?()(?$B!K(B ?))
  (?$B!2(B ?_)(?$B!a(B ?=)(?$B!\(B ?+)(?$B!C(B ?|)(?$B!o(B ?\\)
  (?$B!P(B ?{)(?$B!Q(B ?})(?$B!N(B ?[)(?$B!O(B ?])
  (?$B!'(B ?:)(?$B!((B ?\;)(?$B!I(B ?\")(?$B!G(B ?')(?$B!)(B ??)
  (?$B!1(B ?~)(?$B!.(B ?`)(?$B!c(B ?<)(?$B!d(B ?>)(?$B!?(B ?/)))

(defun kana-to-roma (kanastr)
  (let (romachrs tmp)
    (mapcar
     '(lambda (chr)
	(if (setq tmp (assq chr kana-to-roma-table))
	    (setq romachrs (append romachrs (cdr tmp)))))
     (string-to-char-list kanastr))
    romachrs
    ))

;; (load-library "fep-oak")
