;;;; $Id: fep-egg.el,v 1.13 1995/12/19 05:22:20 ueda Exp ueda $
;;;;
;; Kana Kanji Conversion Protocol Package for Egg
;; Coded by Y.Ueda (ueda@kyoto-u.ac.jp)

;; This file is part of Egg on Mule (Multilingal Environment)

;; Egg is distributed in the forms of patches to GNU
;; Emacs under the terms of the GNU EMACS GENERAL PUBLIC
;; LICENSE which is distributed along with GNU Emacs by the
;; Free Software Foundation.

;; Egg is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied
;; warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
;; PURPOSE.  See the GNU EMACS GENERAL PUBLIC LICENSE for
;; more details.

;; You should have received a copy of the GNU EMACS GENERAL
;; PUBLIC LICENSE along with Nemacs; see the file COPYING.
;; If not, write to the Free Software Foundation, 675 Mass
;; Ave, Cambridge, MA 02139, USA.


;;;
;;; fep-egg.el 
;;;
;;; $B!V$?$^$4!W$N(B fep $B%P!<%8%g%s(B
;;; $B$+$J4A;zJQ49%5!<%P$K(B fepbrdg $B$r;H$$$^$9!#(B
;;;
;;; fep-egg $B$K4X$9$kDs0F!"Cn>pJs$O(B ueda@kyoto-u.ac.jp $B$K$*Aw$j2<$5$$!#(B
;;;
;;;                                                $B>eED(B $B9/M5(B

(require 'egg)
(provide 'fep-egg)
(require 'fep-client)

;;;----------------------------------------------------------------------
;;;
;;; Version control routine
;;;
;;;----------------------------------------------------------------------

(defvar fep-egg-version "3.00" "Version number of this version of Egg. ")
;;; Last modified date: Thu Aug  4 21:18:11 1994

(and (equal (user-full-name) "Kiyoji Ishii")
     (defun fep-egg-version-update (arg)
       (interactive "P")
       (if (equal (buffer-name (current-buffer)) "fep-egg.el")
	   (save-excursion
	    (goto-char (point-min))
	    (re-search-forward "(defvar fep-egg-version \"[0-9]+\\.")
	    (let ((point (point))
		  (minor))
	      (search-forward "\"")
	      (backward-char 1)
	      (setq minor (string-to-int (buffer-substring point (point))))
	      (delete-region point (point))
	      (if (<= minor 8) (insert "0"))
	      (insert  (int-to-string (1+ minor)))
	      (re-search-forward "Last modified date: ")
	      (kill-line)
	      (insert (current-time-string)))
	    (save-buffer)
	    (if arg (byte-compile-file (buffer-file-name)))
	 )))
     )

;;;----------------------------------------------------------------------
;;;
;;; KKCP package: Kana Kanji Conversion Protocol
;;;
;;; KKCP to FEPSERV interface; 
;;;
;;;----------------------------------------------------------------------

(defvar *KKCP:error-flag* t)

(defun KKCP:error (errorCode &rest form)
  (cond((eq errorCode ':FEP_SOCK_OPEN_FAIL)
	(notify "EGG: %s $B>e$K(B FEPSERV $B$,$"$j$^$;$s!#(B" (or (get-fep-host-name) "local"))
	(if debug-on-error
	    (error "EGG: No FEPSERV on %s is running." (or (get-fep-host-name) "local"))
	  (error  "EGG: %s $B>e$K(B FEPSERV $B$,$"$j$^$;$s!#(B" (or (get-fep-host-name) "local"))))
       ((eq errorCode ':FEP_SERVER_DEAD)
	(notify "EGG: %s $B>e$N(BFEPSERV $B$,;`$s$G$$$^$9!#(B" (or (get-fep-host-name) "local"))
	(if debug-on-error
	    (error "EGG: FEPSERV on %s is dead." (or (get-fep-host-name) "local"))
	  (error  "EGG: %s $B>e$N(B FEPSERV $B$,;`$s$G$$$^$9!#(B" (or (get-fep-host-name) "local"))))
       ((and (consp errorCode)
	     (eq (car errorCode) ':FEP_UNKNOWN_HOST))
	(notify "EGG: $B%[%9%H(B %s $B$,$_$D$+$j$^$;$s!#(B" (car(cdr errorCode)))
	(if debug-on-error
	    (error "EGG: Host %s is unknown." (car(cdr errorCode)))
	  (error "EGG: $B%[%9%H(B %s $B$,$_$D$+$j$^$;$s!#(B" (car(cdr errorCode)))))
       ((and (consp errorCode)
	     (eq (car errorCode) ':FEP_UNKNOWN_SERVICE))
	(notify "EGG: Network service %s $B$,$_$D$+$j$^$;$s!#(B" (car(cdr errorCode)))
	(if debug-on-error
	    (error "EGG: Service %s is unknown." (car(cdr errorCode)))
	  (error "EGG: Network service %s $B$,$_$D$+$j$^$;$s!#(B" (cdr errorCode))))
       (t
	(notify "KKCP: $B860x(B %s $B$G(B %s $B$K<:GT$7$^$7$?!#(B" errorCode form)
	(if debug-on-error
	    (error "KKCP: %s failed because of %s." form errorCode)
	  (error  "KKCP: $B860x(B %s $B$G(B %s $B$K<:GT$7$^$7$?!#(B" errorCode form)))))

(defun KKCP:server-open (hostname loginname)
  (let ((result (fep-server-open hostname loginname)))
    (cond((null fep-error-code) result)
	 (t (KKCP:error fep-error-code 'KKCP:server-open hostname loginname)))))

(defun KKCP:henkan-begin (henkan-string)
  (let ((result (fep-server-henkan-begin henkan-string)))
    (cond((null fep-error-code) result)
	 ((eq fep-error-code ':fep-no-connection)
	  (EGG:open-fep)
	  (KKCP:henkan-begin henkan-string))
	 ((null *KKCP:error-flag*) result)
	 (t (KKCP:error fep-error-code 'KKCP:henkan-begin henkan-string)))))

(defun KKCP:henkan-quit ()
  (let ((result (fep-server-henkan-quit)))
    (cond ((null fep-error-code) result)
	  ((eq fep-error-code ':fep-no-connection)
	   (EGG:open-fep)
	   (KKCP:henkan-quit))
	  ((null *KKCP:error-flag*) result)
	  (t (KKCP:error fep-error-code 'KKCP:henkan-quit)))))

(defun KKCP:henkan-end (&optional bunsetuno)
  (let ((result (fep-server-henkan-end bunsetuno)))
    (cond ((null fep-error-code) result)
	  ((eq fep-error-code ':fep-no-connection)
	   (EGG:open-fep)
	   (KKCP:henkan-end bunsetuno))	  
	  ((null *KKCP:error-flag*) result)
	  (t (KKCP:error fep-error-code 'KKCP:henkan-end)))))

(defun KKCP:make-directory (pathname)
  (let ((result (fep-server-make-directory pathname)))
    (cond ((null fep-error-code) result)
	  ((eq fep-error-code ':fep-no-connection)
	   (EGG:open-fep)
	   (KKCP:make-directory pathname))
	  ((null *KKCP:error-flag*) result)
	  (t (KKCP:error fep-error-code 'kkcp:make-directory pathname)))))

(defun KKCP:file-access (pathname mode)
  (let ((result (fep-server-file-access pathname mode)))
    (cond ((null fep-error-code)
	   (if (= result 0) t nil))
	  ((eq fep-error-code ':fep-no-connection)
	   (EGG:open-fep)
	   (KKCP:file-access pathname mode))
	  ((null *KKCP:error-flag*) result)
	  (t (KKCP:error fep-error-code 'kkcp:file-access pathname mode)))))

(defun KKCP:server-close ()
  (let ((result (fep-server-close)))
    (cond ((null fep-error-code) result)
	  ((null *KKCP:error-flag*) result)
	  (t (KKCP:error fep-error-code 'KKCP:server-close)))))

;;;----------------------------------------------------------------------
;;;
;;; Kana Kanji Henkan 
;;;
;;;----------------------------------------------------------------------

;;;
;;; Entry functions for egg-startup-file
;;;

(defmacro push-end (val loc)
  (list 'push-end-internal val (list 'quote loc)))

(defun push-end-internal (val loc)
  (set loc
       (if (eval loc)
	   (nconc (eval loc) (cons val nil))
	 (cons val nil))))


;;;
;;; FEP interface
;;;

(defun get-fep-host-name ()
  (cond((and (boundp 'fep-host-name) (stringp fep-host-name))
	fep-host-name)
       ((and (boundp 'fepserv-host-name) (stringp fepserv-host-name))
	fepserv-host-name)
       (t(getenv "FEPSERV"))))				; 92.7.7 by Y.Kawabe

(fset 'get-fepserv-host-name (symbol-function 'get-fep-host-name))

(defun set-fep-host-name (name)
  (interactive "sHost name: ")
  (let ((*KKCP:error-flag* nil))
    (disconnect-fep))
  (setq fep-host-name name)
  )

(defvar egg-default-startup-file "eggrc-fep"
  "*Egg startup file name (system default)")

(defvar egg-startup-file ".eggrc-fep"
  "*Egg startup file name.")

(defvar egg-startup-file-search-path (append '("~" ".") load-path)
  "*List of directories to search for start up file to load.")

(defun egg:search-file (filename searchpath)
  (let ((result nil))
    (if (null (file-name-directory filename))
	(let ((path searchpath))
	  (while (and path (null result ))
	    (let ((file (substitute-in-file-name
			 (expand-file-name filename (if (stringp (car path)) (car path) nil)))))
	      (if (file-exists-p file) (setq result file)
		(setq path (cdr path))))))
      (let((file (substitute-in-file-name (expand-file-name filename))))
	(if (file-exists-p file) (setq result file))))
    result))

(defun EGG:open-fep ()
  (KKCP:server-open (or (get-fep-host-name) (system-name))
  		    (user-login-name))
  (notify "$B%[%9%H(B %s $B$N(B FEP $B$r5/F0$7$^$7$?!#(B" (or (get-fep-host-name) "local"))
  (let ((eggrc (or (egg:search-file egg-startup-file egg-startup-file-search-path)
		   (egg:search-file egg-default-startup-file load-path))))
    (if eggrc (load-file eggrc)
      (progn
	(KKCP:server-close)
	(error "eggrc-search-path $B>e$K(B egg-startup-file $B$,$"$j$^$;$s!#(B")))))

(defun disconnect-fep ()
  (interactive)
  (KKCP:server-close))

(defun close-fep ()
  (interactive)
  (KKCP:server-close))

;;;
;;; Kanji henkan
;;;

(defvar egg:*kanji-kanabuff* nil)

(defvar *bunsetu-number* nil)

(defun kanji-moji-suu (str)
  (let ((max (length str)) (count 0) (i 0))
    (while (< i max)
      (setq count (1+ count))
      (if (< (aref str i) 128) (setq i (1+ i)) (setq i (+ i 3))))
    count))

(defun bunsetu-kanji (number)
  (fep-bunsetu-kanji number))

(defun bunsetu-kanji-insert (bunsetu-no)
  (fep-bunsetu-kanji bunsetu-no (current-buffer)))

;;;;
;;;; User entry : henkan-region, henkan-paragraph, henkan-sentence
;;;;

;(defconst egg:*bunsetu-face* nil "*$BJ8@aI=<($KMQ$$$k(B face $B$^$?$O(B nil")
(defconst egg:*bunsetu-face* 'region "*$BJ8@aI=<($KMQ$$$k(B face $B$^$?$O(B nil")
(make-variable-buffer-local
 (defvar egg:*bunsetu-overlay* nil "$BJ8@a$NI=<($K;H$&(B overlay"))

(defconst egg:*bunsetu-kugiri* " " "*$BJ8@a$N6h@Z$j$r<($9J8;zNs(B")


(defconst egg:*henkan-face* nil "*$BJQ49NN0h$rI=<($9$k(B face $B$^$?$O(B nil")
(make-variable-buffer-local
 (defvar egg:*henkan-overlay* nil "$BJQ49NN0h$NI=<($K;H$&(B overlay"))

(defconst egg:*henkan-open*  "|" "*$BJQ49$N;OE@$r<($9J8;zNs(B")
(defconst egg:*henkan-close* "|" "*$BJQ49$N=*E@$r<($9J8;zNs(B")

(defun egg:henkan-face-on ()
  (if (overlayp egg:*henkan-overlay*) nil
    (setq egg:*henkan-overlay* (make-overlay 1 1 nil))
    (overlay-put egg:*henkan-overlay* 'face egg:*henkan-face*) )
  (move-overlay egg:*henkan-overlay* egg:*region-start* egg:*region-end* (current-buffer)) )

(defun egg:henkan-face-off ()
  (and (overlayp egg:*henkan-overlay*)
       (delete-overlay egg:*henkan-overlay*) ))

(defun henkan-region (start end)
  (interactive "r")
  (if (interactive-p) (set-mark (point))) ;;; to be fixed
  (henkan-region-internal start end))

(defvar henkan-mode-indicator "$B4A(B")

(defun henkan-region-internal (start end)
  "region$B$r$+$J4A;zJQ49$9$k!#(B"
  (setq egg:*kanji-kanabuff* (buffer-substring start end))
  (if overwrite-mode
      (setq egg:*overwrite-mode-deleted-chars* 
	    (if egg:*henkan-fence-mode* 0
	      (length egg:*kanji-kanabuff*))))
  (setq *bunsetu-number* nil)
  (let ((result (KKCP:henkan-begin egg:*kanji-kanabuff*)))
    (if  result
	(progn
	  (mode-line-egg-mode-update henkan-mode-indicator)
	  (goto-char start)
	  (or (markerp egg:*region-start*)
	      (setq egg:*region-start* (make-marker)))
	  (or (markerp egg:*region-end*)
	      (setq egg:*region-end* (set-marker-type (make-marker) t)))
	  (if (null (marker-position egg:*region-start*))
	      (progn
		;;;(setq egg:*global-map-backup* (current-global-map))
		(setq egg:*local-map-backup* (current-local-map))
		(and (boundp 'disable-undo) (setq disable-undo t))
		(goto-char start)
		(delete-region start end)
		(insert egg:*henkan-open*)
		(set-marker egg:*region-start* (point))
		(insert egg:*henkan-close*)
		(set-marker egg:*region-end* egg:*region-start*)
		(egg:henkan-face-on)
		(goto-char egg:*region-start*)
		)
	    (progn
	      (egg:fence-face-off)
	      (delete-region (- egg:*region-start* (length egg:*fence-open*)) 
			     egg:*region-start*)
	      (delete-region egg:*region-end* (+ egg:*region-end* (length egg:*fence-close*)))
	      (goto-char egg:*region-start*)
	      (insert egg:*henkan-open*)
	      (set-marker egg:*region-start* (point))
	      (goto-char egg:*region-end*)
	      (let ((point (point)))
		(insert egg:*henkan-close*)
		(set-marker egg:*region-end* point))
	      (goto-char start)
	      (delete-region start end)
	      (egg:henkan-face-on))
	    )
	  (henkan-insert-kouho 0)
	  ;;;(use-global-map henkan-mode-map)
	  ;;;(use-local-map nil)
	  (use-local-map henkan-mode-map)
	  )))
  )

(defun henkan-paragraph ()
  "Kana-kanji henkan  paragraph at or after point."
  (interactive )
  (save-excursion
    (forward-paragraph)
    (let ((end (point)))
      (backward-paragraph)
      (henkan-region-internal (point) end ))))

(defun henkan-sentence ()
  "Kana-kanji henkan sentence at or after point."
  (interactive )
  (save-excursion
    (forward-sentence)
    (let ((end (point)))
      (backward-sentence)
      (henkan-region-internal (point) end ))))

(defun henkan-word ()
  "Kana-kanji henkan word at or after point."
  (interactive)
  (save-excursion
    (re-search-backward "\\b\\w" nil t)
    (let ((start (point)))
      (re-search-forward "\\w\\b" nil t)
      (henkan-region-internal start (point)))))

;;;
;;; Kana Kanji Henkan Henshuu mode
;;;

(defun set-egg-henkan-mode-format (open close kugiri &optional henkan-face bunsetu-face)
   "$BJQ49(B mode $B$NI=<(J}K!$r@_Dj$9$k!#(BOPEN $B$OJQ49$N;OE@$r<($9J8;zNs$^$?$O(B nil$B!#(B
CLOSE$B$OJQ49$N=*E@$r<($9J8;zNs$^$?$O(B nil$B!#(B
KUGIRI$B$OJ8@a$N6h@Z$j$rI=<($9$kJ8;zNs$^$?$O(B nil$B!#(B
HENKAN-FACE $B$,;XDj$5$l$F(B nil $B$G$J$1$l$P!"JQ496h4V$rI=<($9$k(B face $B$H$7$F;H$o$l$k!#(B
BUNSETU-FACE $B$,;XDj$5$l$F(B nil $B$G$J$1$l$P!"CmL\$7$F$$$kJ8@a$rI=<($9$k(B face $B$H$7$F;H$o$l$k(B"

  (interactive (list (read-string "$BJQ493+;OJ8;zNs(B: ")
		     (read-string "$BJQ49=*N;J8;zNs(B: ")
		     (read-string "$BJ8@a6h@Z$jJ8;zNs(B: ")
		     (cdr (assoc (completing-read "$BJQ496h4VI=<(B0@-(B: " egg:*face-alist*)
				 egg:*face-alist*))
		     (cdr (assoc (completing-read "$BJ8@a6h4VI=<(B0@-(B: " egg:*face-alist*)
				 egg:*face-alist*))
		     ))

  (if (and (or (stringp open)  (null open))
	   (or (stringp close) (null close))
	   (or (stringp kugiri) (null kugiri))
	   (or (null henkan-face) (memq henkan-face (face-list)))
	   (or (null bunsetu-face) (memq henkan-face (face-list))))
      (progn
	(setq egg:*henkan-open* (or open "")
	      egg:*henkan-close* (or close "")
	      egg:*bunsetu-kugiri* (or kugiri "")
	      egg:*henkan-face* henkan-face
	      egg:*bunsetu-face* bunsetu-face)
	(if (overlayp egg:*henkan-overlay*)
	    (overlay-put egg:*henkan-overlay* 'face egg:*henkan-face*))
	(if (overlayp egg:*bunsetu-overlay*)
	    (overlay-put egg:*bunsetu-overlay* 'face egg:*bunsetu-face*))

	t)
    (error "Wrong type of arguments: %1 %2 %3 %4 %5" open close kugiri henkan-face bunsetu-face)))

(defun henkan-insert-kouho (bunsetu-no)
  (bunsetu-kanji-insert bunsetu-no))

(defun henkan-kakutei ()
  (interactive)
  (egg:henkan-face-off)
  (delete-region (- egg:*region-start* (length egg:*henkan-open*))
		 egg:*region-start*)
  (delete-region egg:*region-start* egg:*region-end*)
  (delete-region egg:*region-end* (+ egg:*region-end* (length egg:*henkan-close*)))
  (goto-char egg:*region-start*)
  (bunsetu-kanji-insert -1)
  (undo-boundary)
  (KKCP:henkan-end)
  (egg:quit-egg-mode)
  )

(defun henkan-kakutei-before-point ()
  (interactive)
  (let ((p0 fep:*bunsetu-start*))
    (if (= p0 0)  nil
      (egg:bunsetu-face-off *bunsetu-number*)
      (egg:henkan-face-off)
      (delete-region egg:*region-start* egg:*region-end*)
      (goto-char (- egg:*region-start* (length egg:*fence-open*)))
      (delete-char (+ (length egg:*fence-open*) (length egg:*fence-close*)))
      (insert (substring fep:*buf-str* 0 p0))
      (undo-boundary)
      (insert egg:*fence-open*)
      (set-marker egg:*region-start* (point))
      (let ((point (point)))
	(insert egg:*fence-close*)
	(set-marker egg:*region-end* point))
      (goto-char egg:*region-start*)
      (fep-bunsetu-kakutei)
      (bunsetu-kanji-insert 0))))

(defun egg:set-bunsetu-face (no face switch)
  (if (not switch)
      (egg:bunsetu-face-off no) ;; JIC
    (if (overlayp egg:*bunsetu-overlay*) nil
      (setq egg:*bunsetu-overlay* (make-overlay 1 1 nil))
      (overlay-put egg:*bunsetu-overlay* 'face egg:*bunsetu-face*))
    (let ((p0 (+ fep:*bunsetu-start* egg:*region-start*))
	  (p1 (+ fep:*bunsetu-end* egg:*region-start*)))
      (if (and p0 p1)
	  (progn
	    (if window-system
		(goto-char p1)
	      (goto-char egg:*region-end*))
	    (move-overlay egg:*bunsetu-overlay*
			  (if (eq face 'modeline)
			      (+ p0 (1+ (char-boundary-p p0)))
			    p0)
			  p1
			  (current-buffer)))))
    ))

(defun egg:bunsetu-face-on (no)
  (egg:set-bunsetu-face no egg:*bunsetu-face* t))

(defun egg:bunsetu-face-off (no)
  ;; ``no'' will be ignored
  (and (overlayp egg:*bunsetu-overlay*)
       (delete-overlay egg:*bunsetu-overlay*)) )

(defun henkan-forward-bunsetu ()
  (interactive)
  (egg:bunsetu-face-off 0)
  (if (eq (- egg:*region-end* egg:*region-start*) fep:*bunsetu-end*)
      (fep-first-bunsetu)
    (fep-forward-bunsetu))
  (egg:bunsetu-face-on 0))

(defun henkan-backward-bunsetu ()
  (interactive)
  (egg:bunsetu-face-off 0)
  (if (eq 0 fep:*bunsetu-start*)
      (fep-last-bunsetu)
    (fep-backward-bunsetu))
  (egg:bunsetu-face-on 0))

(defun henkan-first-bunsetu ()
  (interactive)
  (egg:bunsetu-face-off 0)
  (fep-first-bunsetu)
  (egg:bunsetu-face-on 0))

(defun henkan-last-bunsetu ()
  (interactive)
  (egg:bunsetu-face-off 0)
  (fep-last-bunsetu)
  (egg:bunsetu-face-on 0))
 
(defun check-number-range (i min max)
  (cond((< i min) max)
       ((< max i) min)
       (t i)))

(defun henkan-hiragana ()
  (interactive)
  (delete-region egg:*region-start* egg:*region-end*)
  (fep-bunsetu-hiragana)
  (bunsetu-kanji-insert 0))

(defun henkan-katakana ()
  (interactive)
  (delete-region egg:*region-start* egg:*region-end*)
  (fep-bunsetu-katakana)
  (bunsetu-kanji-insert 0))

(defun henkan-next-kouho ()
  (interactive)
  (delete-region egg:*region-start* egg:*region-end*)
  (fep-next-kouho)
  (bunsetu-kanji-insert 0))

(defun henkan-previous-kouho ()
  (interactive)
  (delete-region egg:*region-start* egg:*region-end*)
  (fep-previous-kouho)
  (bunsetu-kanji-insert 0))

(defun henkan-bunsetu-chijime ()
  (interactive)
  (delete-region egg:*region-start* egg:*region-end*)
  (fep-bunsetu-chijime)
  (bunsetu-kanji-insert 0))

(defun henkan-bunsetu-nobasi ()
  (interactive)
  (delete-region egg:*region-start* egg:*region-end*)
  (fep-bunsetu-nobasi)
  (bunsetu-kanji-insert 0))

(defun henkan-saishou-bunsetu ()
  (interactive))
;  (bunsetu-length-henko 1))

(defun henkan-saichou-bunsetu ()
  (interactive))
;;   (let ((max (bunsetu-su)) (i *bunsetu-number*)
;; 	(l 0))
;;     (while (< i max)
;;       (setq l (+ l (bunsetu-length i)))
;;       (setq i (1+ i)))
;;     (bunsetu-length-henko l)))

(defun henkan-quit ()
  (interactive)
  (egg:bunsetu-face-off *bunsetu-number*)
  (egg:henkan-face-off)
  (delete-region (- egg:*region-start* (length egg:*henkan-open*))
		 egg:*region-start*)
  (delete-region egg:*region-start* egg:*region-end*)
  (delete-region egg:*region-end* (+ egg:*region-end* (length egg:*henkan-close*)))
  (goto-char egg:*region-start*)
  (insert egg:*fence-open*)
  (set-marker egg:*region-start* (point))
  (insert egg:*kanji-kanabuff*)
  (let ((point (point)))
    (insert egg:*fence-close*)
    (set-marker egg:*region-end* point)
    )
  (goto-char egg:*region-end*)
  (egg:fence-face-on)
  (KKCP:henkan-quit)
  (setq egg:*mode-on* t)
  ;;;(use-global-map fence-mode-map)
  ;;;(use-local-map  nil)
  (use-local-map fence-mode-map)
  (egg:mode-line-display)
  )

(defun henkan-select-kouho ()
  (interactive)
  (if (not (eq (selected-window) (minibuffer-window)))
      (let (changed)
	(save-excursion
	  (let (finished
		(echo-keystrokes 0)
		(inhibit-quit t))
	    (fep-kouho-list)
	    (fep:display-minibuffer)
	    (while (not finished)
	      (let ((ch (read-event)))
		(cond 
		 ((or (eq ch 'down) (eq ch ?\C-n))
		  (fep-kouho-list-down)
		  (fep:display-minibuffer))
		 ((or (eq ch 'up) (eq ch ?\C-p))
		  (fep-kouho-list-up)
		  (fep:display-minibuffer))
		 ((or (eq ch 'left) (eq ch ?\C-b))
		  (fep-kouho-list-left)
		  (fep:display-minibuffer))
		 ((or (eq ch 'right) (eq ch ?\C-f))
		  (fep-kouho-list-right)
		  (fep:display-minibuffer))
		 ((or (eq ch 'return) (eq ch ?\C-m))
		  (setq finished (fep-kouho-list-select)
			changed finished)
		  (if (not finished)
		      (beep)))
		 ((and (numberp ch) (>= ch ?0) (<= ch ?9))
		  (setq finished (fep-kouho-list-nth-kouho ch)
			changed finished)
		  (if (not finished)
		      (beep)))
		 ((or (eq ch ?\C-g) (eq ch ?\C-q))
;		      (eq ch 'escape) (eq ch ?\C-\[))
		  (fep-kouho-list-quit)
		  (setq finished t))
		 (t
		  (beep))
		 )))
	    (if fep:*saved-minibuffer*
					;display the saved-buffer in minibuffer
		(progn
		  (set-window-buffer (minibuffer-window) fep:*saved-minibuffer*)
		  (setq fep:*saved-minibuffer nil)
		  ))))
	(if changed
	    (progn
	      (delete-region egg:*region-start* egg:*region-end*)
	      (bunsetu-kanji-insert 0))
	  (beep))
	)))

(defun henkan-kakutei-and-self-insert ()
  (interactive)
  (setq unread-command-events (list last-command-char))
  (henkan-kakutei))


(defvar henkan-mode-map (make-keymap))

(defvar henkan-mode-esc-map (make-keymap))

(let ((ch 0))
  (while (<= ch 127)
    (define-key henkan-mode-map (char-to-string ch) 'undefined)
    (define-key henkan-mode-esc-map (char-to-string ch) 'undefined)
    (setq ch (1+ ch))))

(let ((ch 32))
  (while (< ch 127)
    (define-key henkan-mode-map (char-to-string ch)
      'henkan-kakutei-and-self-insert)
    (define-key henkan-mode-map (make-string 1 ch) 'henkan-kakutei-and-self-insert)
    (setq ch (1+ ch))))
	
(define-key henkan-mode-map "\e"    henkan-mode-esc-map)
(define-key henkan-mode-map "\ei"  'undefined) ;; henkan-inspect-bunsetu
					       ;; not support for fep
(define-key henkan-mode-map "\es"  'henkan-select-kouho)
(define-key henkan-mode-map "\eh"  'henkan-hiragana)
(define-key henkan-mode-map "\ek"  'henkan-katakana)
(define-key henkan-mode-map "\e<"  'henkan-saishou-bunsetu)
(define-key henkan-mode-map "\e>"  'henkan-saichou-bunsetu)
(define-key henkan-mode-map " "    'henkan-next-kouho)
(define-key henkan-mode-map "\C-@" 'henkan-next-kouho)
(define-key henkan-mode-map "\C-a" 'henkan-first-bunsetu)
(define-key henkan-mode-map "\C-b" 'henkan-backward-bunsetu)
(define-key henkan-mode-map "\C-c" 'henkan-quit)
(define-key henkan-mode-map "\C-d" 'undefined)
(define-key henkan-mode-map "\C-e" 'henkan-last-bunsetu)
(define-key henkan-mode-map "\C-f" 'henkan-forward-bunsetu)
(define-key henkan-mode-map "\C-g" 'henkan-quit)
(define-key henkan-mode-map "\C-h" 'help-command)
(define-key henkan-mode-map "\C-i" 'henkan-bunsetu-chijime)
(define-key henkan-mode-map "\C-j" 'undefined)
(define-key henkan-mode-map "\C-k" 'henkan-kakutei-before-point)
(define-key henkan-mode-map "\C-l" 'henkan-kakutei)
(define-key henkan-mode-map "\C-m" 'henkan-kakutei)
(define-key henkan-mode-map "\C-n" 'henkan-next-kouho)
(define-key henkan-mode-map "\C-o" 'henkan-bunsetu-nobasi)
(define-key henkan-mode-map "\C-p" 'henkan-previous-kouho)
(define-key henkan-mode-map "\C-q" 'undefined)
(define-key henkan-mode-map "\C-r" 'undefined)
(define-key henkan-mode-map "\C-s" 'undefined)
(define-key henkan-mode-map "\C-t" 'undefined)
(define-key henkan-mode-map "\C-u" 'undefined)
(define-key henkan-mode-map "\C-v" 'undefined)
(define-key henkan-mode-map "\C-w" 'undefined)
(define-key henkan-mode-map "\C-x" 'undefined)
(define-key henkan-mode-map "\C-y" 'undefined)
(define-key henkan-mode-map "\C-z" 'undefined)
(define-key henkan-mode-map "\177" 'henkan-quit)

(defun henkan-help-command ()
  "Display documentation fo henkan-mode."
  (interactive)
  (with-output-to-temp-buffer "*Help*"
    (princ (substitute-command-keys henkan-mode-document-string))
    (print-help-return-message)))

(defvar henkan-mode-document-string "$B4A;zJQ49%b!<%I(B:
$BJ8@a0\F0(B
  \\[henkan-first-bunsetu]\t$B@hF,J8@a(B\t\\[henkan-last-bunsetu]\t$B8eHxJ8@a(B  
  \\[henkan-backward-bunsetu]\t$BD>A0J8@a(B\t\\[henkan-forward-bunsetu]\t$BD>8eJ8@a(B
$BJQ49JQ99(B
  $B<!8uJd(B    \\[henkan-previous-kouho]  \t$BA08uJd(B    \\[henkan-next-kouho]
  $BJ8@a?-$7(B  \\[henkan-bunsetu-nobasi]  \t$BJ8@a=L$a(B  \\[henkan-bunsetu-chijime]
  $BJQ498uJdA*Br(B  \\[henkan-select-kouho]
$BJQ493NDj(B
  $BA4J8@a3NDj(B  \\[henkan-kakutei]  \t$BD>A0J8@a$^$G3NDj(B  \\[henkan-kakutei-before-point]
$BJQ49Cf;_(B    \\[henkan-quit]
")
  
;;;----------------------------------------------------------------------
;;;
;;; Dictionary management Facility
;;;
;;;----------------------------------------------------------------------

;;;
;;; $B<-=qEPO?(B 
;;;

;;;;
;;;; User entry: toroku-region
;;;;

(defun remove-regexp-in-string (regexp string)
  (cond((not(string-match regexp string))
	string)
       (t(let ((str nil)
	     (ostart 0)
	     (oend   (match-beginning 0))
	     (nstart (match-end 0)))
	 (setq str (concat str (substring string ostart oend)))
	 (while (string-match regexp string nstart)
	   (setq ostart nstart)
	   (setq oend   (match-beginning 0))
	   (setq nstart (match-end 0))
	   (setq str (concat str (substring string ostart oend))))
	 str))))

(defun toroku-region (start end)
  (interactive "r"))
;;   (let*((kanji
;; 	 (remove-regexp-in-string "[\0-\37]" (buffer-substring start end)))
;; 	(yomi (read-hiragana-string
;; 	       (format "$B<-=qEPO?!X(B%s$B!Y(B  $BFI$_(B :" kanji)))
;; 	(type (menu:select-from-menu *fep-bunpo-menu*))
;; 	(dict-no 
;; 	 (menu:select-from-menu (list 'menu "$BEPO?<-=qL>(B:" egg:*dict-menu*))))
;;     ;;;(if (string-match "[\0-\177]" kanji)
;;     ;;;	(error "Kanji string contains hankaku character. %s" kanji))
;;     ;;;(if (string-match "[\0-\177]" yomi)
;;     ;;;	(error "Yomi string contains hankaku character. %s" yomi))
;;     (KKCP:dict-add dict-no kanji yomi type)
;;     (let ((hinshi (nth 1 (assq type *fep-bunpo-code*)))
;; 	  (gobi   (nth 2 (assq type *fep-bunpo-code*)))
;; 	  (dict-name (cdr (assq dict-no egg:*usr-dict*))))
;;       (notify "$B<-=q9`L\!X(B%s$B!Y(B(%s: %s)$B$r(B%s$B$KEPO?$7$^$7$?!#(B"
;; 	      (if gobi (concat kanji " " gobi) kanji)
;; 	      (if gobi (concat yomi  " " gobi) yomi)
;; 	      hinshi dict-name))))



;;; (lsh 1 18)
(defvar *fep-bunpo-menu*
  '(menu "$BIJ;l(B:"
   (("$BL>;l(B"      .
     (menu "$BIJ;l(B:$BL>;l(B:"
	   (("$BL>;l(B"		. 1)
	    ("$BL>;l(B($B$*!D(B)"	. 2)
	    ("$BL>;l(B($B$4!D(B)"	. 3)
	    ("$BL>;l(B($B!DE*(B/$B2=(B)"	. 4)
	    ("$BL>;l(B($B$*!D$9$k(B)"	. 5)
	    ("$BL>;l(B($B!D$9$k(B)"	. 6)
	    ("$BL>;l(B($B$4!D$9$k(B)"	. 7)
	    ("$BL>;l(B($B!D$J(B/$B$K(B)"	. 8)
	    ("$BL>;l(B($B$*!D$J(B/$B$K(B)"	. 9)
	    ("$BL>;l(B($B$4!D$J(B/$B$K(B)"	. 10)
	    ("$BL>;l(B($BI{;l(B)"	. 11))))
    ("$BBeL>;l(B"    . 12)
    ("$BID;z(B"      . 21)
    ("$BL>A0(B"      . 22)
    ("$BCOL>(B"      . 24)
    ("$B8)(B/$B6hL>(B"   . 25)
    ("$BF0;l(B"      .
	  (menu "$BIJ;l(B:$BF0;l(B:"
		(("$B%5JQ8l44(B"      . 80)
		 ("$B%6JQ8l44(B"      . 81)
		 ("$B0lCJITJQ2=It(B"  . 90)
		 ("$B%+9T8^CJ8l44(B"  . 91)
		 ("$B%,9T8^CJ8l44(B"  . 92)   
		 ("$B%59T8^CJ8l44(B"  . 93)   
		 ("$B%?9T8^CJ8l44(B"  . 94)   
		 ("$B%J9T8^CJ8l44(B"  . 95)   
		 ("$B%P9T8^CJ8l44(B"  . 96)   
		 ("$B%^9T8^CJ8l44(B"  . 97)   
		 ("$B%i9T8^CJ8l44(B"  . 98)   
		 ("$B%o9T8^CJ8l44(B"  . 99))))   
    ("$BO"BN;l(B"         . 26)
    ("$B@\B3;l(B"         . 27)
    ("$B=u?t;l(B"         . 29)
    ("$B?t;l(B"           . 30)
    ("$B@\F,8l(B"         . 31)
    ("$B@\Hx8l(B"         . 36)
    ("$BI{;l(B"           . 45)
    ("$BI{;l(B2"          . 46)
    ("$B7AMF;l8l44(B"     . 60)
    ("$B7AMFF0;l8l44(B"   . 71)
    ("$BC14A;z(B"         . 189))))

(defvar *fep-bunpo-code*
  '(
    ( 1   "$BL>;l(B" )
    ( 2   "$BL>;l(B($B$*!D(B)" )
    ( 3   "$BL>;l(B($B$4!D(B)" )
    ( 4   "$BL>;l(B($B!DE*(B/$B2=(B)" "$BE*(B" nil)
    ( 5   "$BL>;l(B($B$*!D$9$k(B)" "$B$9$k(B" nil)
    ( 6   "$BL>;l(B($B!D$9$k(B)" "$B$9$k(B" nil)
    ( 7   "$BL>;l(B($B$4!D$9$k(B)" "$B$9$k(B" nil)
    ( 8   "$BL>;l(B($B!D$J(B/$B$K(B)" "$B$J(B/$B$K(B" nil)
    ( 9   "$BL>;l(B($B$*!D$J(B/$B$K(B)" "$B$J(B/$B$K(B" nil)
    ( 10  "$BL>;l(B($B$4!D$J(B/$B$K(B)" "$B$J(B/$B$K(B" nil)
    ( 11  "$BL>;l(B($BI{;l(B)" )
    ( 12  "$BBeL>;l(B" )
    ( 21  "$BID;z(B" )
    ( 22  "$BL>A0(B" )
    ( 24  "$BCOL>(B" )
    ( 25  "$B8)(B/$B6hL>(B" )
    ( 26  "$BO"BN;l(B" )
    ( 27  "$B@\B3;l(B" )
    ( 29  "$B=u?t;l(B" )
    ( 30  "$B?t;l(B"   )
    ( 31  "$B@\F,8l(B" )
    ( 36  "$B@\Hx8l(B" )
    ( 45  "$BI{;l(B" )
    ( 46  "$BI{;l(B2" )
    ( 60  "$B7AMF;l8l44(B"           "$B$$(B" ("" "" "" "" ""))
    ( 71  "$B7AMFF0;l8l44(B"         "$B$K(B" ("" "" "" "" "") )
    ( 80  "$B%5JQ8l44(B"             "$B$9$k(B" ("" "" "" "" ""))
    ( 81  "$B%6JQ8l44(B"             "$B$:$k(B" ("" "" "" "" ""))
    ( 90  "$B0lCJITJQ2=It(B"         "$B$k(B" ("" "" "" "" ""))
    ( 91  "$B%+9T8^CJ8l44(B"         "$B$/(B" ("$B$+$J$$(B" "$B$-$^$9(B" "$B$/(B" "$B$/$H$-(B" "$B$1(B"))
    ( 92  "$B%,9T8^CJ8l44(B"         "$B$0(B" ("$B$,$J$$(B" "$B$.$^$9(B" "" "" ""))
    ( 93  "$B%59T8^CJ8l44(B"         "$B$9(B" ("" "" "" "" ""))
    ( 94  "$B%?9T8^CJ8l44(B"         "$B$D(B" ("" "" "" "" ""))
    ( 95  "$B%J9T8^CJ8l44(B"         "$B$L(B" ("" "" "" "" ""))   
    ( 96  "$B%P9T8^CJ8l44(B"         "$B$V(B" ("" "" "" "" ""))   
    ( 97  "$B%^9T8^CJ8l44(B"         "$B$`(B" ("" "" "" "" ""))   
    ( 98  "$B%i9T8^CJ8l44(B"         "$B$k(B" ("" "" "" "" ""))   
    ( 99  "$B%o9T8^CJ8l44(B"         "$B$&(B" ("" "" "" "" ""))   
    ( 189  "$BC14A;z(B"  )
    ( 190  "$BITDj(B"  )
    ( 1000  "$B$=$NB>(B"  )
    ))

;;;
;;; $B<-=qJT=87O(B DicEd
;;;

;;;;;
;;;;; User entry : edit-dict
;;;;;

(defun edit-dict ()
  (interactive))

;;; End of fep-egg.el
;; 92.7.7 by Y.Kawabe -- commented out
;; (if (boundp 'FEP) 
;;    (load-library "fepfns"))
