/*
  fepkey.h : I/F of Fep Bridge 1.33
    Copyright (C) Yutaka Tamiya (tamy@flab.fujitsu.co) 1996/11/13

    Definition of Key Codes
	Values are the same as MS-Windows
*/
#ifndef fepkey_H
#define fepkey_H

/* bit mask for shift status */
#define VK_SHIFT_MASK	0x01
#define VK_CONTROL_MASK	0x02
#define VK_ALT_MASK	0x04

/* "VK_US_*" for 101 US Keyboard */
#define VK_LBUTTON	0x01	/* Left mouse button */
#define VK_RBUTTON	0x02	/* Right mouse button */
#define VK_CANCEL	0x03	/* Used for control-break processing */
#define VK_MBUTTON	0x04	/* Middle mouse button (three-button mouse) */
/* 	--		0x05-0x07  Undefined */
#define VK_BACK		0x08	/* BACKSPACE key */
#define VK_TAB		0x09	/* TAB key */
/* 	--		0x0A-0x0B  Undefined */
#define VK_CLEAR	0x0C	/* CLEAR key */
#define VK_RETURN	0x0D	/* ENTER key */
/*	--		0x0E-0x0F  Undefined */
#define VK_SHIFT	0x10	/* SHIFT key */
#define VK_CONTROL	0x11	/* CTRL key */
#define VK_MENU		0x12	/* ALT key */
#define VK_PAUSE	0x13	/* PAUSE key */
#define VK_CAPITAL	0x14	/* CAPS LOCK key */
#define VK_KANA		0x15	/* Kakana/Hiragana key */
/*	--		0x16-0x17  Reserved for Kanji systems */
#define VK_FINAL	0x18	/* for Korea (not supported yet) */
#define VK_KANJI	0x19	/* Kanji/Hankaku/Zenkaku key */
/*	--		0x1A	   Undefined */
#define VK_ESCAPE	0x1B	/* ESC key */
/*	--		0x1C-0x1F  Reserved for Kanji systems */
#define VK_CONVERT	0x1C	/* Henkan key */
#define VK_NONCONVERT	0x1D	/* MuHenkan key */
#define VK_ACCEPT	0x1E	/* ?? not supported yet */
#define VK_MODECHANGE	0x1F	/* ?? not supported yet */
#define VK_SPACE	0x20	/* SPACEBAR */
#define VK_PRIOR	0x21	/* PAGE UP key */
#define VK_NEXT		0x22	/* PAGE DOWN key */
#define VK_END		0x23	/* END key */
#define VK_HOME		0x24	/* HOME key */
#define VK_LEFT		0x25	/* LEFT ARROW key */
#define VK_UP		0x26	/* UP ARROW key */
#define VK_RIGHT	0x27	/* RIGHT ARROW key */
#define VK_DOWN		0x28	/* DOWN ARROW key */
#define VK_SELECT	0x29	/* SELECT key */
/*	--		0x2A	   OEM specific */
#define VK_EXECUTE	0x2B	/* EXECUTE key */
#define VK_SNAPSHOT	0x2C	/* PRINT SCREEN key for Windows 3.0 and later */
#define VK_INSERT	0x2D	/* INS key */
#define VK_DELETE	0x2E	/* DEL key */
#define VK_HELP		0x2F	/* HELP key */
#define VK_0		0x30	/* 0 key */
#define VK_1		0x31	/* 1 key */
#define VK_2		0x32	/* 2 key */
#define VK_3		0x33	/* 3 key */
#define VK_4		0x34	/* 4 key */
#define VK_5		0x35	/* 5 key */
#define VK_6		0x36	/* 6 key */
#define VK_7		0x37	/* 7 key */
#define VK_8		0x38	/* 8 key */
#define VK_9		0x39	/* 9 key */
/*	--		0x3A-0x40  Undefined */
#define VK_A		0x41	/* A key */
#define VK_B		0x42	/* B key */
#define VK_C		0x43	/* C key */
#define VK_D		0x44	/* D key */
#define VK_E		0x45	/* E key */
#define VK_F		0x46	/* F key */
#define VK_G		0x47	/* G key */
#define VK_H		0x48	/* H key */
#define VK_I		0x49	/* I key */
#define VK_J		0x4A	/* J key */
#define VK_K		0x4B	/* K key */
#define VK_L		0x4C	/* L key */
#define VK_M		0x4D	/* M key */
#define VK_N		0x4E	/* N key */
#define VK_O		0x4F	/* O key */
#define VK_P		0x50	/* P key */
#define VK_Q		0x51	/* Q key */
#define VK_R		0x52	/* R key */
#define VK_S		0x53	/* S key */
#define VK_T		0x54	/* T key */
#define VK_U		0x55	/* U key */
#define VK_V		0x56	/* V key */
#define VK_W		0x57	/* W key */
#define VK_X		0x58	/* X key */
#define VK_Y		0x59	/* Y key */
#define VK_Z		0x5A	/* Z key */
/* 	--		0x5B-0x5F  Undefined */
#define VK_NUMPAD0	0x60	/* Numeric keypad 0 key */
#define VK_NUMPAD1	0x61	/* Numeric keypad 1 key */
#define VK_NUMPAD2	0x62	/* Numeric keypad 2 key */
#define VK_NUMPAD3	0x63	/* Numeric keypad 3 key */
#define VK_NUMPAD4	0x64	/* Numeric keypad 4 key */
#define VK_NUMPAD5	0x65	/* Numeric keypad 5 key */
#define VK_NUMPAD6	0x66	/* Numeric keypad 6 key */
#define VK_NUMPAD7	0x67	/* Numeric keypad 7 key */
#define VK_NUMPAD8	0x68	/* Numeric keypad 8 key */
#define VK_NUMPAD9	0x69	/* Numeric keypad 9 key */
#define VK_MULTIPLY	0x6A	/* Multiply key */
#define VK_ADD		0x6B	/* Add key */
#define VK_SEPARATOR	0x6C	/* Separator key */
#define VK_SUBTRACT	0x6D	/* Subtract key */
#define VK_DECIMAL	0x6E	/* Decimal key */
#define VK_DIVIDE	0x6F	/* Divide key */
#define VK_F1		0x70	/* F1 key */
#define VK_F2		0x71	/* F2 key */
#define VK_F3		0x72	/* F3 key */
#define VK_F4		0x73	/* F4 key */
#define VK_F5		0x74	/* F5 key */
#define VK_F6		0x75	/* F6 key */
#define VK_F7		0x76	/* F7 key */
#define VK_F8		0x77	/* F8 key */
#define VK_F9		0x78	/* F9 key */
#define VK_F10		0x79	/* F10 key */
#define VK_F11		0x7A	/* F11 key */
#define VK_F12		0x7B	/* F12 key */
#define VK_F13		0x7C	/* F13 key */
#define VK_F14		0x7D	/* F14 key */
#define VK_F15		0x7E	/* F15 key */
#define VK_F16		0x7F	/* F16 key */
#define VK_F17		0x80	/* F17 key */
#define VK_F18		0x81	/* F18 key */
#define VK_F19		0x82	/* F19 key */
#define VK_F20		0x83	/* F20 key */
#define VK_F21		0x84	/* F21 key */
#define VK_F22		0x85	/* F22 key */
#define VK_F23		0x86	/* F23 key */
#define VK_F24		0x87	/* F24 key */
/*	--		0x88-0x8F  Unassigned */
#define VK_NUMLOCK	0x90	/* NUM LOCK key */
#define VK_SCROLL	0x91	/* SCROLL LOCK key */
/*	--		0x92-0xB9  Unassigned */
/*	--		0xBA-0xC0  OEM specific */
#define VK_US_SEMICOLON		0xBA	/* ; key */
#define VK_US_EQUAL		0xBB	/* = key */
#define VK_US_COMMA		0xBC	/* , key */
#define VK_US_MINUS		0xBD	/* - key */
#define VK_US_PERIOD		0xBE	/* . key */
#define VK_US_SLASH		0xBF	/* / key */
#define VK_US_GRAVE		0xC0	/* ` key */
/*	--		0xC1-0xDA  Unassigned */
/*	--		0xDB-0xE4  OEM specific */
#define VK_US_BRACELEFT		0xDB	/* [ key */
#define VK_US_BACKSLASH		0xDC	/* \ key */
#define VK_US_BRACERIGHT	0xDD	/* ] key */
#define VK_US_APOSTROPHE	0xDE	/* ' key */
/*	--		0xE5	   Unassigned */
/*	--		0xE6	   OEM specific */
/*	--		0xE7-0xE8  Unassigned */
/*	--		0xE9-0xF5  OEM specific */
/*	--		0xF6-0xEF  UUnassigned */

#endif /* fepkey_H */
