/*
  feplib.h : I/F of Fep Bridge 1.33
    Copyright (C) Yutaka Tamiya (tamy@flab.fujitsu.co) 1996/11/13
 */
#ifndef feplib_H
#define feplib_H

#include <stdio.h>
#include "fepkey.h"

#ifndef WCHAR_DEFINED
#	define WCHAR_DEFINED
	typedef unsigned short	wchar;
#endif

typedef struct {
    int		ApiLevel;	/* MS-KANJI API Level		*/
    char *	HostName;	/* Server name of of Fep Bridge	*/
    int		FepSocket;	/* Socket Handle of Fep Bridge	*/
    int		LenResultStr;	/* Length of KKC result string	*/
    int		LenModeStr;	/* Length of KKC mode string	*/
    int		LenSystemStr;	/* Length of KKC system string	*/
    int		LenBufStr;	/* Length of KKC buffer string	*/
    int		BufCursor;	/* Position of the cursor in BufStr	*/
    int		LenInitModeStr;	/* Length of initial mode	*/
    int		LenInitSystemStr;/* Length of initial system	*/
    /* Strings are encoded in WString */
    wchar *	ResultStr;	/* KKC result string		*/
    wchar *	ModeStr;	/* KKC mode string		*/
    wchar *	SystemStr;	/* KKC system string		*/
    wchar *	BufStr;		/* KKC (editting )buffer string	*/
    wchar *	InitModeStr;	/* Initial KKC mode string	*/
    wchar *	InitSystemStr;	/* Initial KKC system string	*/
    /* Attributes of each characters (DOS/V EGA mode)
       LSB bit0: Foreground Color Bule
           bit1: Foreground Color Green
           bit2: Foreground Color Red
           bit3: Foreground Color Intensity
           bit4: Background Color Bule
           bit5: Background Color Green
           bit6: Background Color Red
       MSB bit7: Background Color Intensity
    */
    char *	ModeAttr;	/* Attribute of KKC ModeStr	*/
    char *	SystemAttr;	/* Attribute of KKC SystemStr	*/
    char *	BufAttr;	/* Attribute of KKC BufStr	*/
    char *	InitModeAttr;	/* Attribute of initial KKC ModeStr	*/
    char *	InitSystemAttr;	/* Attribute of initial KKC SystemStr	*/

    int		MaxResultStr;	/* Maximum length of ResultStr	*/
    int		MaxModeStr;	/* Maximum length of ModeStr	*/
    int		MaxSystemStr;	/* Maximum length of SystemStr	*/
    int		MaxBufStr;	/* Maximum length of BufStr	*/

} FepResult;

typedef struct {
    char	scan;
    char	ascii;
    char	shift;
} FepKeyCode;

typedef enum {
    FepWaitForever,	/* Wait until making connection to Fep Bridge	*/
    FepDetectTimeOut,	/* Try to connect to Fep Bridge for 1 sec.	*/
    			/*   If failed, FepInitialize() will abort	*/ 
    FepNoAuth		/* Open connection, but not send FI_Auth.	*/
} FepConnectMode;

/* Bit masks of conversion flags for FepOn() and FepKKC() */
#define	FepMaskKanaHanToZen	(1<<0)	/* convert hankaku kana to zenkaku */
#define	FepMaskAsciiZenToHan	(1<<1)	/* convert zenkaku ascii to hankaku */

/*
    I/F Functions to Fep Bridge
*/
/* Initialize & terminatates Fep Bridge I/F */
extern int	FepInitialize(FepResult *result, const char *fepHost,
							FepConnectMode mode);
extern void	FepTerminate(FepResult *result);

/* Ask if Fep Bridge is alive */
extern int	FepIsAlive(FepResult *result);

/* Command for the Fep Bridge */
extern int	FepInfo(FepResult *result, int convFlags);
extern int	FepReset(FepResult *result, int convFlags);
extern int	FepConv(FepResult *result, const FepKeyCode *keyCode,
							    int convFlags);
extern void	FepKill(const FepResult *result);

/* Print out user information */
extern void	FepUserList(FepResult *result, FILE* fpOut);

#endif /* feplib_H */
