/*
  fepstat.c : I/F of Fep Bridge 1.33
    Copyright (C) Yutaka Tamiya (tamy@flab.fujitsu.co) 1996/11/13
 */
#ifndef FIFO_DIR
#	define FIFO_DIR	"/var/lib/fepbrdg"
#endif
#define FIFO_FEP_IN	"fepin"
#define FIFO_FEP_OUT	"fepout"

#include "feplib.h"
#include "fepbrdg.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <setjmp.h>
#include <signal.h>
#include <fcntl.h>

static int	isFepAliveFifo(void);

/* time out handler */
static jmp_buf	Env;

static void
timeOutHandler(int sig)
{
    signal(sig, timeOutHandler);
    longjmp(Env, 1);
}

int
main(int ac, char *av[])
{
    FepResult	res;
    int		i;
    const char	*fepHost = 0;
    int		userList = 0;
    int		fifoAccess = 0;

    static void	timeOutHandler(int sig);

    for (i = 1; i < ac; i ++) {
	if (*av[i] == '-') {
	    if (strcmp("-fephost", av[i]) == 0) {
		fepHost = av[++i];
	    } else if (strcmp("-list", av[i]) == 0) {
		userList = 1;
	    } else if (strcmp("-fifo", av[i]) == 0) {
		fifoAccess = 1;
	    } else {
		printf("%s: illeagal option %s\n", av[0], av[i]);
		printf("%s [-fephost <hostname>] [-list] [-fifo]\n", av[0]);
		return 1;
	    }
	} else {
	    fepHost = av[i];
	}
    }

    signal(SIGALRM, timeOutHandler);
    
    if (setjmp(Env) == 0) {
	alarm(1);	/* wait 1 sec for the response of fepbrdg */
    } else {
	/* detect TimeOut */
	/* return from timeOutHandler() */
	printf("Fep Bridge cannot be Connected (Maybe Dead).\n");
	return 1;
    }

    if (fifoAccess == 0) {
	int	ans = 0;
	if (FepInitialize(&res, fepHost, FepNoAuth) == 0) {
	    printf("Fep Bridge is NOT Available.\n");
	    return 1;
	} else if (userList) {
	    FepUserList(&res, stdout);
	} else {
	    if (FepIsAlive(&res) == 0) {
	  	printf("Fep Bridge is NOT Available.\n");
	        ans = 1;
	    } else {
		printf("Fep Bridge is Available.\n");
	    }
	}
	FepTerminate(&res);
	return ans;
    } else {
	if (isFepAliveFifo()) {
	    printf("Fep Bridge is Available.\n");
	    return 0;
	} else {
	    printf("Fep Bridge is Dead.\n");
	    return 1;
	}
    }
}

static int
max(int x, int y)
{
    return (x > y) ? x : y;
}

static int
isFepAliveFifo(void)
{
    int		ans;
    char	c;
    char	*buf = 0;
    int		len, pathLen;
    int		fepIn = -1, fepOut = -1;
    
    len = strlen(FIFO_DIR);
    pathLen = len + 1 + max(strlen(FIFO_FEP_IN), strlen(FIFO_FEP_OUT)) + 1;
    if ((buf = (char *)malloc(pathLen)) == 0) {
	perror("");
	return 0;
    }
    strcpy(buf, FIFO_DIR);
    buf[len] = '/';

    strcpy(&buf[len+1], FIFO_FEP_IN);
    if ((fepIn = open(buf, O_WRONLY)) < 0) {
	perror(buf);
	ans = 0;
	goto Exit;
    }
    strcpy(&buf[len+1], FIFO_FEP_OUT);
    if ((fepOut = open(buf, O_RDONLY)) < 0) {
	perror(buf);
	ans = 0;
	goto Exit;
    }

    c = FI_Stat;
    if (write(fepIn, &c, 1) != 1
    || read(fepOut, &c, 1) != 1) {
	ans = 0;
	goto Exit;
    }

    if (c == FO_Alive) {
	ans = 1;
    } else {
	/* wrong answer. may be insane. */
	ans = 0;
    }
    
Exit:
    if (fepIn >= 0)   close(fepIn);
    if (fepOut >= 0)  close(fepOut);
    if (buf != 0)  free(buf);
    return ans;
}
