/*
  This file is taken from screen-3.2b
*/
/* Copyright (c) 1991
 *      Juergen Weigert (jnweiger@immd4.informatik.uni-erlangen.de)
 *      Michael Schroeder (mlschroe@immd4.informatik.uni-erlangen.de)
 * Copyright (c) 1987 Oliver Laumann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Noteworthy contributors to screen's design and implementation:
 *	Wayne Davison (davison@borland.com)
 *	Patrick Wolfe (pat@kai.com, kailand!pat)
 *	Bart Schaefer (schaefer@cse.ogi.edu)
 *	Nathan Glasser (nathan@brokaw.lcs.mit.edu)
 *	Larry W. Virden (lwv27%cas.BITNET@CUNYVM.CUNY.Edu)
 *	Howard Chu (hyc@hanauma.jpl.nasa.gov)
 *	Tim MacKenzie (tym@dibbler.cs.monash.edu.au)
 *	Markku Jarvinen (mta@{cc,cs,ee}.tut.fi)
 *	Marc Boucher (marc@CAM.ORG)
 *
 ****************************************************************
 */

#include <sys/param.h>
#include <ctype.h>
#include <fcntl.h>
#ifdef sgi
# include <sys/sysmacros.h>
#endif /* sgi */

#include <sys/types.h>
#ifndef sgi
# include <sys/file.h>
#endif /* sgi */
#ifndef sun
# include <sys/ioctl.h>
#endif /* sun */

#ifdef ISC
# include <sys/tty.h>
# include <sys/sioctl.h>
# include <sys/pty.h>
#endif

/*
 * The characters to try when finding pty names to open,
 * on some types of systems.
 * Try each char in PTY_FIRST_CHARS followed by each char in
 * PTY_SECOND_CHARS in succession until finding a free pty.
 */
#ifdef MIPS
#define PTY_FIRST_CHARS "zyxwvutsrqp"
#endif

#ifdef hpux
#define PTY_FIRST_CHARS "pqrstuvw"
#endif

#ifdef sun
#define PTY_FIRST_CHARS "qprstuvwxyzPQRST"
#endif

#ifndef PTY_FIRST_CHARS
#define PTY_FIRST_CHARS "qpr"
#endif

#ifndef PTY_SECOND_CHARS
#define PTY_SECOND_CHARS "0123456789abcdef"
#endif

char PtyName[32];
char TtyName[32];

#ifdef hpux
char PtyProto[] = "/dev/ptym/ptyXY";
char TtyProto[] = "/dev/pty/ttyXY";
#else
# if !(defined(sequent) || defined(_SEQUENT_) || defined(SVR4))
static char PtyProto[] = "/dev/ptyXY";
static char TtyProto[] = "/dev/ttyXY";
# endif
#endif /* hpux */

#if defined(sequent) || defined(_SEQUENT_) || defined(SVR4)

int OpenPTY(void)
{
  char *m, *s;
  register int f;
# ifdef SVR4
  char *ptsname();
  sig_t (*sigcld)();

  if ((f = open("/dev/ptmx", O_RDWR)) == -1)
    return(-1);

  /*
   * SIGCLD set to SIG_DFL for grantpt() because it fork()s and
   * exec()s pt_chmod
   */
  sigcld = signal(SIGCLD, SIG_DFL);
       
  if ((m = ptsname(f)) == NULL || unlockpt(f) || grantpt(f))
    {
      signal(SIGCLD, sigcld);
      close(f);
      return(-1);
    } 
  signal(SIGCLD, sigcld);
  strncpy(TtyName, m, sizeof TtyName);
# else /* SVR4 */
  if ((f = getpseudotty(&s, &m)) < 0)
    return(-1);
  strncpy(PtyName, m, sizeof PtyName);
  strncpy(TtyName, s, sizeof TtyName);
# endif /* SVR4 */
# ifdef POSIX
  tcflush(f, TCIOFLUSH);
# else
  (void) ioctl(f, TIOCFLUSH, (char *) 0);
# endif
  (void) ioctl(f, TIOCEXCL, (char *) 0);
  return (f);
}

#else /* defined(sequent) || defined(_SEQUENT_) || defined(SVR4) */
# ifdef MIPS

int OpenPTY(void)
{
  register char *p, *l, *d;
  register f, tf;
  register my_minor;
  struct stat buf;
   
  strcpy(PtyName, PtyProto);
  for (p = PtyName; *p != 'X'; ++p)
    ;
  for (l = PTY_FIRST_CHARS; *p = *l; ++l)
    {
      for (d = PTY_SECOND_CHARS; p[1] = *d; ++d)
	{
	  if ((f = open(PtyName, O_RDWR)) != -1)
	    {
	      fstat(f, &buf);
	      my_minor = minor(buf.st_rdev);
	      sprintf(TtyName, "/dev/ttyq%d", my_minor);
	      if ((tf = open(TtyName, O_RDWR)) != -1)
		{
		  close(tf);
		  (void) ioctl(f, TIOCEXCL, (char *)0);
		  return f;
		}
	      close(f);
	    }
	}
    }
  return -1;
}

# else  /* MIPS */
#  ifdef sgi

int OpenPTY(void)
{
  register f;
  register my_minor;
  struct stat buf;
   
  strcpy(PtyName, "/dev/ptc");
  f = open(PtyName, O_RDWR|O_NDELAY);
  if (f >= 0)
    {
      if (fstat(f, &buf) < 0)
	{
	  close(f);
	  return -1;
	}
      my_minor = minor(buf.st_rdev);
      sprintf(TtyName, "/dev/ttyq%d", my_minor);
    }
  return f;
}

#  else /* sgi */
#   ifdef _AIX /* RS6000 */

int OpenPTY(void)
{
  register int i, f, tf;

  strcpy (PtyName, "/dev/ptc");
  if ((f = open (PtyName, O_RDWR)) != -1)
    {
      strcpy (TtyName, ttyname (f));
      strcpy (PtyName, TtyName);
      PtyName [7] = 'c';
      if (ioctl (f, TIOCEXCL, (char *) 0) == -1)
	return -1;
      if ((tf = open (TtyName, O_RDWR)) != -1)
	{
	  close (tf);
	  return f;
	}
      close (f);
    }
  return -1;
}

#   else /* _AIX, RS6000 */
#     ifdef linux

int OpenPTY(void)
{
  register char *p, *q, *l, *d;
  register int f, tf;

  strcpy(PtyName, PtyProto);
  strcpy(TtyName, TtyProto);
  for (p = PtyName; *p != 'X'; ++p)
    ;
  for (q = TtyName; *q != 'X'; ++q)
    ;
  for (l = PTY_FIRST_CHARS; (*p = *l) != '\0'; ++l)
    {
      for (d = PTY_SECOND_CHARS; (p[1] = *d) != '\0'; ++d)
	{
	  if ((f = open(PtyName, O_RDWR)) != -1)
	    {
	      q[0] = *l;
	      q[1] = *d;
	      (void) ioctl(f, TIOCEXCL, (char *) 0);
	      return f;
	    }
	}
    }
  return -1;
}

#     else /* linux */

int OpenPTY(void)
{
  register char *p, *q, *l, *d;
  register int f, tf;

#    if !defined(hpux)
  debug("Hello, You are none of: sequent, _SEQUENT_, SVR4, MIPS, sgi, AIX, linux\n");
  debug("       This OpenPTY() is for hpux, ... and for you?\n");
#    endif
  strcpy(PtyName, PtyProto);
  strcpy(TtyName, TtyProto);
  for (p = PtyName; *p != 'X'; ++p)
    ;
  for (q = TtyName; *q != 'X'; ++q)
    ;
  for (l = PTY_FIRST_CHARS; (*p = *l) != '\0'; ++l)
    {
      for (d = PTY_SECOND_CHARS; (p[1] = *d) != '\0'; ++d)
	{
	  if ((f = open(PtyName, O_RDWR)) != -1)
	    {
	      q[0] = *l;
	      q[1] = *d;
	      if ((tf = open(TtyName, O_RDWR)) != -1)
		{
		  /* close tf, thus we also get rid of an unwanted
		   * controlling terminal! 
		   */
		  close(tf);
		  (void) ioctl(f, TIOCEXCL, (char *) 0);
		  return f;
		}
	      close(f);
	    }
	}
    }
  return -1;
}

#    endif /* linux */
#   endif /* _AIX, RS6000 */
#  endif /* sgi */
# endif /* MIPS */
#endif
