/*
  Fep.c : I/F of Fep Bridge 1.33
    Copyright (C) Yutaka Tamiya (tamy@flab.fujitsu.co) 1996/11/13
 */
/*
 * Copyright (c) 1990  Software Research Associates, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 *
 * Author:  Makoto Ishisone, Software Research Associates, Inc., Japan
 */

/* Copyright 1991 NEC Corporation, Tokyo, Japan.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of NEC Corporation
 * not be used in advertising or publicity pertaining to distribution
 * of the software without specific, written prior permission.  NEC 
 * Corporation makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express
 * or implied warranty.
 *
 * NEC CORPORATION DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN 
 * NO EVENT SHALL NEC CORPORATION BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF 
 * USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR 
 * OTHER TORTUOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR 
 * PERFORMANCE OF THIS SOFTWARE. 
 *
 * Author: Akira Kon, NEC Corporation. (kon@d1.bs2.mt.nec.co.jp)
 */

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xmu/Atoms.h>
#include <X11/keysym.h>
#include <X11/Xfuncs.h>
#include "FepP.h"
#include "feplib.h"

static XtResource resources[] = {
#define offset(field) XtOffset(FepObject, fep.field)
    { XtNconversionEndKeys, XtCConversionEndKeys, XtRString, sizeof(String),
	  offset(ConvEndKeysStr), XtRImmediate, (XtPointer)NULL},
    { XtNfepHost, XtCFepHost, XtRString, sizeof(String),
	  offset(FepHost), XtRImmediate, (XtPointer)NULL},
    { XtNhankakuKanaToZenkaku, XtCHankakuKanaToZenkaku, XtRBoolean,
	  sizeof(Boolean),
	  offset(KanaHanToZen), XtRImmediate, (XtPointer)False },
    { XtNzenkakuAsciiToHankaku, XtCZenkakuAsciiToHankaku, XtRBoolean,
	  sizeof(Boolean),
	  offset(AsciiZenToHan), XtRImmediate, (XtPointer)False },
    { XtNalwaysShowSystemStr, XtCAlwaysShowSystemStr, XtRBoolean,
	  sizeof(Boolean),
	  offset(AlwaysShowSystemStr), XtRImmediate, (XtPointer)False },
    { XtNbufAttrMask, XtCBufAttrMask, XtRInt, sizeof(int),
	  offset(BufAttrMask), XtRImmediate, (XtPointer)0xf0 },
    { XtNsystemAttrMask, XtCSystemAttrMask, XtRInt, sizeof(int),
	  offset(SystemAttrMask), XtRImmediate, (XtPointer)0xf0 },
    { XtNswapBackDeleteKeys, XtCSwapBackDeleteKeys, XtRBoolean,
	  sizeof(Boolean),
	  offset(SwapBackDeleteKeys), XtRImmediate, (XtPointer)False },
    /*
      Default End Conversion Keys is not defined here, but in
      "static FepTriggerKey DefaultEndKey", which is declared later
    */
#undef offset
};

/*
  I/F Functions to Kinput2
*/
static void ClassInitialize(void);
static void Initialize(Widget req, Widget new,
			ArgList args, Cardinal *num_args);
static void Destroy(Widget w);
static Boolean SetValues(Widget cur, Widget req, Widget wid,
			ArgList args, Cardinal *num_args);
static int InputEvent(Widget w, XEvent *event);
static ICString *GetMode(Widget w);
static int CursorPos(Widget w, Cardinal *nsegp, Cardinal *ncharp);
static int NumSegments(Widget w);
static ICString *GetSegment(Widget w, Cardinal n);
static int CompareSegment(Widget w,
			ICString *seg1, ICString *seg2, Cardinal *n);
static ICString *GetItemList(Widget w, Cardinal *n);
static int SelectItem(Widget w, int n);
static int ConvertedString(Widget w, Atom *encoding, int *format,
					int *length, XtPointer *string);
static int ClearConversion(Widget w);
static ICString *GetAuxSegments(Widget w, Cardinal *n,
					Cardinal *ns, Cardinal *nc);

FepClassRec fepClassRec = {
  { /* object fields */
    /* superclass		*/	(WidgetClass) &inputConvClassRec,
    /* class_name		*/	"Fep",
    /* widget_size		*/	sizeof(FepRec),
    /* class_initialize		*/	ClassInitialize,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* obj1			*/	NULL,
    /* obj2			*/	NULL,
    /* obj3			*/	0,
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* obj4			*/	FALSE,
    /* obj5			*/	FALSE,
    /* obj6			*/	FALSE,
    /* obj7			*/	FALSE,
    /* destroy			*/	Destroy,
    /* obj8			*/	NULL,
    /* obj9			*/	NULL,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* obj10			*/	NULL,
    /* get_values_hook		*/	NULL,
    /* obj11			*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* obj12			*/	NULL,
    /* obj13			*/	NULL,
    /* obj14			*/	NULL,
    /* extension		*/	NULL
  },
  { /* inputConv fields */
    /* InputEvent		*/	InputEvent,
    /* GetMode			*/	GetMode,
    /* CursorPos		*/	CursorPos,
    /* NumSegments		*/	NumSegments,
    /* GetSegment		*/	GetSegment,
    /* CompareSegment		*/	CompareSegment,
    /* GetItemList		*/	GetItemList,
    /* SelectItem		*/	SelectItem,
    /* GetConvertedString	*/	ConvertedString,
    /* ClearConversion		*/	ClearConversion,
    /* GetAuxSegments		*/	GetAuxSegments,
    /* SupportMultipleObjects	*/	True,
#if defined(XtInheritGetTriggerKeys)	/* after Kinput2 ver 2.0 fix 2 */
    /* GetTriggerKeys		*/	XtInheritGetTriggerKeys,
    /* num_trigger_keys		*/	0,
    /* trigger_keys		*/	NULL,
#endif
    /* NoMoreObjects		*/	False,
  },
  { /* fep fields */
      {/* Result */
	/* ApiLevel		*/	(int)0,
	/* HostName		*/	(char *)NULL,
	/* FepSocket		*/	(int)-1,
	/* LenResultStr		*/	(int)0,
        /* LenModeStr		*/	(int)0,
        /* LenSystemStr		*/	(int)0,
        /* LenBufStr		*/	(int)0,
        /* BufCursor		*/	(int)0,
        /* LenInitModeStr	*/	(int)0,
        /* LenInitSystemStr	*/	(int)0,
        /* ResultStr		*/	(wchar *)NULL,
        /* ModeStr		*/	(wchar *)NULL,
        /* SystemStr		*/	(wchar *)NULL,
        /* BufStr		*/	(wchar *)NULL,
        /* InitModeStr		*/	(wchar *)NULL,
        /* InitSystemStr	*/	(wchar *)NULL,
        /* ModeAttr		*/	(char *)NULL,
        /* SystemAttr		*/	(char *)NULL,
        /* BufAttr		*/	(char *)NULL,
        /* InitModeAttr		*/	(char *)NULL,
        /* InitSystemAttr	*/	(char *)NULL,
        /* MaxResultStr		*/	(int)0,
        /* MaxModeStr		*/	(int)0,
        /* MaxSystemStr		*/	(int)0,
        /* MaxBufStr		*/	(int)0,
    },				
    /* ShowSystemStr		*/	False,
    /* NumModeSeg		*/	(int)0,
    /* IndexModeSeg		*/	(int *)NULL,
    /* NumSystemSeg		*/	(int)0,
    /* IndexSystemSeg		*/	(int *)NULL,
    /* NumBufSeg		*/	(int)0,
    /* IndexBufSeg		*/	(int *)NULL,
    /* PrevModeStr		*/	(wchar *)NULL,
  }
};

WidgetClass fepObjectClass = (WidgetClass)&fepClassRec;

/* local functions */
static int	initializeFep(FepObject obj);
static void	terminateFep(FepObject obj);
static void	makeSegment(FepObject obj);
static void	displayPreEdit(FepObject obj);
static int	convKeyEventToFepKeyCode(KeySym ks, const XKeyEvent *xkey,
						     FepKeyCode *keyCode);
static void	getTriggerKeys(FepObject obj);

/*
  CallBack routines
*/
static int
InputEvent(Widget w, XEvent *event)
{
    FepObject obj = (FepObject)w;
    FepObjectClass class = (FepObjectClass)XtClass(w);
    char	buf[1024];
    KeySym	ks;
    const FepTriggerKey	*endKey;
    int		i;
    XComposeStatus   composeStatus;
    FepKeyCode	keyCode;
    int		rtn = -1;

    if (event->type != KeyPress) return 0;

    if (XLookupString(&event->xkey, buf, 1024, &ks, &composeStatus) == 0) {
	buf[0] = 0;
    }

    /* Check end conversion */
    endKey = obj->fep.ConvEndKeys;
    for (i = 0; i < obj->fep.NumConvEndKeys; i++, endKey ++) {
	if (ks == endKey->keysym
	&& (event->xkey.state & endKey->check_modifiers) == endKey->modifiers){
	    /* Notify End conversion */
	    XtCallCallbackList(w, obj->inputConv.endcallback, (XtPointer)NULL);
	    class->fep_class.ShowSystemStr = False;
	    /* endcallback pops down the aux region */
	    return 0;
	}
    }

    keyCode.ascii = buf[0];
    if (! convKeyEventToFepKeyCode(ks, &event->xkey, &keyCode))
	return 1;

    rtn = FepConv(&class->fep_class.Result, &keyCode, obj->fep.ConvFlags);
    	/* return 1, if Event should be sent back */
    	/*        0, otherwise			  */
    if (rtn == 0) {
	makeSegment(obj);
	if (class->fep_class.Result.LenResultStr > 0) {
	    /* Notify Fix */
	    XtCallCallbackList(w, obj->inputConv.fixcallback, (XtPointer)NULL);
	}
	displayPreEdit(obj);
    }

    return rtn;
}

static ICString *
GetMode(Widget w)
{
    FepObject obj = (FepObject)w;
    FepObjectClass class = (FepObjectClass)XtClass(w);
    static ICString icstr;

    if (class->fep_class.Result.LenModeStr > 0) {
	icstr.data = (char *)class->fep_class.Result.ModeStr;
	icstr.nchars = class->fep_class.Result.LenModeStr;
	icstr.nbytes = icstr.nchars * sizeof(wchar);
	icstr.attr = ICAttrNormalString;
	return &icstr;
    } else {
	return NULL;
    }
}

static int
CursorPos(Widget w, Cardinal *nsegp, Cardinal *ncharp)
{
    FepObject obj = (FepObject)w;
    FepObjectClass class = (FepObjectClass)XtClass(w);

    *nsegp = 0;
    *ncharp = class->fep_class.Result.BufCursor;
    return 0;
}

static int
NumSegments(Widget w)
{
    FepObject obj = (FepObject)w;
    FepObjectClass class = (FepObjectClass)XtClass(w);

    return class->fep_class.NumBufSeg;
}

static ICString *
GetSegment(Widget w, Cardinal n)
{
    FepObject obj = (FepObject)w;
    FepObjectClass class = (FepObjectClass)XtClass(w);
    static ICString icstr;
    const int	index = class->fep_class.IndexBufSeg[n];
    const int	nextIndex = class->fep_class.IndexBufSeg[n+1];

    icstr.data = (char *)&class->fep_class.Result.BufStr[index];
    icstr.nchars = nextIndex - index;
    icstr.nbytes = icstr.nchars * sizeof(wchar);
    if (class->fep_class.Result.BufAttr[index] & obj->fep.BufAttrMask) {
	icstr.attr = ICAttrConverted | ICAttrCurrentSegment;
    } else {
	icstr.attr = ICAttrNotConverted;
    }
    return &icstr;
}

/* ARGSUSED */
static int
CompareSegment(Widget w, ICString *seg1, ICString *seg2, Cardinal *n)
{
    wchar *p, *q;
    int len, nsame;
    int result = 0;

    if (seg1->attr != seg2->attr) result |= ICAttrChanged;

    len = seg1->nchars > seg2->nchars ? seg2->nchars : seg1->nchars;
    nsame = 0;
    p = (wchar *)seg1->data;
    q = (wchar *)seg2->data;
    while (nsame < len && *p++ == *q++) nsame++;

    if (nsame != len || len != seg1->nchars || len != seg2->nchars)
	result |= ICStringChanged;

    if (n) *n = nsame;

    return result;
}

static ICString *
GetItemList(Widget w, Cardinal *n)
{
    FepObject obj = (FepObject)w;
    FepObjectClass class = (FepObjectClass)XtClass(w);

    XtAppWarning(XtWidgetToApplicationContext(w),
		 "Fep Object: GetItemList should not be called");
    *n = 0;
    return NULL;
}

static int
SelectItem(Widget w, int n)
{
    FepObject obj = (FepObject)w;
    FepObjectClass class = (FepObjectClass)XtClass(w);

    XtAppWarning(XtWidgetToApplicationContext(w),
		 "Fep Object: SelectItem should not be called");
    return 0;
}

static int
ConvertedString(Widget w, Atom *encoding, int *format, int *length,
		XtPointer *string)
{
    FepObject obj = (FepObject)w;
    FepObjectClass class = (FepObjectClass)XtClass(w);
    int	len;
    wchar  *wp;


    /* Convert to COMPOUND_TEXT */
    *encoding = XA_COMPOUND_TEXT(XtDisplayOfObject((Widget)obj));
    *format = 8;
    /* Convert '\r' to '\n' */
    for (wp = class->fep_class.Result.ResultStr; *wp != 0; wp++) {
	if (*wp == '\r') *wp = '\n';
    }
    *length = len = convJWStoCT(class->fep_class.Result.ResultStr,
				(unsigned char *)NULL, 0);
    *string = XtMalloc(len + 1);
    if (*string == NULL) {
	XtAppError(XtWidgetToApplicationContext((Widget)obj),
		   "FepObject: Not enough memory");
    }
    (void)convJWStoCT(class->fep_class.Result.ResultStr,
		      (unsigned char *)*string, 0);
    return 0;
}

static int
ClearConversion(Widget w)
{
    FepObject obj = (FepObject)w;
    FepObjectClass class = (FepObjectClass)XtClass(w);

    FepReset(&class->fep_class.Result, obj->fep.ConvFlags);
    makeSegment(obj);

    return 0;
}

static ICString *
GetAuxSegments(Widget w, Cardinal *n, Cardinal *ns, Cardinal *nc)
{
    FepObject obj = (FepObject)w;
    FepObjectClass class = (FepObjectClass)XtClass(w);
    int		i;
    static ICString *icstr = NULL;
    static allocatedIcs = 0;

    *n = class->fep_class.NumSystemSeg;
    if (*n > allocatedIcs) {
	if (icstr != NULL) XtFree((char *)icstr);
	icstr = (ICString *)XtMalloc(*n * sizeof(ICString));
	if (icstr == NULL) {
	    XtAppError(XtWidgetToApplicationContext(w),
		       "FepObject: Not enough memory");
	}
	allocatedIcs = *n;
    }
    *ns = 0; *nc = 0;
    for (i = 0; i < *n; i++) {
	const int	index = class->fep_class.IndexSystemSeg[i];
	const int	nextIndex = class->fep_class.IndexSystemSeg[i+1];

	icstr[i].data = (char *)&class->fep_class.Result.SystemStr[index];
	icstr[i].nchars = nextIndex - index;
	icstr[i].nbytes = icstr[i].nchars * sizeof(wchar);
	if (class->fep_class.Result.SystemAttr[index]
	   & obj->fep.SystemAttrMask) {
	    icstr[i].attr = ICAttrConverted | ICAttrCurrentSegment;
	} else {
	    icstr[i].attr = ICAttrNormalString;
	}
    }
    return icstr;
}


/* ARGSUSED */
static void
ClassInitialize(void)
{
    /* nothing */
}

static int	NumObjectCreated = 0;	/* instance counter */
static int	SwapBackDeleteKeys;	/* Swap BackSpace & Delete keys ? */

/* ARGSUSED */
static void
Initialize(Widget req, Widget new, ArgList args, Cardinal *num_args)
{
    FepObject obj = (FepObject)new;
    FepObjectClass class = (FepObjectClass)XtClass(new);

    obj->fep.ConvFlags = 0;
    SwapBackDeleteKeys = obj->fep.SwapBackDeleteKeys;
    if (obj->fep.KanaHanToZen)  obj->fep.ConvFlags |= FepMaskKanaHanToZen;
    if (obj->fep.AsciiZenToHan) obj->fep.ConvFlags |= FepMaskAsciiZenToHan;

    if (NumObjectCreated ++ == 0) {
	/* Create the first object */
	/* Wait until making connection to the Fep Bridge */
	if (FepInitialize(&class->fep_class.Result,
				    obj->fep.FepHost, FepWaitForever) == 0) {
	    XtAppError(XtWidgetToApplicationContext((Widget)new),
			 "Fep Object: cannot connect to the Fep Bridge host.");
	    return;
	}
	if (FepInfo(&class->fep_class.Result, obj->fep.ConvFlags) == 0) {
	    XtAppError(XtWidgetToApplicationContext((Widget)new),
			 "Fep Object: cannot get Fep Bridge information.");
	    return;
	}
	if (initializeFep(obj) == 0) {
	    XtAppError(XtWidgetToApplicationContext((Widget)new),
			 "Fep Object: cannot initialize the internal data.");
	    return;
	}
	FepReset(&class->fep_class.Result, obj->fep.ConvFlags);
	makeSegment(obj);
    } else {
	/* Create not the first object */
	FepReset(&class->fep_class.Result, obj->fep.ConvFlags);
	makeSegment(obj);
    }
    getTriggerKeys(obj);
}

/* ARGSUSED */
static void
Destroy(Widget w)
{
    FepObject obj = (FepObject)w;
    FepObjectClass class = (FepObjectClass)XtClass(w);

    if (-- NumObjectCreated == 0) {
	FepTerminate(&class->fep_class.Result);
	terminateFep(obj);
    }
}

/* ARGSUSED */
static Boolean
SetValues(Widget cur, Widget req, Widget wid, ArgList args, Cardinal *num_args)
{
    FepObject old = (FepObject)cur;
    FepObject new = (FepObject)wid;

    return False;	     
}

/*
  Conversion from KeyEvent to PC KeyCode
*/
static int
convKeyEventToFepKeyCode(KeySym ks, const XKeyEvent *xkey,
				FepKeyCode *keyCode /* return value */)
{
    int		keycode;
    static int	xKeySymToFepKeyCode(KeySym ks, unsigned int keyState);

    if ((keycode = xKeySymToFepKeyCode(ks, xkey->state)) < 0) {
	return 0;
    }
    keyCode->scan = keycode;
    keyCode->shift = 0;
    if (xkey->state & ShiftMask)    keyCode->shift |= 0x01;
    if (xkey->state & ControlMask)  keyCode->shift |= 0x02;
    if (xkey->state & Mod1Mask)     keyCode->shift |= 0x04;

    return 1;
}

static int
xKeySymToFepKeyCode(KeySym ks, unsigned int keyState)
{
    /* Return MS-Windows key code */
    switch (ks) {
    case XK_Escape:	return VK_ESCAPE;
    case XK_BackSpace:	return ! SwapBackDeleteKeys ? VK_BACK : VK_DELETE;
    case XK_Tab:	return VK_TAB;
    case XK_Return:	return VK_RETURN;
    case XK_F1:		return VK_F1;
    case XK_F2:		return VK_F2;
    case XK_F3:		return VK_F3;
    case XK_F4:		return VK_F4;
    case XK_F5:		return VK_F5;
    case XK_F6:		return VK_F6;
    case XK_F7:		return VK_F7;
    case XK_F8:		return VK_F8;
    case XK_F9:		return VK_F9;
    case XK_F10:	return VK_F10;
    case XK_F11:	return VK_F11;
    case XK_F12:	return VK_F12;
    case XK_F13:	return VK_F13;
    case XK_F14:	return VK_F14;
    case XK_F15:	return VK_F15;
    case XK_F16:	return VK_F16;
    case XK_F17:	return VK_F17;
    case XK_F18:	return VK_F18;
    case XK_F19:	return VK_F19;
    case XK_F20:	return VK_F20;
    case XK_F21:	return VK_F21;
    case XK_F22:	return VK_F22;
    case XK_F23:	return VK_F23;
    case XK_F24:	return VK_F24;
    case XK_KP_Multiply:return VK_MULTIPLY;
    case XK_KP_7:	return VK_NUMPAD7;
    case XK_KP_8:	return VK_NUMPAD8;
    case XK_KP_9:	return VK_NUMPAD9;
    case XK_KP_Subtract:return VK_SUBTRACT;
    case XK_KP_4:	return VK_NUMPAD4;
    case XK_KP_5:	return VK_NUMPAD5;
    case XK_KP_6:	return VK_NUMPAD6;
    case XK_KP_Add:	return VK_ADD;
    case XK_KP_1:	return VK_NUMPAD1;
    case XK_KP_2:	return VK_NUMPAD2;
    case XK_KP_3:	return VK_NUMPAD3;
    case XK_KP_0:	return VK_NUMPAD0;
    case XK_KP_Decimal:	return VK_DECIMAL;
    case XK_KP_Divide:	return VK_DIVIDE;
    case XK_KP_Enter:	return VK_RETURN;

#ifdef XK_KP_Home		/* for R6 */
    case XK_KP_Home:
#endif
    case XK_Home:	return VK_HOME;
#ifdef XK_KP_Up			/* for R6 */
    case XK_KP_Up:
#endif
    case XK_Up:		return VK_UP;
#ifdef XK_KP_Prior		/* for R6 */
    case XK_KP_Prior:
#endif
    case XK_Prior:	return VK_PRIOR;
#ifdef XK_KP_Left		/* for R6 */
    case XK_KP_Left:
#endif
    case XK_Left:	return VK_LEFT;
#ifdef XK_KP_Right		/* for R6 */
    case XK_KP_Right:
#endif
    case XK_Right:	return VK_RIGHT;
#ifdef XK_KP_End		/* for R6 */
    case XK_KP_End:
#endif
    case XK_End:	return VK_END;
#ifdef XK_KP_Down		/* for R6 */
    case XK_KP_Down:
#endif
    case XK_Down:	return VK_DOWN;
#ifdef XK_KP_Next		/* for R6 */
    case XK_KP_Next:
#endif
    case XK_Next:	return VK_NEXT;
#ifdef XK_KP_Insert		/* for R6 */
    case XK_KP_Insert:
#endif
    case XK_Insert:	return VK_INSERT;
#ifdef XK_KP_Delete		/* for R6 */
    case XK_KP_Delete:
#endif
    case XK_Delete:	return ! SwapBackDeleteKeys ? VK_DELETE : VK_BACK;

    case XK_Control_R:	return VK_CONTROL;
    case XK_Print:	return VK_SNAPSHOT;
    case XK_Select:	return VK_SELECT;
    case XK_Execute:	return VK_EXECUTE;
    case XK_Help:	return VK_HELP;

    case XK_Control_L:	return VK_CONTROL;
    case XK_Shift_L:
    case XK_Shift_R:	return VK_SHIFT;
    case XK_Alt_L:
    case XK_Meta_L:
    case XK_Alt_R:
    case XK_Meta_R:	return VK_MENU;
    case XK_Caps_Lock:
    case XK_Shift_Lock:	return VK_CAPITAL;
    case XK_Num_Lock:	return VK_NUMLOCK;
    case XK_Scroll_Lock:
    case XK_Multi_key:	return VK_SCROLL;

    /* Special keys of Japanese 106 keyboard */
    case XK_Kanji:	return VK_KANJI;
    case XK_Hiragana_Katakana:
    case XK_Hiragana:
    case XK_Katakana:	return VK_KANA;
    case XK_Henkan:	return VK_CONVERT;
    case XK_Muhenkan:	return VK_NONCONVERT;

    default:		return 0;
    }
}

static int
isInitSystemStr(const FepResult *result)
{
    if (result->LenSystemStr == result->LenInitSystemStr
    && wstrcmp(result->SystemStr, result->InitSystemStr) == 0) {
       return 1;
    } else {
       return 0;
    }
}

/*
  Notify Kinput2 to display ModeStr, BufStr, and SystemStr (if necessary)
*/
static void
displayPreEdit(FepObject obj)
{
    Widget w = (Widget)obj;
    FepObjectClass class = (FepObjectClass)XtClass(w);
    ICAuxControlArg         arg;

    XtCallCallbackList(w, obj->inputConv.textchangecallback, (XtPointer)NULL);
    if (wstrcmp(class->fep_class.PrevModeStr,
		class->fep_class.Result.ModeStr) != 0) {
	XtCallCallbackList(w, obj->inputConv.modechangecallback,
			   (XtPointer)NULL);
	wstrcpy(class->fep_class.PrevModeStr, class->fep_class.Result.ModeStr);
    }

    /* Display SystemStr if it differs from InitSystemStr */
    if (class->fep_class.ShowSystemStr == False) {
	if (class->fep_class.Result.LenSystemStr > 0
        && ! isInitSystemStr(&class->fep_class.Result)) {
	    class->fep_class.ShowSystemStr = True;
	    arg.command = ICAuxStart;
	    XtCallCallbackList(w,obj->inputConv.auxcallback, (XtPointer)&arg);
	} else {
	    /* nothing */
	}
    } else {
	if (class->fep_class.Result.LenSystemStr > 0
	&& ! isInitSystemStr(&class->fep_class.Result)) {
	    arg.command = ICAuxChange;
	    XtCallCallbackList(w, obj->inputConv.auxcallback, (XtPointer)&arg);
	} else {
	    class->fep_class.ShowSystemStr = False;
	    arg.command = ICAuxEnd;
	    XtCallCallbackList(w, obj->inputConv.auxcallback, (XtPointer)&arg);
	}
    }
}

/*
  initializeFep : allocate spaces for segments, and initialize other varialbes

  Note: this routine must be called after the first call of FepInfo(),
        AND, before the first call of makeSegment().
*/
static int
initializeFep(FepObject obj)
{
    FepObjectClass class = (FepObjectClass)XtClass((Widget)obj);

    /* init other variables */
    class->fep_class.ShowSystemStr = False;
    class->fep_class.NumModeSeg = 0;
    class->fep_class.NumSystemSeg = 0;
    class->fep_class.NumBufSeg = 0;

    /* alloc segment index arrays */
    /* In the worst case, # of segments == max length of string */
    class->fep_class.IndexModeSeg
	= (int *)XtMalloc((class->fep_class.Result.MaxModeStr+1)*sizeof(int));
    if (class->fep_class.IndexModeSeg == NULL)	return 0;

    class->fep_class.IndexSystemSeg
	= (int*)XtMalloc((class->fep_class.Result.MaxSystemStr+1)*sizeof(int));
    if (class->fep_class.IndexSystemSeg == NULL) return 0;

    class->fep_class.IndexBufSeg
	= (int *)XtMalloc((class->fep_class.Result.MaxBufStr+1) * sizeof(int));
    if (class->fep_class.IndexBufSeg == NULL) return 0;

    class->fep_class.PrevModeStr
	    = (wchar *)XtMalloc((class->fep_class.Result.MaxModeStr+1)
				* sizeof(wchar));
    if (class->fep_class.PrevModeStr == NULL)  return 0;
    class->fep_class.PrevModeStr[0] = 0;	/* null string */

    if (obj->fep.AlwaysShowSystemStr == True) {
	class->fep_class.Result.LenInitSystemStr = -1;
	/* Always LenInitSystemStr != LenSystemStr */
    }
    return 1;
}

/*
  terminateFep : clear allocated arrays
*/
static void
terminateFep(FepObject obj)
{
    FepObjectClass class = (FepObjectClass)XtClass((Widget)obj);

    /* init other variables */
    class->fep_class.ShowSystemStr = False;
    class->fep_class.NumModeSeg = 0;
    class->fep_class.NumSystemSeg = 0;
    class->fep_class.NumBufSeg = 0;

    /* alloc segment index arrays */
    XtFree((char *)class->fep_class.IndexModeSeg);
    class->fep_class.IndexModeSeg = NULL;
    XtFree((char *)class->fep_class.IndexSystemSeg);
    class->fep_class.IndexSystemSeg = NULL;
    XtFree((char *)class->fep_class.IndexBufSeg);
    class->fep_class.IndexBufSeg = NULL;
    XtFree((char *)class->fep_class.PrevModeStr);
    class->fep_class.PrevModeStr = NULL;
}

/*
  Parse conversion end keys
*/
static FepTriggerKey DefaultEndKey = {XK_space, ShiftMask, ShiftMask};

static void
getTriggerKeys(FepObject obj)
{
    char	*key_str;
    FepTriggerKey keys[100];
    int		num_keys;
    int		c;

    key_str = obj->fep.ConvEndKeysStr;
    num_keys = 0;

    if (key_str != NULL) {
	do {
	    char buf[256];
	    char *p = buf;
	    KeySym keysym;
	    long mods, chk_mods;

	    while ((c = *key_str++) != '\0' && c != '\n') {
		*p++ = c;
	    }
	    *p = '\0';
	    if (ParseKeyEvent(buf, &keysym, &mods, &chk_mods)) {
		keys[num_keys].keysym = keysym;
		keys[num_keys].modifiers = mods;
		keys[num_keys].check_modifiers = chk_mods;
		num_keys++;
	    }
	} while  (c != '\0' && num_keys < 100);
    }

    if (num_keys > 0) {
	const int	size = sizeof(FepTriggerKey) * num_keys;

	obj->fep.ConvEndKeys = (FepTriggerKey *)XtMalloc(size);
	bcopy((char *)keys, (char *)obj->fep.ConvEndKeys, size);
	obj->fep.NumConvEndKeys = num_keys;
    } else {
	obj->fep.ConvEndKeys = &DefaultEndKey;
	obj->fep.NumConvEndKeys = 1;
    }
}

/*
  makeSegment: make segments of ModeStr, SystemStr, and BufStr
  Note: this must be called every time after calling FepKKC(), and FepOn()
*/
static void
makeSegment(FepObject obj)
{
    FepObjectClass class = (FepObjectClass)XtClass((Widget)obj);
    FepResult *const r = &class->fep_class.Result;

    static int makeSegmentWstr(int len, const char *attr, int *index);

    class->fep_class.NumModeSeg
	= makeSegmentWstr(r->LenModeStr, r->ModeAttr,
			  class->fep_class.IndexModeSeg);
    class->fep_class.NumSystemSeg
	= makeSegmentWstr(r->LenSystemStr, r->SystemAttr,
			  class->fep_class.IndexSystemSeg);
    class->fep_class.NumBufSeg
	= makeSegmentWstr(r->LenBufStr, r->BufAttr,
			  class->fep_class.IndexBufSeg);
}

/*
  makeSegmentWstr: According to Attr, divide Wstr into segments
                   Sub-routine of makeSegment()
*/
static int
makeSegmentWstr(int len, const char *attr, int *index)
{
    int		i = 0, numSeg = 0;
    int		prevAttr = -1;

    if (len == 0)  return 0;
    while (1) {
	if (i == len) {
	    *index = i;
	    break;
	}
	if (*attr != prevAttr) {
	    *index = i;
	    prevAttr = *attr;
	    numSeg ++, index ++;
	}
	attr ++, i ++;
    }
    return numSeg;
}
