;;;
;;; Fep Bridge 1.34 (1996/12/15)
;;; Copyright (C) Yutaka Tamiya (tamy@flab.fujitsu.co.jp) 
;;;
;;;    I thanks very much to these authors for stealing some codes.
;;;        S. Tomura (tomura@etl.go.jp)		: wnn-client.el
;;;        Akira Kon (kon@dl.bs2.mt.nec.co.jp)	: canna.el
;;;        Masahiko Sato (masahiko@sato.riec.tohoku.ac.jp): skk.el	
;;;

;;;
;;; Variables for customization
;;; Please do "setq" the following variables before loading this file.
;;;
(defvar fep:*conversion-key-list*	'("\C-\\" [kanji])
					"Start/End key list for Fep Bridge")
(defvar fep:*unmap-key-list*	nil	"Key list not sent to Fep Bridge")
(defvar fep:*host*	"localhost"	"Fep Bridge host name")
(defvar fep:*service*	6622		"Fep Bridge service port")
(defvar fep:*buf-attr-mask*	240	"Bit mask for FEP's buffer string")
(defvar fep:*system-attr-mask*	240	"Bit mask for FEP's system string")
(defvar fep:*full-mode-line-in-minibuffer* nil
				"Show the whole mode line in minibuffer")

;;;
;;; Variables for Emacs
;;;
(defvar fep:*current-buffer*	nil	"Current buffer for conversion")
(defvar fep:*display-buffer*	nil	"Current display buffer")
(defvar fep:*japanese-mode*	nil	"Fep Bridge japanese mode.")
(defvar fep:*preedit-start* (make-marker)
					"Start point of the preedit region")
(defvar fep:*preedit-end* (make-marker) "End point of the preedit region")

(defvar fep:*attr-underline*	nil	"Underline attr in the preedit region")
(defvar fep:*buf-seg-list*	nil	"Segments of the preedit region")
(defvar fep:*system-seg-list*	nil	"Segments of system region")
(copy-face 'region 'fep:*region*)	; color definitions for "region"
(defvar fep:*buffer*  nil	"Buffer associated with Fep Bridge process.")
(defvar fep:*process* nil	"Fep Bridge process.")
(defvar fep:*command* nil	"Command buffer to be sent to Fep Bridge.")
(defvar fep:*mode-line-format*	nil	"Mode line format for Fep Bridge.")
(defvar fep:*mode-line-format-backup*	nil	"Backup of mode line format.")
(defvar fep:*local-map-backup*	nil	"Backup of keymap")
(defvar fep:*minibuffer-map-backup*	nil	"Backup of minibuffer keymap")
(defvar fep:*mode-map* (make-keymap)	"Keymap for Fep Bridge")
(defvar fep:*menu-buffer*	nil	"Buffer to display fep:*system-str*")
(defvar fep:*saved-minibuffer*	nil	"Saved buffer displayed in minibuffer")
(defvar fep:*backup-undo-list*	nil	"Saved undo-list of the buffer")
(defvar fep:*isearch-japanese-mode*	nil	"japanese mode in isearch")

(make-variable-buffer-local 'fep:*japanese-mode*)
(make-variable-buffer-local 'fep:*display-buffer*)
(make-variable-buffer-local 'fep:*local-map-backup*)
(make-variable-buffer-local 'fep:*preedit-start*)
(make-variable-buffer-local 'fep:*preedit-end*)
(make-variable-buffer-local 'fep:*local-map-backup*)
(make-variable-buffer-local 'fep:*backup-undo-list*)

;;;
;;; Constants & variables for communication with Fep Beridg
;;;
; Commands of Protocol
(defconst fep:FO_Info	1 "Information of Fep Bridge")
(defconst fep:FO_Ack	2 "Acknoledgement to connect with Fep Bridge")
(defconst fep:FO_Result	5 "Result of japanese conversion")
(defconst fep:FO_Alive	3 "Fep Bridge is alive")
(defconst fep:FO_Dead	4 "Fep Bridge is dead")
(defconst fep:FI_Info	1 "Send me information of Fep Bridge")
(defconst fep:FI_Reset	2 "Reset Fep Bridg")
(defconst fep:FI_Auth	3 "User authorization check")
(defconst fep:FI_Conv1	4 "Request for japanese conversion")
(defconst fep:FI_ConvN	6 "N-key request for japanese conversion")
(defconst fep:FI_Stat	5 "Is Fep Bridge alive ?")
(defconst fep:FI_Kill	7 "Kill Fep Bridge")
(defconst fep:FI_UserList	8 "Send me user list")
(defconst fep:FI_Nop	127 "A meaningless packet")

; Virtual key codes of MS-Windows
(defconst fep:VK_SHIFT_MASK	1   "Shift modifier")
(defconst fep:VK_CONTROL_MASK	2   "Control modifier")
(defconst fep:VK_ALT_MASK	4   "Alt modifier")

(defconst fep:VK_LBUTTON	1   "Left mouse button")
(defconst fep:VK_RBUTTON	2   "Right mouse button")
(defconst fep:VK_CANCEL		3   "Used for control-break processing")
(defconst fep:VK_MBUTTON	4   "Middle mouse button (three-button mouse)")
(defconst fep:VK_BACK		8   "BACKSPACE key")
(defconst fep:VK_TAB		9   "TAB key")
(defconst fep:VK_CLEAR		12  "CLEAR key")
(defconst fep:VK_RETURN		13  "ENTER key")
(defconst fep:VK_SHIFT		16  "SHIFT key")
(defconst fep:VK_CONTROL	17  "CTRL key")
(defconst fep:VK_MENU		18  "ALT key")
(defconst fep:VK_PAUSE		19  "PAUSE key")
(defconst fep:VK_CAPITAL	20  "CAPS LOCK key")
(defconst fep:VK_KANA		21  "Kakana/Hiragana key")
(defconst fep:VK_KANJI		25  "Kanji key")
(defconst fep:VK_ESCAPE		27  "ESC key")
(defconst fep:VK_CONVERT	28  "Henkan key")
(defconst fep:VK_NONCONVERT	29  "MuHenkan key")
(defconst fep:VK_SPACE		32  "SPACEBAR")
(defconst fep:VK_PRIOR		33  "PAGE UP key")
(defconst fep:VK_NEXT		34  "PAGE DOWN key")
(defconst fep:VK_END		35  "END key")
(defconst fep:VK_HOME		36  "HOME key")
(defconst fep:VK_LEFT		37  "LEFT ARROW key")
(defconst fep:VK_UP		38  "UP ARROW key")
(defconst fep:VK_RIGHT		39  "RIGHT ARROW key")
(defconst fep:VK_DOWN		40  "DOWN ARROW key")
(defconst fep:VK_SELECT		41  "SELECT key")
(defconst fep:VK_EXECUTE	43  "EXECUTE key")
(defconst fep:VK_SNAPSHOT	44  "PRINT SCREEN key")
(defconst fep:VK_INSERT		45  "INS key")
(defconst fep:VK_DELETE		46  "DEL key")
(defconst fep:VK_HELP		47  "HELP key")
(defconst fep:VK_0		48  "0 key")
(defconst fep:VK_1		49  "1 key")
(defconst fep:VK_2		50  "2 key")
(defconst fep:VK_3		51  "3 key")
(defconst fep:VK_4		52  "4 key")
(defconst fep:VK_5		53  "5 key")
(defconst fep:VK_6		54  "6 key")
(defconst fep:VK_7		55  "7 key")
(defconst fep:VK_8		56  "8 key")
(defconst fep:VK_9		57  "9 key")
(defconst fep:VK_A		65  "A key")
(defconst fep:VK_B		66  "B key")
(defconst fep:VK_C		67  "C key")
(defconst fep:VK_D		68  "D key")
(defconst fep:VK_E		69  "E key")
(defconst fep:VK_F		70  "F key")
(defconst fep:VK_G		71  "G key")
(defconst fep:VK_H		72  "H key")
(defconst fep:VK_I		73  "I key")
(defconst fep:VK_J		74  "J key")
(defconst fep:VK_K		75  "K key")
(defconst fep:VK_L		76  "L key")
(defconst fep:VK_M		77  "M key")
(defconst fep:VK_N		78  "N key")
(defconst fep:VK_O		79  "O key")
(defconst fep:VK_P		80  "P key")
(defconst fep:VK_Q		81  "Q key")
(defconst fep:VK_R		82  "R key")
(defconst fep:VK_S		83  "S key")
(defconst fep:VK_T		84  "T key")
(defconst fep:VK_U		85  "U key")
(defconst fep:VK_V		86  "V key")
(defconst fep:VK_W		87  "W key")
(defconst fep:VK_X		88  "X key")
(defconst fep:VK_Y		89  "Y key")
(defconst fep:VK_Z		90  "Z key")
(defconst fep:VK_NUMPAD0	96  "Numeric keypad 0 key")
(defconst fep:VK_NUMPAD1	97  "Numeric keypad 1 key")
(defconst fep:VK_NUMPAD2	98  "Numeric keypad 2 key")
(defconst fep:VK_NUMPAD3	99  "Numeric keypad 3 key")
(defconst fep:VK_NUMPAD4	100 "Numeric keypad 4 key")
(defconst fep:VK_NUMPAD5	101 "Numeric keypad 5 key")
(defconst fep:VK_NUMPAD6	102 "Numeric keypad 6 key")
(defconst fep:VK_NUMPAD7	103 "Numeric keypad 7 key")
(defconst fep:VK_NUMPAD8	104 "Numeric keypad 8 key")
(defconst fep:VK_NUMPAD9	105 "Numeric keypad 9 key")
(defconst fep:VK_MULTIPLY	106 "Multiply key")
(defconst fep:VK_ADD		107 "Add key")
(defconst fep:VK_SEPARATOR	108 "Separator key")
(defconst fep:VK_SUBTRACT	109 "Subtract key")
(defconst fep:VK_DECIMAL	110 "Decimal key")
(defconst fep:VK_DIVIDE		111 "Divide key")
(defconst fep:VK_F1		112 "F1 key")
(defconst fep:VK_F2		113 "F2 key")
(defconst fep:VK_F3		114 "F3 key")
(defconst fep:VK_F4		115 "F4 key")
(defconst fep:VK_F5		116 "F5 key")
(defconst fep:VK_F6		117 "F6 key")
(defconst fep:VK_F7		118 "F7 key")
(defconst fep:VK_F8		119 "F8 key")
(defconst fep:VK_F9		120 "F9 key")
(defconst fep:VK_F10		121 "F10 key")
(defconst fep:VK_F11		122 "F11 key")
(defconst fep:VK_F12		123 "F12 key")
(defconst fep:VK_F13		124 "F13 key")
(defconst fep:VK_F14		125 "F14 key")
(defconst fep:VK_F15		126 "F15 key")
(defconst fep:VK_F16		127 "F16 key")
(defconst fep:VK_F17		128 "F17 key")
(defconst fep:VK_F18		129 "F18 key")
(defconst fep:VK_F19		130 "F19 key")
(defconst fep:VK_F20		131 "F20 key")
(defconst fep:VK_F21		132 "F21 key")
(defconst fep:VK_F22		133 "F22 key")
(defconst fep:VK_F23		134 "F23 key")
(defconst fep:VK_F24		135 "F24 key")
(defconst fep:VK_NUMLOCK	144 "NUM LOCK key")
(defconst fep:VK_SCROLL		145 "SCROLL LOCK key")
; for US 101 keyboard
(defconst fep:VK_US_SEMICOLON	186 "; key")
(defconst fep:VK_US_EQUAL	187 "= key")
(defconst fep:VK_US_COMMA	188 ", key")
(defconst fep:VK_US_MINUS	189 "- key")
(defconst fep:VK_US_PERIOD	190 ". key")
(defconst fep:VK_US_SLASH	191 "/ key")
(defconst fep:VK_US_GRAVE	192 "` key")
(defconst fep:VK_US_BRACELEFT	219 "[ key")
(defconst fep:VK_US_BACKSLASH	220 "\ key")
(defconst fep:VK_US_BRACERIGHT	221 "] key")
(defconst fep:VK_US_APOSTROPHE	222 "' key")

; Variables for storing conversion results
(defvar fep:*packet-id* 	0   "Packet ID from Fep Bridge.")
(defvar fep:*return-code*	0   "Result value of Fep Bridge.")
(defvar fep:*result-len* 	0   "Length of FEP result string.")
(defvar fep:*result-str*	nil "FEP result string.")
(defvar fep:*result-ctl-h-num* 	0   "# of ^H in the result string.")
(defvar fep:*mode-len*		0   "Length of FEP mode string.")
(defvar fep:*mode-str*		nil "FEP mode string.")
;(defvar fep:*mode-attr*	nil "Attributes of FEP mode string.")
(defvar fep:*system-len*	0   "Length of FEP system string.")
(defvar fep:*system-str*	nil "FEP system string.")
(defvar fep:*system-attr*	nil "Attributes of FEP system string.")
(defvar fep:*init-system-str*	nil "Initial FEP system string.")
(defvar fep:*buf-len*		0   "Length of FEP buffer string.")
(defvar fep:*buf-str*		nil "FEP buffer string.")
(defvar fep:*buf-attr*		nil "Attributes of FEP buffer string.")

;;;
;;; Turn on/off Fep Bridge
;;;
(defun fep:toggle-japanese-mode ()
  "Turn on/off Japanese mode"
  (interactive "")
  (if (not buffer-read-only)
      (if (null fep:*japanese-mode*)
	  (fep:japanese-mode-on)
	(fep:japanese-mode-off))
    (message "Buffer is read-only: #<buffer %s>" (current-buffer)))
  )

(defun fep:japanese-mode-on ()
  "Trun on Japanese mode"
  (setq fep:*current-buffer* (current-buffer))
  (if (not fep:*isearch-japanese-mode*)
      (progn
	(setq fep:*japanese-mode* t)
	(setq fep:*local-map-backup* (current-local-map))
	(setq fep:*mode-line-format-backup* mode-line-format)
	(setq mode-line-format fep:*mode-line-format*)
	(setq fep:*display-buffer* fep:*current-buffer*)
	(fep:record-undo-list nil))
    (setq fep:*minibuffer-map-backup* (current-local-map)))

  (set-marker fep:*preedit-start* (point))
  (set-marker fep:*preedit-end* (point))
  (use-local-map fep:*mode-map*)
  (fep:reset)
  (fep:display-preedit)
  )

(defun fep:japanese-mode-off ()
  "Turn off Japanese mode"
  (if (not fep:*isearch-japanese-mode*)
      (progn
	(setq fep:*japanese-mode* nil)
	(use-local-map fep:*local-map-backup*)
	(setq mode-line-format fep:*mode-line-format-backup*)
	(fep:record-undo-list t))
    (use-local-map fep:*minibuffer-map-backup*))

  (setq minibuffer-preprompt nil)
  (fep:delete-preedit)
  (if fep:*saved-minibuffer*
      ;display the saved-buffer in minibuffer
      (progn
	(set-window-buffer (minibuffer-window) fep:*saved-minibuffer*)
	(setq fep:*saved-minibuffer nil)
	))
  (set-buffer-modified-p (buffer-modified-p))
  )

;;;
;;;  Keyboard input for Fep Bridge
;;;
(defun fep:key-normal ()
  "Key command for normal keys"
  (interactive "*")
  (fep:conv (fep:key-normal-sub last-command-event))
  )

(defun fep:key-normal-sub (ch)
  "Generate list of (virtual-keycode ascii-code shift-status)"
  (cond
   ((or (< ch 32) (= ch 127)) (list 0 ch '(control)))	; Ctrl-char
   (t (list 0 ch nil)))
  )

(defun fep:key-unmap ()
  "Key command for unmap keys"
  (interactive "*")
  (if (not fep:*isearch-japanese-mode*)
      (fep:original-key-command (this-command-keys)))
  )


;;;
;;; Key command for special keys
;;;
(defun fep:key-back ()
  (interactive "*")
  (fep:conv (list fep:VK_BACK 0 (event-modifiers last-command-event))))

(defun fep:key-tab ()
  (interactive "*")
  (fep:conv (list fep:VK_TAB 0 (event-modifiers last-command-event))))

(defun fep:key-clear ()
  (interactive "*")
  (fep:conv (list fep:VK_CLEAR 0 (event-modifiers last-command-event))))

(defun fep:key-return ()
  (interactive "*")
  (fep:conv (list fep:VK_RETURN 0 (event-modifiers last-command-event))))

(defun fep:key-pause ()
  (interactive "*")
  (fep:conv (list fep:VK_PAUSE 0 (event-modifiers last-command-event))))

(defun fep:key-capital ()
  (interactive "*")
  (fep:conv (list fep:VK_CAPITAL 0 (event-modifiers last-command-event))))

(defun fep:key-escape ()
  (interactive "*")
  (fep:conv (list fep:VK_ESCAPE 0 (event-modifiers last-command-event))))

(defun fep:key-prior ()
  (interactive "*")
  (fep:conv (list fep:VK_PRIOR 0 (event-modifiers last-command-event))))

(defun fep:key-next ()
  (interactive "*")
  (fep:conv (list fep:VK_NEXT 0 (event-modifiers last-command-event))))

(defun fep:key-end ()
  (interactive "*")
  (fep:conv (list fep:VK_END 0 (event-modifiers last-command-event))))

(defun fep:key-home ()
  (interactive "*")
  (fep:conv (list fep:VK_HOME 0 (event-modifiers last-command-event))))

(defun fep:key-left ()
  (interactive "*")
  (fep:conv (list fep:VK_LEFT 0 (event-modifiers last-command-event))))

(defun fep:key-up ()
  (interactive "*")
  (fep:conv (list fep:VK_UP 0 (event-modifiers last-command-event))))

(defun fep:key-right ()
  (interactive "*")
  (fep:conv (list fep:VK_RIGHT 0 (event-modifiers last-command-event))))

(defun fep:key-down ()
  (interactive "*")
  (fep:conv (list fep:VK_DOWN 0 (event-modifiers last-command-event))))

(defun fep:key-select ()
  (interactive "*")
  (fep:conv (list fep:VK_SELECT 0 (event-modifiers last-command-event))))

(defun fep:key-execute ()
  (interactive "*")
  (fep:conv (list fep:VK_EXECUTE 0 (event-modifiers last-command-event))))

(defun fep:key-snapshot ()
  (interactive "*")
  (fep:conv (list fep:VK_SNAPSHOT 0 (event-modifiers last-command-event))))

(defun fep:key-insert ()
  (interactive "*")
  (fep:conv (list fep:VK_INSERT 0 (event-modifiers last-command-event))))

(defun fep:key-delete ()
  (interactive "*")
  (fep:conv (list fep:VK_DELETE 0 (event-modifiers last-command-event))))

(defun fep:key-help ()
  (interactive "*")
  (fep:conv (list fep:VK_HELP 0 (event-modifiers last-command-event))))

(defun fep:key-numpad0 ()
  (interactive "*")
  (fep:conv (list fep:VK_NUMPAD0 0 (event-modifiers last-command-event))))

(defun fep:key-numpad1 ()
  (interactive "*")
  (fep:conv (list fep:VK_NUMPAD1 0 (event-modifiers last-command-event))))

(defun fep:key-numpad2 ()
  (interactive "*")
  (fep:conv (list fep:VK_NUMPAD2 0 (event-modifiers last-command-event))))

(defun fep:key-numpad3 ()
  (interactive "*")
  (fep:conv (list fep:VK_NUMPAD3 0 (event-modifiers last-command-event))))

(defun fep:key-numpad4 ()
  (interactive "*")
  (fep:conv (list fep:VK_NUMPAD4 0 (event-modifiers last-command-event))))

(defun fep:key-numpad5 ()
  (interactive "*")
  (fep:conv (list fep:VK_NUMPAD5 0 (event-modifiers last-command-event))))

(defun fep:key-numpad6 ()
  (interactive "*")
  (fep:conv (list fep:VK_NAMPAD6 0 (event-modifiers last-command-event))))

(defun fep:key-numpad7 ()
  (interactive "*")
  (fep:conv (list fep:VK_NUMPAD7 0 (event-modifiers last-command-event))))

(defun fep:key-numpad8 ()
  (interactive "*")
  (fep:conv (list fep:VK_NUMPAD8 0 (event-modifiers last-command-event))))

(defun fep:key-numpad9 ()
  (interactive "*")
  (fep:conv (list fep:VK_NUMPAD9 0 (event-modifiers last-command-event))))

(defun fep:key-multiply ()
  (interactive "*")
  (fep:conv (ist fep:VK_MULTIPLY 0 (event-modifiers last-command-event))))

(defun fep:key-add ()
  (interactive "*")
  (fep:conv (list fep:VK_ADD 0 (event-modifiers last-command-event))))

(defun fep:key-separator ()
  (interactive "*")
  (fep:conv (list fep:VK_SEPARATOR 0 (event-modifiers last-command-event))))

(defun fep:key-subtract ()
  (interactive "*")
  (fep:conv (list fep:VK_SUBTRACT 0 (event-modifiers last-command-event))))

(defun fep:key-decimal ()
  (interactive "*")
  (fep:conv (list fep:VK_DECIMAL 0 (event-modifiers last-command-event))))

(defun fep:key-divide ()
  (interactive "*")
  (fep:conv (list fep:VK_DIVIDE 0 (event-modifiers last-command-event))))

(defun fep:key-f1 ()
  (interactive "*")
  (fep:conv (list fep:VK_F1 0 (event-modifiers last-command-event))))

(defun fep:key-f2 ()
  (interactive "*")
  (fep:conv (list fep:VK_F2 0 (event-modifiers last-command-event))))

(defun fep:key-f3 ()
  (interactive "*")
  (fep:conv (list fep:VK_F3 0 (event-modifiers last-command-event))))

(defun fep:key-f4 ()
  (interactive "*")
  (fep:conv (list fep:VK_F4 0 (event-modifiers last-command-event))))

(defun fep:key-f5 ()
  (interactive "*")
  (fep:conv (list fep:VK_F5 0 (event-modifiers last-command-event))))

(defun fep:key-f6 ()
  (interactive "*")
  (fep:conv (list fep:VK_F6 0 (event-modifiers last-command-event))))

(defun fep:key-f7 ()
  (interactive "*")
  (fep:conv (list fep:VK_F7 0 (event-modifiers last-command-event))))

(defun fep:key-f8 ()
  (interactive "*")
  (fep:conv (list fep:VK_F8 0 (event-modifiers last-command-event))))

(defun fep:key-f9 ()
  (interactive "*")
  (fep:conv (list fep:VK_F9 0 (event-modifiers last-command-event))))

(defun fep:key-f10 ()
  (interactive "*")
  (fep:conv (list fep:VK_F10 0 (event-modifiers last-command-event))))

(defun fep:key-f11 ()
  (interactive "*")
  (fep:conv (list fep:VK_F11 0 (event-modifiers last-command-event))))

(defun fep:key-f12 ()
  (interactive "*")
  (fep:conv (list fep:VK_F12 0 (event-modifiers last-command-event))))

(defun fep:key-f13 ()
  (interactive "*")
  (fep:conv (list fep:VK_F13 0 (event-modifiers last-command-event))))

(defun fep:key-f14 ()
  (interactive "*")
  (fep:conv (list fep:VK_F14 0 (event-modifiers last-command-event))))

(defun fep:key-f15 ()
  (interactive "*")
  (fep:conv (list fep:VK_F15 0 (event-modifiers last-command-event))))

(defun fep:key-f16 ()
  (interactive "*")
  (fep:conv (list fep:VK_F16 0 (event-modifiers last-command-event))))

(defun fep:key-f17 ()
  (interactive "*")
  (fep:conv (list fep:VK_F17 0 (event-modifiers last-command-event))))

(defun fep:key-f18 ()
  (interactive "*")
  (fep:conv (list fep:VK_F18 0 (event-modifiers last-command-event))))

(defun fep:key-f19 ()
  (interactive "*")
  (fep:conv (list fep:VK_F19 0 (event-modifiers last-command-event))))

(defun fep:key-f20 ()
  (interactive "*")
  (fep:conv (list fep:VK_F20 0 (event-modifiers last-command-event))))

(defun fep:key-f21 ()
  (interactive "*")
  (fep:conv (list fep:VK_F21 0 (event-modifiers last-command-event))))

(defun fep:key-f22 ()
  (interactive "*")
  (fep:conv (list fep:VK_F22 0 (event-modifiers last-command-event))))

(defun fep:key-f23 ()
  (interactive "*")
  (fep:conv (list fep:VK_F23 0 (event-modifiers last-command-event))))

(defun fep:key-f24 ()
  (interactive "*")
  (fep:conv (list fep:VK_F24 0 (event-modifiers last-command-event))))

;;;
;;;  Display strings of FEP
;;;
(defun fep:display-preedit ()
  "Insert the fixed string (if any). Display the preedit string,
   and system string (if any)" 
  (set-buffer fep:*display-buffer*)
  (fep:delete-preedit)
  (if (> fep:*result-len* 0)
      (progn
	(fep:record-undo-list t)
	(set-buffer-modified-p (buffer-modified-p))
	; When FEP retuns some ^H's in its result string, delete chars
	; of the same number of ^H's.
	(while (> fep:*result-ctl-h-num* 0)
	  (progn
	    (backward-char)
	    ; Delete \n's automatically inserted in auto-fill mode.
	    (if (not (eq (char-after (point)) ?\n))
		(setq fep:*result-ctl-h-num* (1- fep:*result-ctl-h-num*)))
	    (delete-char 1)))
	(insert fep:*result-str*)
	(setq fep:*result-len* 0)
	(if self-insert-after-hook
	    (funcall self-insert-after-hook
		     fep:*preedit-start* fep:*preedit-end*))
	(fep:do-auto-fill)
	(undo-boundary)
	(fep:record-undo-list nil)
	))
  (set-marker fep:*preedit-start* (point))
  (if (> fep:*buf-len* 0)
      (progn
	(insert fep:*buf-str*)
	(setq fep:*buf-seg-list*
	      (fep:make-segments
	       fep:*buf-attr* fep:*buf-attr-mask* fep:*preedit-start*))
	(fep:attribute-on fep:*buf-seg-list* 'fep:*region*)
	))
  (set-marker fep:*preedit-end* (point))
  (if (> fep:*buf-len* 0)
      (fep:attribute-underline-on fep:*preedit-start* fep:*preedit-end*))
  ;(goto-char (+ fep:*preedit-start* fep:*cursor-pos*))
  (fep:display-minibuffer)
  (set-buffer fep:*current-buffer*)
  )

(defun fep:display-minibuffer ()
  "Display the system str if it differs from the initial system str."
  (if (equal fep:*init-system-str* fep:*system-str*)
      (progn
	(if fep:*saved-minibuffer*
	    ;display the saved-buffer in minibuffer
	    (progn
	      (set-window-buffer (minibuffer-window) fep:*saved-minibuffer*)
	      (setq fep:*saved-minibuffer nil)
	      ))
	(if (eq fep:*display-buffer* (window-buffer (minibuffer-window)))
	    (if (boundp 'minibuffer-preprompt)
		(if fep:*full-mode-line-in-minibuffer*
		    (setq minibuffer-preprompt (concat "[" fep:*mode-str* "]"))
		  (setq minibuffer-preprompt (concat "[]"))
		  )))
	)
    ;else: the system str is not the initial system str.
    (progn
      (if (not fep:*saved-minibuffer*)
	  (setq fep:*saved-minibuffer* (window-buffer (minibuffer-window)))
	)
      (setq minibuffer-preprompt nil)
      (set-window-buffer (minibuffer-window) fep:*menu-buffer*)
      (set-buffer fep:*menu-buffer*)
      (fep:attribute-off fep:*system-seg-list*)
      (delete-region (point-min) (point-max))
      (insert fep:*system-str*)
      (setq fep:*system-seg-list*
	    (fep:make-segments fep:*system-attr*
			       fep:*system-attr-mask* (point-min)))
      (fep:attribute-on fep:*system-seg-list* 'fep:*region*)
      (if fep:*system-seg-list*
	  (goto-char (overlay-end (car fep:*system-seg-list*))))
      ))
  )

(defun fep:delete-preedit ()
  (fep:attribute-underline-off fep:*preedit-start* fep:*preedit-end*)
  (if fep:*buf-seg-list*
      (progn
	(fep:attribute-off fep:*buf-seg-list*)
	(setq fep:*buf-seg-list* nil)
	))
  (save-excursion
    (progn
      (set-buffer (marker-buffer fep:*preedit-start*))
      (delete-region fep:*preedit-start* fep:*preedit-end*)
      ))
  )

(defun fep:make-segments (attr-str mask buf-top)
  "Make list of attribute segments from attr-str."
  (let ((rtn nil) (i 0) (s 0) (len (length attr-str)))
    (while (< i len)
      (while (and (< i len)
		  (zerop (logand (aref attr-str i) mask)))
	(setq i (1+ i)))
      (if (< i len)
	  (progn
	    (setq s i)
	    (while
		(and (< i len)
		     (not (zerop (logand (aref attr-str i) mask))))
	      (setq i (1+ i)))
	    (setq rtn
		  (cons (make-overlay (+ buf-top s) (+ buf-top i) nil t) rtn))
	    )))
    rtn
    ))

(defun fep:attribute-underline-on (start end)
  (if (overlayp fep:*attr-underline*)
      (move-overlay fep:*attr-underline* start end)
    (overlay-put (setq fep:*attr-underline* (make-overlay start end nil t))
		 'face 'underline)
    ))

(defun fep:attribute-underline-off (start end)
  (and (overlayp fep:*attr-underline*)
       (delete-overlay fep:*attr-underline*)))

(defun fep:attribute-on (seg-list attr)
  (let ((segs seg-list))
    (while segs
      (progn
	(overlay-put (car segs) 'face attr)
	(setq segs (cdr segs)))))
  )

(defun fep:attribute-off (seg-list)
  (let ((segs seg-list))
    (while segs
      (progn
	(and (overlayp (car segs))
	     (delete-overlay (car segs)))
	(setq segs (cdr segs))))
    ))

;;;
;;; auto fill controll (from egg)
;;;
(defun fep:do-auto-fill ()
  (if (and auto-fill-function (not buffer-read-only)
	   (> (current-column) fill-column))
      (let ((ocolumn (current-column)))
	(funcall auto-fill-function)
	(while (and (< fill-column (current-column))
		    (< (current-column) ocolumn))
  	  (setq ocolumn (current-column))
	  (funcall auto-fill-function)))))

;;;
;;; Switch to stop/re-start recording undo list
;;;
(defun fep:record-undo-list (on)
  "Stop/Re-start recording undo list"
  (let ((saved-buffer (current-buffer)))
    (if on
	(progn
	  (set-buffer fep:*current-buffer*)
	  (buffer-enable-undo)
	  (setq buffer-undo-list fep:*backup-undo-list*))
      (progn
	(set-buffer fep:*current-buffer*)
	(setq fep:*backup-undo-list* buffer-undo-list)
	(buffer-disable-undo)))
    (set-buffer saved-buffer))
  )

;;;
;;;
;;; Fep Bridge Interface:  open, reset, info, conv
;;;
(defun fep:open1 ()
  "Open the connection to FEP."
  (let (host)
    (setq host (getenv "FEPHOST"))
    (if (null host) (setq host fep:*host*))

    (if (or (null fep:*process*) (eq (process-status fep:*process*) 'closed))
	; First try with service name of "fepbrdg".
	(progn
	  (setq fep:*process*
		(condition-case var
		    (open-network-stream "*fepbrdg*"
					 " [Fep Bridge Result Buffer] "
					 host "fepbrdg")
		  (error nil)))
	  (if (null fep:*process*)
	      ; Second try with service port address.
	      (setq fep:*process*
		    (condition-case var
			(open-network-stream "*fepbrdg*"
					     " [Fep Bridge Result Buffer] "
					     host fep:*service*)
		      (error nil))))))
    (if (null fep:*process*) nil
      (progn
	(setq fep:*command*
	      (get-buffer-create " [Fep Bridge Command Buffer] "))
	(set-buffer fep:*command*)
	(setq kanji-flag nil)
	(setq mc-flag nil)
	(buffer-disable-undo fep:*command*)
	(setq fep:*buffer* (process-buffer fep:*process*))
	(set-buffer fep:*buffer*)
	(set-current-process-coding-system *noconv* *noconv*)
	(setq kanji-flag nil)
	(setq mc-flag nil)
	(buffer-disable-undo fep:*buffer*)
	(process-kill-without-query fep:*process*)
	(condition-case nil
	    (if (fep:authority)
		(progn
		 (fep:info)	; get the initial system str
		 fep:*process*)
	      (setq fep:*process* nil))
	  (error (setq fep:*process* nil)))
	)))
  )

(defun fep:authority ()
  "Send the user name to Fep Bridge."
  (let ((name (user-real-login-name)))
    (fep:command-start fep:FI_Auth)
    (fep:put-int2 (length name))
    (insert name)
    (fep:command-end)
    (set-buffer fep:*buffer*)
    (and (eq (fep:get-1byte) fep:FO_Ack) (eq (fep:get-1byte) 0)))
  )
  
(defun fep:info ()
  "Get information of Fep Bridge."
  (fep:command-start fep:FI_Info)
  (fep:command-end)
  (fep:get-info)
  )

(defun fep:open ()
  "Open connection to Fep Bridge, if it is not opened yet, or closed."
  (save-excursion
  (let ((i 0))
    (if (not (and fep:*process*
		  (eq (process-status fep:*process*) 'open)))
	(progn
	  (while (not (fep:open1))
	    (sit-for 1)
	    (setq i (1+ i))
	    (if (= i 5)
		(error "Fep Bridge may be dead."))
	    ))
      ))
  )
  fep:*process*)

(defun fep:reset ()
  "Reset Fep Bridge."
  (fep:open)
  (progn
    (fep:command-start fep:FI_Reset)
    (fep:command-end)
    (fep:get-result))
  (set-buffer fep:*current-buffer*)
  )
  
(defun fep:conv (kcode)
  "Send key input to FEP and show result of conversion. If FEP does not
   accept the key input, execute Emacs command assigned to the key input."
  (fep:open)
  (fep:command-start fep:FI_Conv1)
  (fep:put-1byte (car kcode))
  (fep:put-1byte (car (cdr kcode)))
  (fep:put-1byte (fep:convert-modifiers (car (cdr (cdr kcode)))))
  (fep:command-end)
  (fep:get-result)
  (set-buffer fep:*current-buffer*)
  (if (eq fep:*return-code* 0)
      (fep:display-preedit)
    ;else : FEP did not accept the key. Execute the Emacs command.
    (if (not fep:*isearch-japanese-mode*)
	(fep:original-key-command (this-command-keys))
      ))
  )

(defun fep:convert-modifiers (mod-list)
  (let ((c 0))
    (while mod-list
       (if (eq 'shift (car mod-list)) (setq c (+ c fep:VK_SHIFT_MASK)))
       (if (eq 'control (car mod-list)) (setq c (+ c fep:VK_CONTROL_MASK)))
       (if (or (eq 'meta (car mod-list)) (eq 'alt (car mod-list)))
	   (setq c (+ c fep:VK_ALT_MASK)))
       (setq mod-list (cdr mod-list)))
    c
    ))

(defun fep:original-key-command (com-keys)
  "Do original Emacs key command associated with com-keys"
  (let (com (saved-current-buffer fep:*current-buffer*))
    (setq unread-command-events (listify-key-sequence com-keys))
    (use-local-map fep:*local-map-backup*)
    (fep:record-undo-list t)
    (setq com (key-binding (read-key-sequence nil)))
    (if com
	(progn
	  (call-interactively com)
	  (save-excursion
	    (set-buffer saved-current-buffer)
	    (fep:record-undo-list nil)
	    (use-local-map fep:*mode-map*)
	    ))
      (fep:record-undo-list nil)
      (use-local-map fep:*mode-map*))
    ))

;;;
;;; Fep Bridge Command Primitives
;;;
(defun fep:command-start (command)
  (set-buffer fep:*command*)
  (erase-buffer)
  (fep:put-1byte command)
  (point)
  )

(defun fep:command-end ()
  (set-buffer fep:*buffer*)
  (erase-buffer)
  (set-buffer fep:*command*)
  (process-send-region fep:*process* (point-min) (point))
  )

;;;
;;; Get conversion result
;;;
(defun fep:get-result ()
  (let ((str nil) (attr nil))
    (set-buffer fep:*buffer*)
    (prog2
     (setq fep:*packet-id*	(fep:get-1byte))
     (setq fep:*return-code*	(fep:get-1byte))
     (if (eq fep:*return-code* 0)
	 (progn
	   (setq fep:*result-len*	(fep:get-int2))
	   (setq fep:*mode-len*		(fep:get-int2))
	   (setq fep:*system-len*	(fep:get-int2))
	   (setq fep:*buf-len*		(fep:get-int2))
	   (setq fep:*cursor-pos*	(fep:get-int2))

	   (setq str (fep:get-nbyte fep:*result-len*))
	   (setq fep:*result-ctl-h-num* (fep:count-ctl-h str))
	   (setq fep:*result-str*
		 (code-convert-string
		  (substring str fep:*result-ctl-h-num*) *sjis* *internal*))

	   (setq str	(fep:get-nbyte fep:*mode-len*))
	   (setq attr	(fep:get-nbyte fep:*mode-len*))
	   ;(setq fep:*mode-attr*
	   ;	   (fep:attr-convert fep:*mode-len* str attr))
	   (setq fep:*mode-str*	(code-convert-string str *sjis* *internal*))

	   (setq str	(fep:get-nbyte fep:*system-len*))
	   (setq attr	(fep:get-nbyte fep:*system-len*))
	   (setq fep:*system-attr*
		 (fep:attr-convert fep:*system-len* str attr))
	   (setq fep:*system-str* (code-convert-string str *sjis* *internal*))

	   (setq str	(fep:get-nbyte fep:*buf-len*))
	   (setq attr	(fep:get-nbyte fep:*buf-len*))
	   (setq fep:*buf-attr* (fep:attr-convert fep:*buf-len* str attr))
	   (setq fep:*buf-str*	(code-convert-string str *sjis* *internal*))
           ;(setq fep:*cursor-pos*
           ;	   (fep:pos-convert str fep:*cursor-pos*))
	   ))
     ))
  )

;;;
;;; Get info
;;;
(defun fep:get-info ()
  (let ((init-mode-len) (init-system-len))
    (set-buffer fep:*buffer*)
    (progn
      (fep:get-nbyte 11)
      (setq init-mode-len	(fep:get-int2))
      (setq init-system-len	(fep:get-int2))
      (fep:get-nbyte (* 2 init-mode-len))
      (setq fep:*init-system-str*
	    (code-convert-string (fep:get-nbyte init-system-len)
				 *sjis* *internal*))
      (fep:get-nbyte init-system-len)
      ))
  )

;;;
;;;  Put data into the command buffer 
;;;
(defun fep:put-1byte (integer)
  (insert (logand 255 integer)))

(defun fep:put-int2 (integer)
  (insert (logand 255 (lsh integer -8))
	  (logand 255 integer)))

;;;
;;; Get data from the process buffer
;;;
(defun fep:get-1byte ()
  (let ((c 0) (point (point)))
    (while (< (point-max) (+ point 1))
      (accept-process-output)
      (if (= c 10000) (error "Count exceed."))
      (setq c (1+ c)))
    (goto-char point))
  (let ((point (point)))
    (prog1
	(char-after (point))
      (goto-char (+ (point) 1)))))

(defun fep:get-int2 ()
  (let ((c 0) (point (point)))
    ;;;(goto-char (point-min))
    (while (< (point-max) (+ point 2))
      (accept-process-output)
      (if (= c 10000) (error "Count exceed."))
      (setq c (1+ c)))
    (goto-char point))
  (let ((point (point)))
    (prog1
	(logior 
	 (lsh (char-after point)       8)
	 (lsh (char-after (1+ point)) 0))
      (goto-char (+ (point) 2)))))

(defun fep:get-nbyte (len)
  (let ((point (point)) (c 0))
      (while (< (point-max) (+ point len))
	(accept-process-output)
	(if (= c 10000) (error "Count exceed."))
	(setq c (1+ c)))
      (progn 
	(goto-char (+ point len))
	(buffer-substring point (point)))))

;;;
;;; Count # of char ^H's in the result string from the head.
;;;    Some FEP return ^H as the result string in order to delete
;;;    already fixed string.
;;;
(defun fep:count-ctl-h (str)
  "Count # of ^H's in the result string from the head."
  (let ((len (length str)) (i 0))
    (while (and (< i len) (eq (aref str i) ?\b))
      (setq i (1+ i)))
    i)
  )

;;;
;;; Convert attribute string and cursor positoin for Mule buffer
;;;  Kanji, Hankaku-kana, and ascii character take 3, 2, and 1 byte
;;;  in Mule buffer, respectively.
;;;
(defun fep:attr-convert (len str attr)
  (let ((rtn (make-string (* len 2) 0)) ; enough length
	(i 0) (j 0) (c 0) (d 0))
    (while (< i len)
      (progn
	(setq c (aref str i))
	(setq d (aref attr i))
	(cond
	 ((or (and (<= 129 c) (<= c 159)) (and (<= 224 c) (<= c 252)))
	  ; Sjis-Kanji
	  (aset rtn j d)
	  (aset rtn (+ j 1) d)
	  (aset rtn (+ j 2) d)
	  (setq j (+ j 3))
	  (setq i (+ i 2)))
	 ((and (<= 161 c) (<= c 223))
	  ; Hankaku-Katakana
	  (aset rtn j d)
	  (aset rtn (+ j 1) d)
	  (setq j (+ j 2))
	  (setq i (+ i 1)))
	 (t
	  ; Ascii char
	  (aset rtn j d)
	  (setq j (+ j 1))
	  (setq i (+ i 1))))))
    (substring rtn 0 j)))

(defun fep:pos-convert (str position)
  (let ((pos (min (length str) position))
	(i 0) (j 0) (prev-j 0) (c 0))
    (while (< i pos)
      (progn
	(setq c (aref str i))
	(cond
	 ((or (and (<= 129 c) (<= c 159)) (and (<= 224 c) (<= c 252)))
	  ; Sjis-Kanji
	  (setq prev-j j)
	  (setq j (+ j 3))
	  (setq i (+ i 2)))
	 ((and (<= 161 c) (<= c 223))
	  ; Hankaku-Katakana
	  (setq prev-j j)
	  (setq j (+ j 2))
	  (setq i (+ i 1)))
	 (t
	  ; Ascii char
	  (setq prev-j j)
	  (setq j (+ j 1))
	  (setq i (+ i 1))))))
    prev-j))

;;;
;;; When the selected window is changed by the mouse or else,
;;; turn off japanese mode
;;;
(defun fep:select-window-hook (old new)
  (if (not (eq old new))
      (progn
        ; If minibuffer is in use, retreive old contents to minibuffer
	(if fep:*saved-minibuffer*
	    (progn
	      (set-window-buffer (minibuffer-window) fep:*saved-minibuffer*)
	      (setq fep:*saved-minibuffer nil)))
	(if (and fep:*isearch-japanese-mode*
		 (not (window-minibuffer-p new)))
	    ; isearch mode with an unfixed string in minibuffer
	    (progn
	      (top-level)	; exit from recursive minibuffer edit
	      (setq fep:*isearch-japanese-mode* nil)))
	; Set up for new buffer
	(set-buffer (window-buffer new))
	(if fep:*japanese-mode*
	    (setq fep:*current-buffer* (current-buffer)))
	))
  )

;;;
;;; Fep Bridge initialize
;;;
(defun fep:initialize ()
  "Open connection and initialize variables and buffers for Fep Bridge."
  (if (boundp 'select-window-hook)
      (setq select-window-hook 'fep:select-window-hook))
  (setq fep:*mode-line-format*
	(list "[" 'fep:*mode-str* "]"
	      (cond
	       ((featurep 'egg) (cdr mode-line-format))
	       ((featurep 'canna) (cdr mode-line-format))
	       (t mode-line-format))))
  (fep:make-keymap)
  (setq fep:*menu-buffer* (get-buffer-create " [Fep Bridge Menu Buffer]"))
  (save-excursion
    (set-buffer fep:*menu-buffer*)
    (fep:attribute-off (overlay-lists))
    (use-local-map fep:*mode-map*)
    )
  (setq fep:*saved-minibuffer*	nil)
  (setq fep:*japanese-mode*	nil)
  (setq fep:*buf-seg-list*	nil)
  (setq fep:*system-seg-list*	nil)
  ; Define colors for "region"
  (if (and window-system (cdr (assq 'reverse default-frame-alist)))
      (set-face-foreground 'fep:*region* "black"))
  (if (not (and fep:*process*
		(eq (process-status fep:*process*) 'open)))
      (fep:open))
  )

(defun fep:make-keymap ()
  "Create keymap for Fep Bridge"
  (let ((c 0))
    ; Normal keys
    (while (<= c 127)
      (define-key fep:*mode-map* (char-to-string c) 'fep:key-normal)
      (setq c (1+ c))))

  ; Function keys
  (define-key fep:*mode-map* [backspace] 'fep:key-back)
  (define-key fep:*mode-map* [tab] 'fep:key-tab)
  (define-key fep:*mode-map* [return] 'fep:key-return)
  (define-key fep:*mode-map* [escape] 'fep:key-escape)
  (define-key fep:*mode-map* [prior] 'fep:key-prior)
  (define-key fep:*mode-map* [next] 'fep:key-next)
  (define-key fep:*mode-map* [end] 'fep:key-end)
  (define-key fep:*mode-map* [home] 'fep:key-home)
  (define-key fep:*mode-map* [left] 'fep:key-left)
  (define-key fep:*mode-map* [up] 'fep:key-up)
  (define-key fep:*mode-map* [right] 'fep:key-right)
  (define-key fep:*mode-map* [down] 'fep:key-down)
  (define-key fep:*mode-map* [select] 'fep:key-select)
  (define-key fep:*mode-map* [execute] 'fep:key-execute)
  (define-key fep:*mode-map* [print] 'fep:key-snapshot)
  (define-key fep:*mode-map* [insert] 'fep:key-insert)
  (define-key fep:*mode-map* [delete] 'fep:key-delete)
  (define-key fep:*mode-map* [kp_0] 'fep:key-numpad0)
  (define-key fep:*mode-map* [kp_1] 'fep:key-numpad1)
  (define-key fep:*mode-map* [kp_2] 'fep:key-numpad2)
  (define-key fep:*mode-map* [kp_3] 'fep:key-numpad3)
  (define-key fep:*mode-map* [kp_4] 'fep:key-numpad4)
  (define-key fep:*mode-map* [kp_5] 'fep:key-numpad5)
  (define-key fep:*mode-map* [kp_6] 'fep:key-numpad6)
  (define-key fep:*mode-map* [kp_7] 'fep:key-numpad7)
  (define-key fep:*mode-map* [kp_8] 'fep:key-numpad8)
  (define-key fep:*mode-map* [kp_9] 'fep:key-numpad9)
  (define-key fep:*mode-map* [kp_multiply] 'fep:key-multiply)
  (define-key fep:*mode-map* [kp_add] 'fep:key-add)
  (define-key fep:*mode-map* [kp_separator] 'fep:key-separator)
  (define-key fep:*mode-map* [kp_subtract] 'fep:key-subtract)
  (define-key fep:*mode-map* [kp_decimal] 'fep:key-decimal)
  (define-key fep:*mode-map* [kp_divide] 'fep:key-divide)
  (define-key fep:*mode-map* [f1] 'fep:key-f1)
  (define-key fep:*mode-map* [f2] 'fep:key-f2)
  (define-key fep:*mode-map* [f3] 'fep:key-f3)
  (define-key fep:*mode-map* [f4] 'fep:key-f4)
  (define-key fep:*mode-map* [f5] 'fep:key-f5)
  (define-key fep:*mode-map* [f6] 'fep:key-f6)
  (define-key fep:*mode-map* [f7] 'fep:key-f7)
  (define-key fep:*mode-map* [f8] 'fep:key-f8)
  (define-key fep:*mode-map* [f9] 'fep:key-f9)
  (define-key fep:*mode-map* [f10] 'fep:key-f10)
  (define-key fep:*mode-map* [f11] 'fep:key-f11)
  (define-key fep:*mode-map* [f12] 'fep:key-f12)
  (define-key fep:*mode-map* [f13] 'fep:key-f13)
  (define-key fep:*mode-map* [f14] 'fep:key-f14)
  (define-key fep:*mode-map* [f15] 'fep:key-f15)
  (define-key fep:*mode-map* [f16] 'fep:key-f16)
  (define-key fep:*mode-map* [f17] 'fep:key-f17)
  (define-key fep:*mode-map* [f18] 'fep:key-f18)
  (define-key fep:*mode-map* [f19] 'fep:key-f19)
  (define-key fep:*mode-map* [f20] 'fep:key-f20)
  (define-key fep:*mode-map* [f21] 'fep:key-f21)
  (define-key fep:*mode-map* [f22] 'fep:key-f22)
  (define-key fep:*mode-map* [f23] 'fep:key-f23)
  (define-key fep:*mode-map* [f24] 'fep:key-f24)

  (define-key fep:*mode-map* [S-backspace] 'fep:key-back)
  (define-key fep:*mode-map* [S-tab] 'fep:key-tab)
  (define-key fep:*mode-map* [S-return] 'fep:key-return)
  (define-key fep:*mode-map* [S-escape] 'fep:key-escape)
  (define-key fep:*mode-map* [S-prior] 'fep:key-prior)
  (define-key fep:*mode-map* [S-next] 'fep:key-next)
  (define-key fep:*mode-map* [S-end] 'fep:key-end)
  (define-key fep:*mode-map* [S-home] 'fep:key-home)
  (define-key fep:*mode-map* [S-left] 'fep:key-left)
  (define-key fep:*mode-map* [S-up] 'fep:key-up)
  (define-key fep:*mode-map* [S-right] 'fep:key-right)
  (define-key fep:*mode-map* [S-down] 'fep:key-down)
  (define-key fep:*mode-map* [S-select] 'fep:key-select)
  (define-key fep:*mode-map* [S-execute] 'fep:key-execute)
  (define-key fep:*mode-map* [S-print] 'fep:key-snapshot)
  (define-key fep:*mode-map* [S-insert] 'fep:key-insert)
  (define-key fep:*mode-map* [S-delete] 'fep:key-delete)
  (define-key fep:*mode-map* [S-kp_0] 'fep:key-numpad0)
  (define-key fep:*mode-map* [S-kp_1] 'fep:key-numpad1)
  (define-key fep:*mode-map* [S-kp_2] 'fep:key-numpad2)
  (define-key fep:*mode-map* [S-kp_3] 'fep:key-numpad3)
  (define-key fep:*mode-map* [S-kp_4] 'fep:key-numpad4)
  (define-key fep:*mode-map* [S-kp_5] 'fep:key-numpad5)
  (define-key fep:*mode-map* [S-kp_6] 'fep:key-numpad6)
  (define-key fep:*mode-map* [S-kp_7] 'fep:key-numpad7)
  (define-key fep:*mode-map* [S-kp_8] 'fep:key-numpad8)
  (define-key fep:*mode-map* [S-kp_9] 'fep:key-numpad9)
  (define-key fep:*mode-map* [S-kp_multiply] 'fep:key-multiply)
  (define-key fep:*mode-map* [S-kp_add] 'fep:key-add)
  (define-key fep:*mode-map* [S-kp_separator] 'fep:key-separator)
  (define-key fep:*mode-map* [S-kp_subtract] 'fep:key-subtract)
  (define-key fep:*mode-map* [S-kp_decimal] 'fep:key-decimal)
  (define-key fep:*mode-map* [S-kp_divide] 'fep:key-divide)
  (define-key fep:*mode-map* [S-f1] 'fep:key-f1)
  (define-key fep:*mode-map* [S-f2] 'fep:key-f2)
  (define-key fep:*mode-map* [S-f3] 'fep:key-f3)
  (define-key fep:*mode-map* [S-f4] 'fep:key-f4)
  (define-key fep:*mode-map* [S-f5] 'fep:key-f5)
  (define-key fep:*mode-map* [S-f6] 'fep:key-f6)
  (define-key fep:*mode-map* [S-f7] 'fep:key-f7)
  (define-key fep:*mode-map* [S-f8] 'fep:key-f8)
  (define-key fep:*mode-map* [S-f9] 'fep:key-f9)
  (define-key fep:*mode-map* [S-f10] 'fep:key-f10)
  (define-key fep:*mode-map* [S-f11] 'fep:key-f11)
  (define-key fep:*mode-map* [S-f12] 'fep:key-f12)
  (define-key fep:*mode-map* [S-f13] 'fep:key-f13)
  (define-key fep:*mode-map* [S-f14] 'fep:key-f14)
  (define-key fep:*mode-map* [S-f15] 'fep:key-f15)
  (define-key fep:*mode-map* [S-f16] 'fep:key-f16)
  (define-key fep:*mode-map* [S-f17] 'fep:key-f17)
  (define-key fep:*mode-map* [S-f18] 'fep:key-f18)
  (define-key fep:*mode-map* [S-f19] 'fep:key-f19)
  (define-key fep:*mode-map* [S-f20] 'fep:key-f20)
  (define-key fep:*mode-map* [S-f21] 'fep:key-f21)
  (define-key fep:*mode-map* [S-f22] 'fep:key-f22)
  (define-key fep:*mode-map* [S-f23] 'fep:key-f23)
  (define-key fep:*mode-map* [S-f24] 'fep:key-f24)

  (define-key fep:*mode-map* [C-backspace] 'fep:key-back)
  (define-key fep:*mode-map* [C-tab] 'fep:key-tab)
  (define-key fep:*mode-map* [C-return] 'fep:key-return)
  (define-key fep:*mode-map* [C-escape] 'fep:key-escape)
  (define-key fep:*mode-map* [C-prior] 'fep:key-prior)
  (define-key fep:*mode-map* [C-next] 'fep:key-next)
  (define-key fep:*mode-map* [C-end] 'fep:key-end)
  (define-key fep:*mode-map* [C-home] 'fep:key-home)
  (define-key fep:*mode-map* [C-left] 'fep:key-left)
  (define-key fep:*mode-map* [C-up] 'fep:key-up)
  (define-key fep:*mode-map* [C-right] 'fep:key-right)
  (define-key fep:*mode-map* [C-down] 'fep:key-down)
  (define-key fep:*mode-map* [C-select] 'fep:key-select)
  (define-key fep:*mode-map* [C-execute] 'fep:key-execute)
  (define-key fep:*mode-map* [C-print] 'fep:key-snapshot)
  (define-key fep:*mode-map* [C-insert] 'fep:key-insert)
  (define-key fep:*mode-map* [C-delete] 'fep:key-delete)
  (define-key fep:*mode-map* [C-kp_0] 'fep:key-numpad0)
  (define-key fep:*mode-map* [C-kp_1] 'fep:key-numpad1)
  (define-key fep:*mode-map* [C-kp_2] 'fep:key-numpad2)
  (define-key fep:*mode-map* [C-kp_3] 'fep:key-numpad3)
  (define-key fep:*mode-map* [C-kp_4] 'fep:key-numpad4)
  (define-key fep:*mode-map* [C-kp_5] 'fep:key-numpad5)
  (define-key fep:*mode-map* [C-kp_6] 'fep:key-numpad6)
  (define-key fep:*mode-map* [C-kp_7] 'fep:key-numpad7)
  (define-key fep:*mode-map* [C-kp_8] 'fep:key-numpad8)
  (define-key fep:*mode-map* [C-kp_9] 'fep:key-numpad9)
  (define-key fep:*mode-map* [C-kp_multiply] 'fep:key-multiply)
  (define-key fep:*mode-map* [C-kp_add] 'fep:key-add)
  (define-key fep:*mode-map* [C-kp_separator] 'fep:key-separator)
  (define-key fep:*mode-map* [C-kp_subtract] 'fep:key-subtract)
  (define-key fep:*mode-map* [C-kp_decimal] 'fep:key-decimal)
  (define-key fep:*mode-map* [C-kp_divide] 'fep:key-divide)
  (define-key fep:*mode-map* [C-f1] 'fep:key-f1)
  (define-key fep:*mode-map* [C-f2] 'fep:key-f2)
  (define-key fep:*mode-map* [C-f3] 'fep:key-f3)
  (define-key fep:*mode-map* [C-f4] 'fep:key-f4)
  (define-key fep:*mode-map* [C-f5] 'fep:key-f5)
  (define-key fep:*mode-map* [C-f6] 'fep:key-f6)
  (define-key fep:*mode-map* [C-f7] 'fep:key-f7)
  (define-key fep:*mode-map* [C-f8] 'fep:key-f8)
  (define-key fep:*mode-map* [C-f9] 'fep:key-f9)
  (define-key fep:*mode-map* [C-f10] 'fep:key-f10)
  (define-key fep:*mode-map* [C-f11] 'fep:key-f11)
  (define-key fep:*mode-map* [C-f12] 'fep:key-f12)
  (define-key fep:*mode-map* [C-f13] 'fep:key-f13)
  (define-key fep:*mode-map* [C-f14] 'fep:key-f14)
  (define-key fep:*mode-map* [C-f15] 'fep:key-f15)
  (define-key fep:*mode-map* [C-f16] 'fep:key-f16)
  (define-key fep:*mode-map* [C-f17] 'fep:key-f17)
  (define-key fep:*mode-map* [C-f18] 'fep:key-f18)
  (define-key fep:*mode-map* [C-f19] 'fep:key-f19)
  (define-key fep:*mode-map* [C-f20] 'fep:key-f20)
  (define-key fep:*mode-map* [C-f21] 'fep:key-f21)
  (define-key fep:*mode-map* [C-f22] 'fep:key-f22)
  (define-key fep:*mode-map* [C-f23] 'fep:key-f23)
  (define-key fep:*mode-map* [C-f24] 'fep:key-f24)

  (define-key fep:*mode-map* [C-S-backspace] 'fep:key-back)
  (define-key fep:*mode-map* [C-S-tab] 'fep:key-tab)
  (define-key fep:*mode-map* [C-S-return] 'fep:key-return)
  (define-key fep:*mode-map* [C-S-escape] 'fep:key-escape)
  (define-key fep:*mode-map* [C-S-prior] 'fep:key-prior)
  (define-key fep:*mode-map* [C-S-next] 'fep:key-next)
  (define-key fep:*mode-map* [C-S-end] 'fep:key-end)
  (define-key fep:*mode-map* [C-S-home] 'fep:key-home)
  (define-key fep:*mode-map* [C-S-left] 'fep:key-left)
  (define-key fep:*mode-map* [C-S-up] 'fep:key-up)
  (define-key fep:*mode-map* [C-S-right] 'fep:key-right)
  (define-key fep:*mode-map* [C-S-down] 'fep:key-down)
  (define-key fep:*mode-map* [C-S-select] 'fep:key-select)
  (define-key fep:*mode-map* [C-S-execute] 'fep:key-execute)
  (define-key fep:*mode-map* [C-S-print] 'fep:key-snapshot)
  (define-key fep:*mode-map* [C-S-insert] 'fep:key-insert)
  (define-key fep:*mode-map* [C-S-delete] 'fep:key-delete)
  (define-key fep:*mode-map* [C-S-kp_0] 'fep:key-numpad0)
  (define-key fep:*mode-map* [C-S-kp_1] 'fep:key-numpad1)
  (define-key fep:*mode-map* [C-S-kp_2] 'fep:key-numpad2)
  (define-key fep:*mode-map* [C-S-kp_3] 'fep:key-numpad3)
  (define-key fep:*mode-map* [C-S-kp_4] 'fep:key-numpad4)
  (define-key fep:*mode-map* [C-S-kp_5] 'fep:key-numpad5)
  (define-key fep:*mode-map* [C-S-kp_6] 'fep:key-numpad6)
  (define-key fep:*mode-map* [C-S-kp_7] 'fep:key-numpad7)
  (define-key fep:*mode-map* [C-S-kp_8] 'fep:key-numpad8)
  (define-key fep:*mode-map* [C-S-kp_9] 'fep:key-numpad9)
  (define-key fep:*mode-map* [C-S-kp_multiply] 'fep:key-multiply)
  (define-key fep:*mode-map* [C-S-kp_add] 'fep:key-add)
  (define-key fep:*mode-map* [C-S-kp_separator] 'fep:key-separator)
  (define-key fep:*mode-map* [C-S-kp_subtract] 'fep:key-subtract)
  (define-key fep:*mode-map* [C-S-kp_decimal] 'fep:key-decimal)
  (define-key fep:*mode-map* [C-S-kp_divide] 'fep:key-divide)
  (define-key fep:*mode-map* [C-S-f1] 'fep:key-f1)
  (define-key fep:*mode-map* [C-S-f2] 'fep:key-f2)
  (define-key fep:*mode-map* [C-S-f3] 'fep:key-f3)
  (define-key fep:*mode-map* [C-S-f4] 'fep:key-f4)
  (define-key fep:*mode-map* [C-S-f5] 'fep:key-f5)
  (define-key fep:*mode-map* [C-S-f6] 'fep:key-f6)
  (define-key fep:*mode-map* [C-S-f7] 'fep:key-f7)
  (define-key fep:*mode-map* [C-S-f8] 'fep:key-f8)
  (define-key fep:*mode-map* [C-S-f9] 'fep:key-f9)
  (define-key fep:*mode-map* [C-S-f10] 'fep:key-f10)
  (define-key fep:*mode-map* [C-S-f11] 'fep:key-f11)
  (define-key fep:*mode-map* [C-S-f12] 'fep:key-f12)
  (define-key fep:*mode-map* [C-S-f13] 'fep:key-f13)
  (define-key fep:*mode-map* [C-S-f14] 'fep:key-f14)
  (define-key fep:*mode-map* [C-S-f15] 'fep:key-f15)
  (define-key fep:*mode-map* [C-S-f16] 'fep:key-f16)
  (define-key fep:*mode-map* [C-S-f17] 'fep:key-f17)
  (define-key fep:*mode-map* [C-S-f18] 'fep:key-f18)
  (define-key fep:*mode-map* [C-S-f19] 'fep:key-f19)
  (define-key fep:*mode-map* [C-S-f20] 'fep:key-f20)
  (define-key fep:*mode-map* [C-S-f21] 'fep:key-f21)
  (define-key fep:*mode-map* [C-S-f22] 'fep:key-f22)
  (define-key fep:*mode-map* [C-S-f23] 'fep:key-f23)
  (define-key fep:*mode-map* [C-S-f24] 'fep:key-f24)

  (define-key fep:*mode-map* [M-backspace] 'fep:key-back)
  (define-key fep:*mode-map* [M-tab] 'fep:key-tab)
  (define-key fep:*mode-map* [M-return] 'fep:key-return)
  (define-key fep:*mode-map* [M-escape] 'fep:key-escape)
  (define-key fep:*mode-map* [M-prior] 'fep:key-prior)
  (define-key fep:*mode-map* [M-next] 'fep:key-next)
  (define-key fep:*mode-map* [M-end] 'fep:key-end)
  (define-key fep:*mode-map* [M-home] 'fep:key-home)
  (define-key fep:*mode-map* [M-left] 'fep:key-left)
  (define-key fep:*mode-map* [M-up] 'fep:key-up)
  (define-key fep:*mode-map* [M-right] 'fep:key-right)
  (define-key fep:*mode-map* [M-down] 'fep:key-down)
  (define-key fep:*mode-map* [M-select] 'fep:key-select)
  (define-key fep:*mode-map* [M-execute] 'fep:key-execute)
  (define-key fep:*mode-map* [M-print] 'fep:key-snapshot)
  (define-key fep:*mode-map* [M-insert] 'fep:key-insert)
  (define-key fep:*mode-map* [M-delete] 'fep:key-delete)
  (define-key fep:*mode-map* [M-kp_0] 'fep:key-numpad0)
  (define-key fep:*mode-map* [M-kp_1] 'fep:key-numpad1)
  (define-key fep:*mode-map* [M-kp_2] 'fep:key-numpad2)
  (define-key fep:*mode-map* [M-kp_3] 'fep:key-numpad3)
  (define-key fep:*mode-map* [M-kp_4] 'fep:key-numpad4)
  (define-key fep:*mode-map* [M-kp_5] 'fep:key-numpad5)
  (define-key fep:*mode-map* [M-kp_6] 'fep:key-numpad6)
  (define-key fep:*mode-map* [M-kp_7] 'fep:key-numpad7)
  (define-key fep:*mode-map* [M-kp_8] 'fep:key-numpad8)
  (define-key fep:*mode-map* [M-kp_9] 'fep:key-numpad9)
  (define-key fep:*mode-map* [M-kp_multiply] 'fep:key-multiply)
  (define-key fep:*mode-map* [M-kp_add] 'fep:key-add)
  (define-key fep:*mode-map* [M-kp_separator] 'fep:key-separator)
  (define-key fep:*mode-map* [M-kp_subtract] 'fep:key-subtract)
  (define-key fep:*mode-map* [M-kp_decimal] 'fep:key-decimal)
  (define-key fep:*mode-map* [M-kp_divide] 'fep:key-divide)
  (define-key fep:*mode-map* [M-f1] 'fep:key-f1)
  (define-key fep:*mode-map* [M-f2] 'fep:key-f2)
  (define-key fep:*mode-map* [M-f3] 'fep:key-f3)
  (define-key fep:*mode-map* [M-f4] 'fep:key-f4)
  (define-key fep:*mode-map* [M-f5] 'fep:key-f5)
  (define-key fep:*mode-map* [M-f6] 'fep:key-f6)
  (define-key fep:*mode-map* [M-f7] 'fep:key-f7)
  (define-key fep:*mode-map* [M-f8] 'fep:key-f8)
  (define-key fep:*mode-map* [M-f9] 'fep:key-f9)
  (define-key fep:*mode-map* [M-f10] 'fep:key-f10)
  (define-key fep:*mode-map* [M-f11] 'fep:key-f11)
  (define-key fep:*mode-map* [M-f12] 'fep:key-f12)
  (define-key fep:*mode-map* [M-f13] 'fep:key-f13)
  (define-key fep:*mode-map* [M-f14] 'fep:key-f14)
  (define-key fep:*mode-map* [M-f15] 'fep:key-f15)
  (define-key fep:*mode-map* [M-f16] 'fep:key-f16)
  (define-key fep:*mode-map* [M-f17] 'fep:key-f17)
  (define-key fep:*mode-map* [M-f18] 'fep:key-f18)
  (define-key fep:*mode-map* [M-f19] 'fep:key-f19)
  (define-key fep:*mode-map* [M-f20] 'fep:key-f20)
  (define-key fep:*mode-map* [M-f21] 'fep:key-f21)
  (define-key fep:*mode-map* [M-f22] 'fep:key-f22)
  (define-key fep:*mode-map* [M-f23] 'fep:key-f23)
  (define-key fep:*mode-map* [M-f24] 'fep:key-f24)

  (define-key fep:*mode-map* [M-S-backspace] 'fep:key-back)
  (define-key fep:*mode-map* [M-S-tab] 'fep:key-tab)
  (define-key fep:*mode-map* [M-S-return] 'fep:key-return)
  (define-key fep:*mode-map* [M-S-escape] 'fep:key-escape)
  (define-key fep:*mode-map* [M-S-prior] 'fep:key-prior)
  (define-key fep:*mode-map* [M-S-next] 'fep:key-next)
  (define-key fep:*mode-map* [M-S-end] 'fep:key-end)
  (define-key fep:*mode-map* [M-S-home] 'fep:key-home)
  (define-key fep:*mode-map* [M-S-left] 'fep:key-left)
  (define-key fep:*mode-map* [M-S-up] 'fep:key-up)
  (define-key fep:*mode-map* [M-S-right] 'fep:key-right)
  (define-key fep:*mode-map* [M-S-down] 'fep:key-down)
  (define-key fep:*mode-map* [M-S-select] 'fep:key-select)
  (define-key fep:*mode-map* [M-S-execute] 'fep:key-execute)
  (define-key fep:*mode-map* [M-S-print] 'fep:key-snapshot)
  (define-key fep:*mode-map* [M-S-insert] 'fep:key-insert)
  (define-key fep:*mode-map* [M-S-delete] 'fep:key-delete)
  (define-key fep:*mode-map* [M-S-kp_0] 'fep:key-numpad0)
  (define-key fep:*mode-map* [M-S-kp_1] 'fep:key-numpad1)
  (define-key fep:*mode-map* [M-S-kp_2] 'fep:key-numpad2)
  (define-key fep:*mode-map* [M-S-kp_3] 'fep:key-numpad3)
  (define-key fep:*mode-map* [M-S-kp_4] 'fep:key-numpad4)
  (define-key fep:*mode-map* [M-S-kp_5] 'fep:key-numpad5)
  (define-key fep:*mode-map* [M-S-kp_6] 'fep:key-numpad6)
  (define-key fep:*mode-map* [M-S-kp_7] 'fep:key-numpad7)
  (define-key fep:*mode-map* [M-S-kp_8] 'fep:key-numpad8)
  (define-key fep:*mode-map* [M-S-kp_9] 'fep:key-numpad9)
  (define-key fep:*mode-map* [M-S-kp_multiply] 'fep:key-multiply)
  (define-key fep:*mode-map* [M-S-kp_add] 'fep:key-add)
  (define-key fep:*mode-map* [M-S-kp_separator] 'fep:key-separator)
  (define-key fep:*mode-map* [M-S-kp_subtract] 'fep:key-subtract)
  (define-key fep:*mode-map* [M-S-kp_decimal] 'fep:key-decimal)
  (define-key fep:*mode-map* [M-S-kp_divide] 'fep:key-divide)
  (define-key fep:*mode-map* [M-S-f1] 'fep:key-f1)
  (define-key fep:*mode-map* [M-S-f2] 'fep:key-f2)
  (define-key fep:*mode-map* [M-S-f3] 'fep:key-f3)
  (define-key fep:*mode-map* [M-S-f4] 'fep:key-f4)
  (define-key fep:*mode-map* [M-S-f5] 'fep:key-f5)
  (define-key fep:*mode-map* [M-S-f6] 'fep:key-f6)
  (define-key fep:*mode-map* [M-S-f7] 'fep:key-f7)
  (define-key fep:*mode-map* [M-S-f8] 'fep:key-f8)
  (define-key fep:*mode-map* [M-S-f9] 'fep:key-f9)
  (define-key fep:*mode-map* [M-S-f10] 'fep:key-f10)
  (define-key fep:*mode-map* [M-S-f11] 'fep:key-f11)
  (define-key fep:*mode-map* [M-S-f12] 'fep:key-f12)
  (define-key fep:*mode-map* [M-S-f13] 'fep:key-f13)
  (define-key fep:*mode-map* [M-S-f14] 'fep:key-f14)
  (define-key fep:*mode-map* [M-S-f15] 'fep:key-f15)
  (define-key fep:*mode-map* [M-S-f16] 'fep:key-f16)
  (define-key fep:*mode-map* [M-S-f17] 'fep:key-f17)
  (define-key fep:*mode-map* [M-S-f18] 'fep:key-f18)
  (define-key fep:*mode-map* [M-S-f19] 'fep:key-f19)
  (define-key fep:*mode-map* [M-S-f20] 'fep:key-f20)
  (define-key fep:*mode-map* [M-S-f21] 'fep:key-f21)
  (define-key fep:*mode-map* [M-S-f22] 'fep:key-f22)
  (define-key fep:*mode-map* [M-S-f23] 'fep:key-f23)
  (define-key fep:*mode-map* [M-S-f24] 'fep:key-f24)

  (define-key fep:*mode-map* [M-C-backspace] 'fep:key-back)
  (define-key fep:*mode-map* [M-C-tab] 'fep:key-tab)
  (define-key fep:*mode-map* [M-C-return] 'fep:key-return)
  (define-key fep:*mode-map* [M-C-escape] 'fep:key-escape)
  (define-key fep:*mode-map* [M-C-prior] 'fep:key-prior)
  (define-key fep:*mode-map* [M-C-next] 'fep:key-next)
  (define-key fep:*mode-map* [M-C-end] 'fep:key-end)
  (define-key fep:*mode-map* [M-C-home] 'fep:key-home)
  (define-key fep:*mode-map* [M-C-left] 'fep:key-left)
  (define-key fep:*mode-map* [M-C-up] 'fep:key-up)
  (define-key fep:*mode-map* [M-C-right] 'fep:key-right)
  (define-key fep:*mode-map* [M-C-down] 'fep:key-down)
  (define-key fep:*mode-map* [M-C-select] 'fep:key-select)
  (define-key fep:*mode-map* [M-C-execute] 'fep:key-execute)
  (define-key fep:*mode-map* [M-C-print] 'fep:key-snapshot)
  (define-key fep:*mode-map* [M-C-insert] 'fep:key-insert)
  (define-key fep:*mode-map* [M-C-delete] 'fep:key-delete)
  (define-key fep:*mode-map* [M-C-kp_0] 'fep:key-numpad0)
  (define-key fep:*mode-map* [M-C-kp_1] 'fep:key-numpad1)
  (define-key fep:*mode-map* [M-C-kp_2] 'fep:key-numpad2)
  (define-key fep:*mode-map* [M-C-kp_3] 'fep:key-numpad3)
  (define-key fep:*mode-map* [M-C-kp_4] 'fep:key-numpad4)
  (define-key fep:*mode-map* [M-C-kp_5] 'fep:key-numpad5)
  (define-key fep:*mode-map* [M-C-kp_6] 'fep:key-numpad6)
  (define-key fep:*mode-map* [M-C-kp_7] 'fep:key-numpad7)
  (define-key fep:*mode-map* [M-C-kp_8] 'fep:key-numpad8)
  (define-key fep:*mode-map* [M-C-kp_9] 'fep:key-numpad9)
  (define-key fep:*mode-map* [M-C-kp_multiply] 'fep:key-multiply)
  (define-key fep:*mode-map* [M-C-kp_add] 'fep:key-add)
  (define-key fep:*mode-map* [M-C-kp_separator] 'fep:key-separator)
  (define-key fep:*mode-map* [M-C-kp_subtract] 'fep:key-subtract)
  (define-key fep:*mode-map* [M-C-kp_decimal] 'fep:key-decimal)
  (define-key fep:*mode-map* [M-C-kp_divide] 'fep:key-divide)
  (define-key fep:*mode-map* [M-C-f1] 'fep:key-f1)
  (define-key fep:*mode-map* [M-C-f2] 'fep:key-f2)
  (define-key fep:*mode-map* [M-C-f3] 'fep:key-f3)
  (define-key fep:*mode-map* [M-C-f4] 'fep:key-f4)
  (define-key fep:*mode-map* [M-C-f5] 'fep:key-f5)
  (define-key fep:*mode-map* [M-C-f6] 'fep:key-f6)
  (define-key fep:*mode-map* [M-C-f7] 'fep:key-f7)
  (define-key fep:*mode-map* [M-C-f8] 'fep:key-f8)
  (define-key fep:*mode-map* [M-C-f9] 'fep:key-f9)
  (define-key fep:*mode-map* [M-C-f10] 'fep:key-f10)
  (define-key fep:*mode-map* [M-C-f11] 'fep:key-f11)
  (define-key fep:*mode-map* [M-C-f12] 'fep:key-f12)
  (define-key fep:*mode-map* [M-C-f13] 'fep:key-f13)
  (define-key fep:*mode-map* [M-C-f14] 'fep:key-f14)
  (define-key fep:*mode-map* [M-C-f15] 'fep:key-f15)
  (define-key fep:*mode-map* [M-C-f16] 'fep:key-f16)
  (define-key fep:*mode-map* [M-C-f17] 'fep:key-f17)
  (define-key fep:*mode-map* [M-C-f18] 'fep:key-f18)
  (define-key fep:*mode-map* [M-C-f19] 'fep:key-f19)
  (define-key fep:*mode-map* [M-C-f20] 'fep:key-f20)
  (define-key fep:*mode-map* [M-C-f21] 'fep:key-f21)
  (define-key fep:*mode-map* [M-C-f22] 'fep:key-f22)
  (define-key fep:*mode-map* [M-C-f23] 'fep:key-f23)
  (define-key fep:*mode-map* [M-C-f24] 'fep:key-f24)

  (define-key fep:*mode-map* [M-C-S-backspace] 'fep:key-back)
  (define-key fep:*mode-map* [M-C-S-tab] 'fep:key-tab)
  (define-key fep:*mode-map* [M-C-S-return] 'fep:key-return)
  (define-key fep:*mode-map* [M-C-S-escape] 'fep:key-escape)
  (define-key fep:*mode-map* [M-C-S-prior] 'fep:key-prior)
  (define-key fep:*mode-map* [M-C-S-next] 'fep:key-next)
  (define-key fep:*mode-map* [M-C-S-end] 'fep:key-end)
  (define-key fep:*mode-map* [M-C-S-home] 'fep:key-home)
  (define-key fep:*mode-map* [M-C-S-left] 'fep:key-left)
  (define-key fep:*mode-map* [M-C-S-up] 'fep:key-up)
  (define-key fep:*mode-map* [M-C-S-right] 'fep:key-right)
  (define-key fep:*mode-map* [M-C-S-down] 'fep:key-down)
  (define-key fep:*mode-map* [M-C-S-select] 'fep:key-select)
  (define-key fep:*mode-map* [M-C-S-execute] 'fep:key-execute)
  (define-key fep:*mode-map* [M-C-S-print] 'fep:key-snapshot)
  (define-key fep:*mode-map* [M-C-S-insert] 'fep:key-insert)
  (define-key fep:*mode-map* [M-C-S-delete] 'fep:key-delete)
  (define-key fep:*mode-map* [M-C-S-kp_0] 'fep:key-numpad0)
  (define-key fep:*mode-map* [M-C-S-kp_1] 'fep:key-numpad1)
  (define-key fep:*mode-map* [M-C-S-kp_2] 'fep:key-numpad2)
  (define-key fep:*mode-map* [M-C-S-kp_3] 'fep:key-numpad3)
  (define-key fep:*mode-map* [M-C-S-kp_4] 'fep:key-numpad4)
  (define-key fep:*mode-map* [M-C-S-kp_5] 'fep:key-numpad5)
  (define-key fep:*mode-map* [M-C-S-kp_6] 'fep:key-numpad6)
  (define-key fep:*mode-map* [M-C-S-kp_7] 'fep:key-numpad7)
  (define-key fep:*mode-map* [M-C-S-kp_8] 'fep:key-numpad8)
  (define-key fep:*mode-map* [M-C-S-kp_9] 'fep:key-numpad9)
  (define-key fep:*mode-map* [M-C-S-kp_multiply] 'fep:key-multiply)
  (define-key fep:*mode-map* [M-C-S-kp_add] 'fep:key-add)
  (define-key fep:*mode-map* [M-C-S-kp_separator] 'fep:key-separator)
  (define-key fep:*mode-map* [M-C-S-kp_subtract] 'fep:key-subtract)
  (define-key fep:*mode-map* [M-C-S-kp_decimal] 'fep:key-decimal)
  (define-key fep:*mode-map* [M-C-S-kp_divide] 'fep:key-divide)
  (define-key fep:*mode-map* [M-C-S-f1] 'fep:key-f1)
  (define-key fep:*mode-map* [M-C-S-f2] 'fep:key-f2)
  (define-key fep:*mode-map* [M-C-S-f3] 'fep:key-f3)
  (define-key fep:*mode-map* [M-C-S-f4] 'fep:key-f4)
  (define-key fep:*mode-map* [M-C-S-f5] 'fep:key-f5)
  (define-key fep:*mode-map* [M-C-S-f6] 'fep:key-f6)
  (define-key fep:*mode-map* [M-C-S-f7] 'fep:key-f7)
  (define-key fep:*mode-map* [M-C-S-f8] 'fep:key-f8)
  (define-key fep:*mode-map* [M-C-S-f9] 'fep:key-f9)
  (define-key fep:*mode-map* [M-C-S-f10] 'fep:key-f10)
  (define-key fep:*mode-map* [M-C-S-f11] 'fep:key-f11)
  (define-key fep:*mode-map* [M-C-S-f12] 'fep:key-f12)
  (define-key fep:*mode-map* [M-C-S-f13] 'fep:key-f13)
  (define-key fep:*mode-map* [M-C-S-f14] 'fep:key-f14)
  (define-key fep:*mode-map* [M-C-S-f15] 'fep:key-f15)
  (define-key fep:*mode-map* [M-C-S-f16] 'fep:key-f16)
  (define-key fep:*mode-map* [M-C-S-f17] 'fep:key-f17)
  (define-key fep:*mode-map* [M-C-S-f18] 'fep:key-f18)
  (define-key fep:*mode-map* [M-C-S-f19] 'fep:key-f19)
  (define-key fep:*mode-map* [M-C-S-f20] 'fep:key-f20)
  (define-key fep:*mode-map* [M-C-S-f21] 'fep:key-f21)
  (define-key fep:*mode-map* [M-C-S-f22] 'fep:key-f22)
  (define-key fep:*mode-map* [M-C-S-f23] 'fep:key-f23)
  (define-key fep:*mode-map* [M-C-S-f24] 'fep:key-f24)

  ; Setup for incremental search: it is still buggy !
  (fep:isearch-initialize)
  
  ; Keys to turn on/off Fep Bridge
  (cond
   ((let ((keys fep:*conversion-key-list*) (ok nil))
     (while keys
       (progn
	 (global-set-key (car keys) 'fep:toggle-japanese-mode)
	 (define-key fep:*mode-map* (car keys) 'fep:toggle-japanese-mode)
	 (define-key isearch-mode-map (car keys) 'fep:isearch-fep)
	 (setq keys (cdr keys))
	 (setq ok t))
       ) ok))
   (t ; Default key
    (global-set-key "\C-\\" 'fep:toggle-japanese-mode)
    (define-key fep:*mode-map* "\C-\\" 'fep:toggle-japanese-mode)
    (define-key isearch-mode-map "\C-\\" 'fep:isearch-fep)))

  ; Unmap keys. These keys are not sent to Fep Bridge.
  (let ((keys fep:*unmap-key-list*))
     (while keys
       (progn
	 (define-key fep:*mode-map* (car keys) 'fep:key-unmap)
	 (setq keys (cdr keys))
	 (setq ok t))
       ))
  )

;;;
;;; isearch
;;;	I've stolen some code from isearch-ext.el of Mule2.3
;;;
(defun fep:isearch-initialize ()
  "Setup for isearch command."
  (require 'isearch-ext)	; load isearch-ext.el
  ; Overwrite isearch-fep-table defined in "isearch-ext.el"
  (if (not (boundp 'isearch-fep-table))
      (setq isearch-fep-table
       '((isearch-fep-string isearch-fep-prompt-string isearch-fep-read-string)
	 (isearch-fep-egg    isearch-fep-prompt-egg    isearch-fep-read-egg)
	 (isearch-fep-canna  isearch-fep-prompt-canna  isearch-fep-read-canna)
	 (isearch-fep-quail  isearch-fep-prompt-quail  isearch-fep-read-quail))
       ))
  (setq isearch-fep-table
      (append isearch-fep-table
	  '((fep:isearch-fep fep:isearch-fep-prompt fep:isearch-fep-read))))
  )

(defun fep:isearch-fep ()
  "Read string for incremental search by using Fep Bridge."
  (interactive)
  (isearch-fep-mode)
  )

(defun fep:isearch-fep-prompt ()
  (if fep:*full-mode-line-in-minibuffer*
      (format "[%s]" fep:*mode-str*)
    (format "[]"))
  )

(defun fep:isearch-exit-minibuffer (from to)
  ; Turn Off Japanese mode
  (setq fep:*isearch-japanese-mode* nil)
  (setq minibuffer-preprompt nil)
  (exit-minibuffer))

(defun fep:isearch-minibuffer-setup ()
  (setq fep:*current-buffer* (current-buffer))
  (setq fep:*display-buffer* (window-buffer (minibuffer-window)))
  )

(defun fep:isearch-minibuffer-exit ()
  ; nop
  )

(defun fep:isearch-fep-read (first-str)
  (if (= (minibuffer-depth) 0)
      (save-excursion
	(set-buffer (window-buffer (minibuffer-window)))
	(let ((display-minibuffer-mode-in-minibuffer t)
	      (fep:*isearch-japanese-mode* t)
	      ;; In GNU Emacs 19.29 or later, isearch.el doesn't
	      ;; make overriding-terminal-local-map buffer local.
	      ;; So we must bind explicitly it to nil to disable it
	      ;; while reading from minibuffer.
	      (overriding-terminal-local-map nil)
	      (self-insert-after-hook 'fep:isearch-exit-minibuffer))
	  (setq unread-command-events (listify-key-sequence first-str))
	  (set-marker fep:*preedit-start* (point))
	  (set-marker fep:*preedit-end* (point))
	  (fep:reset)
	  (fep:display-preedit)
	  (unwind-protect
	      (progn
		(add-hook 'minibuffer-setup-hook 'fep:isearch-minibuffer-setup)
		(add-hook 'minibuffer-exit-hook 'fep:isearch-minibuffer-exit)
		(read-from-minibuffer (isearch-message) nil fep:*mode-map*))
	    (remove-hook 'minibuffer-setup-hook 'fep:isearch-minibuffer-setup)
	    (remove-hook 'minibuffer-exit-hook 'fep:isearch-minibuffer-exit)
	    (fep:delete-preedit)
	    (if fep:*saved-minibuffer*
		;display the saved-buffer in minibuffer
		(progn
		  (set-window-buffer (minibuffer-window) fep:*saved-minibuffer*)
		  (setq fep:*saved-minibuffer nil)
		  ))
	    (setq disable-undo nil))))
    ""))

;;;
;;; Run
;;;
(if (and (boundp 'MULE)
	 (boundp 'emacs-major-version) (>= emacs-major-version 19))
    (fep:initialize)
  (error "This version of Emacs cannot use \"fep.el\".")
  )
